# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
from django.views.decorators.http import require_GET
from django.shortcuts import get_object_or_404

from common.utils import (
    JSONResponse,
    MySerializer,
)

from bootspeed.models import (
    Image,
    ImageResult,
    Machine,
    ImageMachineAgg,
)

MAX_RESULTS = 180


@require_GET
def overall(request, arch):
    # do some reversed sorting to get the latest images.
    images = reversed(
        Image.objects.filter(arch=arch, image_type='daily').order_by(
            '-build_number')[:MAX_RESULTS]
    )

    images_data = []
    for image in images:
        # Ignore images without results.
        if image.result_count() == 0:
            continue
        images_data.append({
            'min_boot': image.min_boot,
            'max_boot': image.max_boot,
            'avg_boot': image.avg_boot,
            'bottom': image.min_boot,
            'middle': image.avg_boot - image.min_boot,
            'top': image.max_boot - image.avg_boot,
            'build_number': image.build_number,
        })

    data = json.dumps(images_data)

    return JSONResponse(data)


@require_GET
def machine_list(request):
    machines = Machine.objects.all()

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize(machines)

    return JSONResponse(data)


@require_GET
def machine_detail(request, machine_id):
    machine = get_object_or_404(Machine, id=machine_id)

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize([machine])

    return JSONResponse(data)


@require_GET
def machine_results(request, machine_id, arch):
    res = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='daily'
    ).order_by(
        'image__build_number'
    )[:MAX_RESULTS]

    last_releases = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='release',
    ).order_by(
        '-image__build_number'
    )
    if last_releases.count() > 0:
        last_release_avg = last_releases[0].boot_avg
        last_release_name = last_releases[0].image.release
    else:
        last_release_avg = 0.0
        last_release_name = 'Unknown'

    bundle = {
        'last_release_avg': last_release_avg,
        'last_release_name': last_release_name,
    }

    image_data = []
    for r in res:
        image_info = dict(
            bottom=r.boot_min,
            min=r.boot_min,
            max=r.boot_max,
            avg=r.boot_avg,
            std=r.boot_std,
            build_number=r.image.build_number,
        )
        image_data.append(image_info)

    bundle['count'] = res.count()
    bundle['image_data'] = image_data
    data = json.dumps(bundle)

    return JSONResponse(data)


@require_GET
def machine_kernel_results(request, machine_id, arch):
    res = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='daily',
    ).order_by('image__build_number')[:MAX_RESULTS]

    last_releases = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='release',
    ).order_by('-image__build_number')

    if last_releases.count() > 0:
        last_release_avg = last_releases[0].kernel_avg
        last_release_name = last_releases[0].image.release
    else:
        last_release_avg = 0.0
        last_release_name = 'Unknown'

    bundle = {
        'last_release_avg': last_release_avg,
        'last_release_name': last_release_name,
    }

    image_data = []
    for r in res:
        image_info = dict(
            bottom=r.kernel_min,
            min=r.kernel_min,
            max=r.kernel_max,
            avg=r.kernel_avg,
            std=r.kernel_std,
            build_number=r.image.build_number,
        )
        image_data.append(image_info)

    bundle['count'] = res.count()
    bundle['image_data'] = image_data
    data = json.dumps(bundle)

    return JSONResponse(data)


@require_GET
def machine_kernel_init_results(request, machine_id, arch):
    res = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='daily',
    ).order_by('image__build_number')[:MAX_RESULTS]

    last_releases = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='release',
    ).order_by('-image__build_number')

    if last_releases.count() > 0:
        last_release_avg = last_releases[0].kernel_init_avg
        last_release_name = last_releases[0].image.release
    else:
        last_release_avg = 0.0
        last_release_name = 'Unknown'

    bundle = {
        'last_release_avg': last_release_avg,
        'last_release_name': last_release_name,
    }

    image_data = []
    for r in res:
        image_info = dict(
            bottom=r.kernel_init_min,
            min=r.kernel_init_min,
            max=r.kernel_init_max,
            avg=r.kernel_init_avg,
            std=r.kernel_init_std,
            build_number=r.image.build_number,
        )
        image_data.append(image_info)

    bundle['count'] = res.count()
    bundle['image_data'] = image_data
    data = json.dumps(bundle)

    return JSONResponse(data)


@require_GET
def machine_plumbing_results(request, machine_id, arch):
    res = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='daily',
    ).order_by('image__build_number')[:MAX_RESULTS]

    last_releases = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='release',
    ).order_by('-image__build_number')

    if last_releases.count() > 0:
        last_release_avg = last_releases[0].plumbing_avg
        last_release_name = last_releases[0].image.release
    else:
        last_release_avg = 0.0
        last_release_name = 'Unknown'

    bundle = {
        'last_release_avg': last_release_avg,
        'last_release_name': last_release_name,
    }

    image_data = []
    for r in res:
        image_info = dict(
            bottom=r.plumbing_min,
            min=r.plumbing_min,
            max=r.plumbing_max,
            avg=r.plumbing_avg,
            std=r.plumbing_std,
            build_number=r.image.build_number,
        )
        image_data.append(image_info)

    bundle['count'] = res.count()
    bundle['image_data'] = image_data
    data = json.dumps(bundle)

    return JSONResponse(data)


@require_GET
def machine_xorg_results(request, machine_id, arch):
    res = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='daily',
    ).order_by('image__build_number')[:MAX_RESULTS]

    last_releases = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='release',
    ).order_by('-image__build_number')

    if last_releases.count() > 0:
        last_release_avg = last_releases[0].xorg_avg
        last_release_name = last_releases[0].image.release
    else:
        last_release_avg = 0.0
        last_release_name = 'Unknown'

    bundle = {
        'last_release_avg': last_release_avg,
        'last_release_name': last_release_name,
    }

    image_data = []
    for r in res:
        image_info = dict(
            bottom=r.xorg_min,
            min=r.xorg_min,
            max=r.xorg_max,
            avg=r.xorg_avg,
            std=r.xorg_std,
            build_number=r.image.build_number,
        )
        image_data.append(image_info)

    bundle['count'] = res.count()
    bundle['image_data'] = image_data
    data = json.dumps(bundle)

    return JSONResponse(data)


@require_GET
def machine_desktop_results(request, machine_id, arch):
    res = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='daily',
    ).order_by('image__build_number')[:MAX_RESULTS]

    last_releases = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='release',
    ).order_by('-image__build_number')

    if last_releases.count() > 0:
        last_release_avg = last_releases[0].desktop_avg
        last_release_name = last_releases[0].image.release
    else:
        last_release_avg = 0.0
        last_release_name = 'Unknown'

    bundle = {
        'last_release_avg': last_release_avg,
        'last_release_name': last_release_name,
    }

    image_data = []
    for r in res:
        image_info = dict(
            bottom=r.desktop_min,
            min=r.desktop_min,
            max=r.desktop_max,
            avg=r.desktop_avg,
            std=r.desktop_std,
            build_number=r.image.build_number,
        )
        image_data.append(image_info)

    bundle['count'] = res.count()
    bundle['image_data'] = image_data
    data = json.dumps(bundle)

    return JSONResponse(data)


@require_GET
def machine_xorg_plumbing_results(request, machine_id, arch):
    res = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
    ).order_by('image__build_number')[:MAX_RESULTS]

    last_releases = ImageMachineAgg.objects.filter(
        machine=machine_id,
        image__arch=arch,
        image__image_type='release',
    ).order_by('-image__build_number')

    if last_releases.count() > 0:
        last_release_avg = last_releases[0].xorg_plumbing_avg
        last_release_name = last_releases[0].image.release
    else:
        last_release_avg = 0.0
        last_release_name = 'Unknown'

    bundle = {
        'last_release_avg': last_release_avg,
        'last_release_name': last_release_name,
    }

    image_data = []
    for r in res:
        image_info = dict(
            bottom=r.xorg_plumbing_min,
            min=r.xorg_plumbing_min,
            max=r.xorg_plumbing_max,
            avg=r.xorg_plumbing_avg,
            std=r.xorg_plumbing_std,
            build_number=r.image.build_number,
        )
        image_data.append(image_info)

    bundle['count'] = res.count()
    bundle['image_data'] = image_data

    data = json.dumps(bundle)

    return JSONResponse(data)


@require_GET
def image_list(request):
    images = Image.objects.all()

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize(images)

    return JSONResponse(data)


@require_GET
def image_detail(request, image_id):
    image = get_object_or_404(Image, id=image_id)

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize([image])

    return JSONResponse(data)


@require_GET
def image_result_list(request, image_id):
    results = ImageResult.objects.filter(image=image_id)

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize(results)

    return JSONResponse(data)


@require_GET
def arch_list(request):
    """
    List images filtered by arch (i.e. i386, amd64)
    """
    image_arches = Image.objects.values('arch').distinct()

    arches = []
    for arch in image_arches:
        arches.append(arch['arch'])

    data = json.dumps({'arches': arches})

    return JSONResponse(data)


@require_GET
def result_count(request):
    error_data = {
        'result': 'error',
        'message': 'failure',
    }

    machine_name = request.GET.get('machine')
    machine_id = request.GET.get('machine_id')
    md5 = request.GET.get('md5')

    if (machine_id is None and machine_name is None) or md5 is None:
        error_data['message'] = "both machine and md5 are required"
        return JSONResponse(json.dumps(error_data))

    machine_token = None
    if machine_id is not None:
        machine_token = machine_id
        machines = Machine.objects.filter(id=machine_id)
    else:
        machine_token = machine_name
        machines = Machine.objects.filter(name=machine_name)

    if machines.count() != 1:
        error_data['message'] = "unknown machine: {}".format(machine_token)
        return JSONResponse(json.dumps(error_data))

    machine = machines[0]
    image = Image.objects.filter(md5=md5)

    if image.count() != 1:
        error_data['message'] = "unknown md5: {}".format(md5)
        return JSONResponse(json.dumps(error_data))

    image_results = image.imageresult_set.filter(machine=machine)

    data = json.dumps({
        "image_id": image.id,
        "machine_id": machine.id,
        "result_count": image_results.count(),
    })

    return JSONResponse(data)
