# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re

from performance.management import (
    get_image_build_number,
    get_run_number,
)

from common.management import (
    jenkins_get,
    JenkinsBaseCommand,
)

from common.utils import regexes

from bootspeed.models import (
    Image,
    Machine,
    JobImage,
    ImageResult,
)

from bootspeed.utils import (
    verify_data,
    MissingField,
    BadMD5,
)


def _get_bootspeed_logs(artifacts):
    """ Check for valid bootspeed results.

    :returns: a list of bootspeed logs.

    """

    logs = {}
    json_regex = re.compile(ur'^(\d*)\/boot.json')
    log_regex = re.compile(
        ur'^(\d*)\/(boot(chart|graph).(png|svg|tgz)|dmesg|syslog|times)'
    )

    for artifact in artifacts:
        path = artifact['relativePath']
        num = get_run_number(path)
        if num is not None and num not in logs:
            logs[num] = {
                'data_file': None,
                'logs': [],
            }

        logging.debug("artifact: {}".format(path))
        if json_regex.match(path):
            logging.debug("found {}".format(path))
            logs[num]['data_file'] = artifact
        if log_regex.match(path):
            logging.debug("found bootspeed log {}".format(path))
            logs[num]['logs'].append(artifact)

    return logs


class Command(JenkinsBaseCommand):

    job_regex = regexes['bootspeed']
    job_types = ['milestone', 'backfill']

    def extract_data(self, name):
        m = self.job_regex.match(name)

        if m:
            self.job_type = m.group(1)
            self.variant = m.group(2)
            self.arch = m.group(3)
            self.machine_name = m.group(4)

            self.install_data = dict(
                job_type=self.job_type,
                variant=self.variant,
                arch=self.arch,
                machine=self.machine_name,
            )

    def process_build(self, build):
        self.image_build_number = get_image_build_number(build)
        if self.job_type in self.job_types:
            self.release = self.job_type
        else:
            self.release = 'UNKNOWN'
        self.job_image, new_job_image = JobImage.objects.get_or_create(
            arch=self.arch,
            build_number=self.image_build_number,
            release=self.release,
            variant=self.variant,
        )
        build_set = self.job_image.imagebuild_set
        image_build, new_image_build = build_set.get_or_create(
            build=self.install_data['jenkins_build'],
        )

    def process_job(self, job):
        self.machine, new_machine = Machine.objects.get_or_create(
            name=self.machine_name,
        )

    def _process_logs(self, build, build_date, install_data, func):

        if not hasattr(func, '__call__'):
            raise Exception("invalid callable: %s", func)

        logs = _get_bootspeed_logs(build['artifacts'])
        for num in logs:
            log_num = logs[num]
            # skip incomplete logs
            if logs[num]['data_file'] is None:
                continue
            log_path = "{}artifact/{}".format(
                build['url'],
                logs[num]['data_file']['relativePath'],
            )
            dashboard_data = jenkins_get(log_path)
            # XXX: fix up some malformed log data
            if 'kernelinit' in dashboard_data:
                dashboard_data['kernel_init'] = dashboard_data['kernelinit']

            if self.job_type == 'milestone':
                dashboard_data['image_type'] = Image.TYPE_CHOICES[1][0]
            if 'build_number' in dashboard_data:
                bootspeed_build = dashboard_data['build_number']
                dashboard_data['image_build_number'] = bootspeed_build
            if dashboard_data['image_build_number'] == 'unknown':
                dashboard_data['image_build_number'] = self.image_build_number
            # Allow machine name from job name to override
            # provided machine name for the case where we
            # have spare machines running jobs like backfill.
            dashboard_data['machine_name'] = self.machine
            dashboard_data['image_variant'] = self.variant

            func(build, dashboard_data, log_num)

    def add_result(self, build, dashboard_data, log_num):

        try:
            verify_data(dashboard_data)
        except MissingField as e:
            logging.error(e)
            return
        except BadMD5 as e:
            # We don't need to worry about bad md5s
            return

        image, new_image = Image.objects.get_or_create(
            build_number=self.image_build_number,
            arch=dashboard_data['image_arch'],
            release=dashboard_data['image_release'],
            variant=dashboard_data['image_variant'],
            defaults={
                'md5': dashboard_data['image_md5'],
                'image_type': dashboard_data.get(
                    'image_type',
                    Image.TYPE_CHOICES[0][0],
                ),
            },
        )

        if new_image:
            logging.info(
                "Created image for build_number %s on arch %s",
                self.image_build_number,
                image.arch,
            )

        image_result, new_result = ImageResult.objects.get_or_create(
            image=image,
            machine=self.machine,
            kernel=dashboard_data['kernel'],
            kernel_init=dashboard_data['kernel_init'],
            plumbing=dashboard_data['plumbing'],
            xorg=dashboard_data['xorg'],
            desktop=dashboard_data['desktop'],
            boot=dashboard_data['boot'],
            ran_at=dashboard_data['ran_at'],
        )

        if new_result:
            logging.info(
                "Created new result for image %s %s on machine %s",
                self.image_build_number,
                image.arch,
                self.machine,
            )

        for log in log_num['logs']:
            path = log['relativePath']
            log, new_log = image_result.imageresultlog_set.get_or_create(
                path=path,
                remote_url="{}artifact/{}".format(build['url'], path),
                display_name=path,
            )

            if new_log:
                logging.debug("Added new result log: %s", log)

    def remove_result(self, build, dashboard_data):
        """
        Remove builds marked 'dash_ignore'.
        Remove a bootspeed result as well as an image or machine if needed.
        """
        logging.info(
            "Skipping build %s for job %s",
            self.image_build_number,
            self.jenkins_job.name,
        )

        jenkins_builds = self.jenkins_job.jenkinsbuild_set.filter(
            build_number=build['number'],
        )

        for jenkins_build in jenkins_builds:
            jenkins_build.bugs.all().delete()
            jenkins_build.delete()

        JobImage.objects.filter(
            arch=self.arch,
            build_number=self.image_build_number,
            release=self.release,
            variant=self.variant,
        ).delete()

        verify_data(dashboard_data)

        image = None
        machine = None

        images = Image.objects.filter(
            build_number=self.image_build_number,
            arch=dashboard_data['image_arch'],
            release=dashboard_data['image_release'],
            variant=dashboard_data['image_variant'],
            md5=dashboard_data['image_md5'],
            image_type=dashboard_data.get(
                'image_type',
                Image.TYPE_CHOICES[0][0],
            ),
        )

        if images.count() > 0:
            image = images[0]  # there should only be one

        machines = Machine.objects.filter(name=dashboard_data['machine_name'])

        if machines.count() > 0:
            machine = machines[0]  # there should only be one

        ImageResult.objects.filter(
            image=image,
            machine=machine,
            kernel=dashboard_data['kernel'],
            kernel_init=dashboard_data['kernel_init'],
            plumbing=dashboard_data['plumbing'],
            xorg=dashboard_data['xorg'],
            desktop=dashboard_data['desktop'],
            boot=dashboard_data['boot'],
            ran_at=dashboard_data['ran_at'],
        ).delete()

        if machine is not None and machine.imageresult_set.count() == 0:
            machine.delete()

        if image is not None and image.imageresult_set.count() == 0:
            image.delete()
