# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import re

from django.db import models

from common.models import (
    DashboardBaseModel,
    JenkinsBuild,
)


# Bootspeed Models
class Image(DashboardBaseModel):
    TYPE_CHOICES = (
        (u'daily', u'Daily Image'),
        (u'release', u'Release Image'),
    )

    class Meta:
        db_table = 'images'
        unique_together = ('release', 'variant', 'arch', 'build_number')

    release = models.CharField(max_length=200)
    variant = models.CharField(max_length=200)
    arch = models.CharField(max_length=200)
    md5 = models.CharField(max_length=200, unique=True)
    build_number = models.CharField(max_length=50)
    image_type = models.CharField(
        max_length=10,
        choices=TYPE_CHOICES,
        default=TYPE_CHOICES[0][0]
    )

    def __unicode__(self):
        return "{}-{}-{}-{}".format(
            self.release,
            self.variant,
            self.arch,
            self.build_number,
        )
    # Used by MySerializer to add property fields to the JSON representation
    # of an instance of this class.
    extra_fields = [
        'min_boot',
        'max_boot',
        'avg_boot',
        'result_count',
    ]

    def get_chart_data(self, start=None, end=None):
        x_data = []
        avg_data = []
        max_data = []
        min_data = []
        low_error = []
        hi_error = []
        end_date = end if end is not None else datetime.datetime.now()
        if start is not None:
            return start
        else:
            start = end_date - datetime.timedelta(days=60)
            return start
        start_date = start
        results = self.imageresult_set.filter(
            ran_at__gte=start_date,
            ran_at__lte=end_date
        ).values('ran_at').annotate(
            avg_boot=models.Avg('boot'),
            max_boot=models.Max('boot'),
            min_boot=models.Min('boot')
        ).order_by('ran_at')
        for result in results:
            x_data.append(result['ran_at'])
            avg_data.append(result['avg_boot'])
            max_data.append(result['max_boot'])
            min_data.append(result['min_boot'])
            low_error.append(result['avg_boot'] - result['min_boot'])
            hi_error.append(result['max_boot'] - result['avg_boot'])

        return x_data, avg_data, max_data, min_data, low_error, hi_error

    def result_count(self, machine_id=None):
        results = self.imageresult_set

        if machine_id is not None:
            results = results.filter(machine_id=machine_id)

        return results.count()

    @property
    def avg_boot(self, machine=None):
        results = self.imageresult_set

        if machine is not None:
            results = results.filter(machine=machine)

        return results.aggregate(models.Avg('boot'))['boot__avg']

    @property
    def min_boot(self, machine=None):
        results = self.imageresult_set

        if machine is not None:
            results = results.filter(machine=machine)
        return results.aggregate(models.Min('boot'))['boot__min']

    @property
    def max_boot(self, machine=None):
        results = self.imageresult_set
        if machine is not None:
            results = results.filter(machine=machine)
        return results.aggregate(models.Max('boot'))['boot__max']

    def last_run_date(self, machine=None):
        results = self.imageresult_set

        if machine is not None:
            results = results.filter(machine=machine)
        last_run = results.aggregate(models.Max('ran_at'))['ran_at__max']
        if last_run is not None:
            last_run = last_run.strftime("%Y-%m-%d %H:%M")

        return last_run


class Machine(DashboardBaseModel):
    class Meta:
        db_table = 'machines'
    name = models.CharField(max_length=200, unique=True)
    mac_address = models.CharField(max_length=200)

    # Used by MySerializer to add property fields to the JSON representation
    # of an instance of this class.
    extra_fields = [
        'result_count',
    ]

    def __unicode__(self):
        return self.name

    @property
    def result_count(self):
        return ImageResult.objects.filter(machine=self).count()


class ImageResult(DashboardBaseModel):
    class Meta:
        db_table = 'images_results'

    image = models.ForeignKey(Image)
    machine = models.ForeignKey(Machine)
    ran_at = models.DateTimeField('date run')

    kernel_init = models.FloatField()
    kernel = models.FloatField()
    plumbing = models.FloatField()
    xorg = models.FloatField()
    desktop = models.FloatField()
    boot = models.FloatField()

    @property
    def log_links(self):
        """ Get links to logs for this result. """

        links = ""
        for log in self.imageresultlog_set.all():
            regex = re.compile(ur'^\d*\/(.*)')
            m = regex.match(log.display_name)
            name = m.group(1) if m else log.display_name

            links = " ".join([
                links,
                '<a href="{}">{}</a>'.format(
                    log.remote_url,
                    name,
                )
            ])

        return links

    def __unicode__(self):
        return "{} - {} - {} Result".format(
            self.image,
            self.machine,
            self.ran_at,
        )

    def save(self, *args, **kwargs):
        """
        Overridden to update aggregation information.
        """

        super(ImageResult, self).save(*args, **kwargs)

        results = ImageResult.objects.filter(
            machine=self.machine,
            image=self.image
        )

        result_nums = []

        for result in results:
            result_data = [
                result.kernel,
                result.kernel_init,
                result.plumbing,
                result.xorg,
                result.desktop,
                result.xorg + result.plumbing,
                result.boot,
            ]
            result_nums.append(result_data)

        from numpy import array
        result_num_array = array(result_nums)

        std_ar = result_num_array.std(axis=0)
        avg_ar = result_num_array.mean(axis=0)
        min_ar = result_num_array.min(axis=0)
        max_ar = result_num_array.max(axis=0)

        kernel_std = std_ar[0]
        kernel_init_std = std_ar[1]
        plumbing_std = std_ar[2]
        xorg_std = std_ar[3]
        desktop_std = std_ar[4]
        xorg_plumbing_std = std_ar[5]
        boot_std = std_ar[6]

        kernel_avg = avg_ar[0]
        kernel_init_avg = avg_ar[1]
        plumbing_avg = avg_ar[2]
        xorg_avg = avg_ar[3]
        desktop_avg = avg_ar[4]
        xorg_plumbing_avg = avg_ar[5]
        boot_avg = avg_ar[6]

        kernel_min = min_ar[0]
        kernel_init_min = min_ar[1]
        plumbing_min = min_ar[2]
        xorg_min = min_ar[3]
        desktop_min = min_ar[4]
        xorg_plumbing_min = min_ar[5]
        boot_min = min_ar[6]

        kernel_max = max_ar[0]
        kernel_init_max = max_ar[1]
        plumbing_max = max_ar[2]
        xorg_max = max_ar[3]
        desktop_max = max_ar[4]
        xorg_plumbing_max = max_ar[5]
        boot_max = max_ar[6]

        agg_data, created = ImageMachineAgg.objects.get_or_create(
            image=self.image,
            machine=self.machine,
        )

        agg_data.update(
            count=results.count(),

            kernel_std=kernel_std,
            kernel_avg=kernel_avg,
            kernel_max=kernel_max,
            kernel_min=kernel_min,

            kernel_init_std=kernel_init_std,
            kernel_init_avg=kernel_init_avg,
            kernel_init_max=kernel_init_max,
            kernel_init_min=kernel_init_min,

            plumbing_std=plumbing_std,
            plumbing_avg=plumbing_avg,
            plumbing_max=plumbing_max,
            plumbing_min=plumbing_min,

            xorg_std=xorg_std,
            xorg_avg=xorg_avg,
            xorg_max=xorg_max,
            xorg_min=xorg_min,

            desktop_std=desktop_std,
            desktop_avg=desktop_avg,
            desktop_max=desktop_max,
            desktop_min=desktop_min,

            xorg_plumbing_std=xorg_plumbing_std,
            xorg_plumbing_avg=xorg_plumbing_avg,
            xorg_plumbing_max=xorg_plumbing_max,
            xorg_plumbing_min=xorg_plumbing_min,

            boot_std=boot_std,
            boot_avg=boot_avg,
            boot_max=boot_max,
            boot_min=boot_min,

        )
        agg_data.save()


class ImageMachineAgg(DashboardBaseModel):
    """
    Aggregation table for Image, Machine combos.
    """
    class Meta:
        db_table = 'image_machine_agg'
        verbose_name = 'Image Machine Aggregation'

    image = models.ForeignKey(Image)
    machine = models.ForeignKey(Machine)

    kernel_min = models.FloatField(default=0)
    kernel_max = models.FloatField(default=0)
    kernel_avg = models.FloatField(default=0)
    kernel_std = models.FloatField(default=0)

    kernel_init_min = models.FloatField(default=0)
    kernel_init_max = models.FloatField(default=0)
    kernel_init_avg = models.FloatField(default=0)
    kernel_init_std = models.FloatField(default=0)

    plumbing_min = models.FloatField(default=0)
    plumbing_max = models.FloatField(default=0)
    plumbing_avg = models.FloatField(default=0)
    plumbing_std = models.FloatField(default=0)

    xorg_min = models.FloatField(default=0)
    xorg_max = models.FloatField(default=0)
    xorg_avg = models.FloatField(default=0)
    xorg_std = models.FloatField(default=0)

    desktop_min = models.FloatField(default=0)
    desktop_max = models.FloatField(default=0)
    desktop_avg = models.FloatField(default=0)
    desktop_std = models.FloatField(default=0)

    xorg_plumbing_min = models.FloatField(default=0)
    xorg_plumbing_max = models.FloatField(default=0)
    xorg_plumbing_avg = models.FloatField(default=0)
    xorg_plumbing_std = models.FloatField(default=0)

    boot_min = models.FloatField(default=0)
    boot_max = models.FloatField(default=0)
    boot_avg = models.FloatField(default=0)
    boot_std = models.FloatField(default=0)

    def __unicode__(self):
        return "{} - {} Agg".format(
            self.image,
            self.machine,
        )

    def update(
        self,
        count,
        kernel_avg,
        kernel_min,
        kernel_max,
        kernel_std,
        kernel_init_avg,
        kernel_init_min,
        kernel_init_max,
        kernel_init_std,
        plumbing_avg,
        plumbing_min,
        plumbing_max,
        plumbing_std,
        xorg_avg,
        xorg_min,
        xorg_max,
        xorg_std,
        desktop_avg,
        desktop_min,
        desktop_max,
        desktop_std,
        xorg_plumbing_avg,
        xorg_plumbing_min,
        xorg_plumbing_max,
        xorg_plumbing_std,
        boot_avg,
        boot_min,
        boot_max,
        boot_std,
    ):
        """ Update the entry with the provided values. """

        self.count = count
        self.kernel_avg = kernel_avg
        self.kernel_min = kernel_min
        self.kernel_max = kernel_max
        self.kernel_std = kernel_std
        self.kernel_init_avg = kernel_init_avg
        self.kernel_init_min = kernel_init_min
        self.kernel_init_max = kernel_init_max
        self.kernel_init_std = kernel_init_std
        self.plumbing_avg = plumbing_avg
        self.plumbing_min = plumbing_min
        self.plumbing_max = plumbing_max
        self.plumbing_std = plumbing_std
        self.xorg_avg = xorg_avg
        self.xorg_min = xorg_min
        self.xorg_max = xorg_max
        self.xorg_std = xorg_std
        self.desktop_avg = desktop_avg
        self.desktop_min = desktop_min
        self.desktop_max = desktop_max
        self.desktop_std = desktop_std
        self.xorg_plumbing_avg = xorg_plumbing_avg
        self.xorg_plumbing_min = xorg_plumbing_min
        self.xorg_plumbing_max = xorg_plumbing_max
        self.xorg_plumbing_std = xorg_plumbing_std
        self.boot_avg = boot_avg
        self.boot_min = boot_min
        self.boot_max = boot_max
        self.boot_std = boot_std


class ImageResultLog(DashboardBaseModel):
    """ Image result logs. """

    class Meta:
        db_table = "image_result_logs"

    display_name = models.CharField(max_length=255, null=True)
    path = models.TextField()
    image_result = models.ForeignKey(ImageResult)
    remote_url = models.URLField()

    def __unicode__(self):
        return self.path


class JobImage(DashboardBaseModel):
    """ Jenkins Job image. """

    class Meta:
        db_table = "job_image"

    build_number = models.CharField(max_length=25)
    arch = models.CharField(max_length=25)
    release = models.CharField(max_length=100)
    variant = models.CharField(max_length=100)

    @property
    def job_bug_count(self):
        """ Get number of job bugs for this image. """

        return self.imagebuild_set.annotate(
            build_bug_count=models.Count('build__bugs')
        ).aggregate(bug_count=models.Sum('build_bug_count'))['bug_count']

    @property
    def job_bugs(self):
        """ Get a list of job bugs for this image. """

        # XXX: try to convert this to a pure QuerySet query if possible.
        bugs = []
        for imagebuild in self.imagebuild_set.all():
            bugs += imagebuild.bugs

        return bugs

    @property
    def job_build_count(self):
        """ Get number of builds for this image. """

        return self.imagebuild_set.count()

    @property
    def job_build_success_count(self):
        """ Get number of successful builds for this image. """

        return self.imagebuild_set.filter(build__failed=False).count()

    @property
    def job_last_build_date(self):
        """ Get most recent build date for this image. """

        return self.imagebuild_set.aggregate(
            last_build_date=models.Max('build__ran_at')
        )['last_build_date']

    @property
    def job_builds(self):
        """ Get list of builds for this image. """

        return self.imagebuild_set.all()


class ImageBuild(DashboardBaseModel):
    """ Image build model. """

    class Meta:
        db_table = "image_builds"

    build = models.ForeignKey(JenkinsBuild)
    job_image = models.ForeignKey(JobImage)

    @property
    def bug_count(self):
        return self.build.bugs.count()

    @property
    def bugs(self):
        return self.build.bugs.all()
