# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import datetime
import json

from django.test import TestCase
from django.test.client import Client

from bootspeed.models import (
    Image,
    Machine,
    ImageResult,
    ImageMachineAgg,
    JobImage,
    ImageBuild,
)
from common.models import (
    JenkinsJob,
    JenkinsBuild,
)

from bootspeed.management.commands.jenkins_pull_bootspeed import (
    _get_bootspeed_logs,
)

from bootspeed.utils import (
    verify_data,
    MissingField,
)

from performance.management import (
    _get_file_contents,
    get_image_build_number,
    get_run_number,
)


class TestBootspeedUtils(TestCase):
    """ Test for bootspeed utilities. """

    def setUp(self):

        self.bootspeed_data = dict(
            kernel=0.1,
            kernel_init=0.2,
            plumbing=0.3,
            xorg=0.4,
            desktop=0.5,
            boot=0.6,
            machine_name="test machine",
            machine_mac="00:00:00:00:01",
            image_build_number='20130201',
            image_arch='amd64',
            image_release='raring',
            image_variant='desktop',
            image_md5='blahblahblah',
            ran_at=datetime.datetime.now(),
        )

    def test_verify_data__valid(self):
        """ Test that verify_data() succeeds on valid data. """

        verify_data(self.bootspeed_data)

    def test_verify_data__invalid(self):
        """ Test that verify_data() fails on invalid data. """

        self.bootspeed_data['image_build_number'] = 'unknown'
        with self.assertRaises(MissingField):
            verify_data(self.bootspeed_data)


class TestBootspeed(TestCase):
    """ Tests for bootspeed models. """

    def setUp(self):
        """ Setup commonly used objects. """

        self.image = Image.objects.create(
            release='raring',
            variant='desktop',
            arch='amd64',
            md5='fakemd5',
            build_number='20130101',
        )

        self.assert_(self.image)

        self.machine = Machine.objects.create(
            name='Test Machine',
            mac_address='00:00:00:00:01',
        )

        self.assert_(self.machine)

        self.image_with_results = Image.objects.create(
            release='raring',
            variant='desktop',
            arch='amd64',
            md5='fakemd5-1',
            build_number='20130102',
        )

        self.assert_(self.image_with_results)

        self.machine_with_results = Machine.objects.create(
            name='Test Machine 2',
            mac_address='00:00:00:00:02',
        )

        self.assert_(self.machine_with_results)

        self.image_result = ImageResult.objects.create(
            image=self.image_with_results,
            machine=self.machine_with_results,
            ran_at=datetime.datetime.now(),
            kernel=1.0,
            kernel_init=2.0,
            plumbing=3.0,
            xorg=4.0,
            desktop=5.0,
            boot=15.0,
        )

        self.assert_(self.image_result)

    def test_image_machine_agg_creation(self):
        """ Test ImageMachineAgg is created for an ImageResult. """

        self.assertEqual(ImageMachineAgg.objects.count(), 1)

    def test_image_default_values(self):
        """ Test default values for Image model. """

        self.assertEqual(self.image.min_boot, None)
        self.assertEqual(self.image.max_boot, None)
        self.assertEqual(self.image.avg_boot, None)
        self.assertEqual(self.image.result_count(), 0)
        self.assertEqual(self.image.last_run_date(), None)

    def test_machine_default_values(self):
        """ Test default values for Machine model. """

        self.assertEqual(self.machine.result_count, 0)

    def test_image_with_results(self):
        """ Test default values for Image model. """

        self.assertEqual(self.image_with_results.min_boot, 15.0)
        self.assertEqual(self.image_with_results.max_boot, 15.0)
        self.assertEqual(self.image_with_results.avg_boot, 15.0)
        self.assertEqual(self.image_with_results.result_count(), 1)
        self.assertEqual(
            self.image_with_results.last_run_date(),
            self.image_result.ran_at.strftime("%Y-%m-%d %H:%M"),
        )

    def test_machine_with_results(self):
        """ Test default values for Machine model. """

        self.assertEqual(self.machine_with_results.result_count, 1)

    def test_image_count_release(self):
        """ Test that releases aren't shown in graph api calls. """

        # Convert the test image to a release image
        self.image_with_results.image_type = 'release'
        self.image_with_results.save()

        c = Client()
        url = '/api/bootspeed/arch/{}/'.format(
            self.image_with_results.arch
        )
        response = c.get(url)
        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        # There shouldn't be any images listed.
        self.assertEqual(len(data), 0)


class TestBootspeedPrivateMethods(TestCase):
    """ Test bootspeed private methods. """

    def test__get_file_contents(self):
        path = "http://jenkins.qa.ubuntu.com/"
        contents = _get_file_contents(path)

        self.assertNotEqual(contents, '')

    def test__get_image_build_number_unknown(self):
        """ Returns 'UNKNOWN' for no build_number. """

        build = {
            'description': 'blah',
            'url': 'http://jenkins.qa.ubuntu.com/',
        }
        response = get_image_build_number(build)

        self.assertEqual(response, "UNKNOWN")

    def test__get_image_build_number_from_desc(self):
        """ Returns build_number from description. """

        build = {
            'description': '20130101',
            'url': 'http://jenkins.qa.ubuntu.com/',
        }
        response = get_image_build_number(build)

        self.assertEqual(response, build['description'])

    def test_get_run_number_success(self):
        """ Test get_run_number() works on a valid string. """

        number = '1'
        path = "{}/blah".format(number)

        run_number = get_run_number(path)

        self.assertEqual(run_number, number)

    def test_get_run_number_invalid(self):
        """ Test get_run_number() works on an invalid string. """

        path = "a/blah"

        run_number = get_run_number(path)

        self.assertEqual(run_number, None)

    def test_get_bootspeed_logs(self):
        """ Test _get_bootspeed_logs() returns a valid. """

        number = '1'
        data_file = '{}/boot.json'.format(number)
        log_file = '{}/bootchart.png'.format(number)

        artifacts = [
            {
                'relativePath': '{}'.format(data_file)
            },
            {
                'relativePath': '{}'.format(log_file)
            },
        ]

        logs = _get_bootspeed_logs(artifacts)

        self.assertIn(number, logs)
        self.assertEqual(len(logs), 1)
        self.assertEqual(logs[number]['data_file']['relativePath'], data_file)
        self.assertEqual(len(logs[number]['logs']), 1)
        self.assertEqual(logs[number]['logs'][0]['relativePath'], log_file)


class TestJenkins(TestCase):
    """ Test the ImageBuild models. """

    def setUp(self):
        """ Common setup. """

        self.image_release = 'raring',
        self.image_variant = 'desktop',
        self.image_arch = 'amd64',
        self.image_build_number = '20130101',

        self.job_image = JobImage.objects.create(
            build_number=self.image_build_number,
            release=self.image_release,
            variant=self.image_variant,
            arch=self.image_arch,
        )
        self.assert_(self.job_image)

        self.job = JenkinsJob.objects.create(
            name="bootspeed-raring-desktop-amd64-acer-veriton-01",
            url="blah",
        )
        self.assert_(self.job)

        self.last_build_date = datetime.datetime.now()
        self.build = self.job.jenkinsbuild_set.create(
            ran_at=self.last_build_date,
            failed=True,
            build_number=1,
        )
        self.assert_(self.build)

        self.image_build = ImageBuild.objects.create(
            job_image=self.job_image,
            build=self.build,
        )
        self.assert_(self.image_build)

        self.bug = self.image_build.build.bugs.create(
            bug_no='1234',
            status='open',
        )
        self.assert_(self.bug)

    def test_image_build_bug_count(self):
        """ Test that an ImageBuild reports the correct bug count. """

        self.assertEqual(self.image_build.bug_count, 1)

    def test_image_job_bug_count(self):
        """ Test image returns correct bug count. """

        self.assertEqual(self.job_image.job_bug_count, 1)

    def test_image_job_bugs(self):
        """ Test image returns the correct list of bugs. """

        bugs = self.job_image.job_bugs
        self.assertEqual(len(bugs), 1)
        self.assertEqual(bugs[0], self.bug)

    def test_image_job_build_count(self):
        """ Test image returns the correct number of builds. """

        self.assertEqual(self.job_image.job_build_count, 1)

    def test_image_job_build_success_count(self):
        """ Test image returns the correct number of successful builds. """

        self.assertEqual(self.job_image.job_build_success_count, 0)

        self.build.failed = False
        self.build.save()

        self.assertEqual(self.job_image.job_build_success_count, 1)

    def test_image_job_builds(self):
        """ Test image returns correct list of builds. """

        builds = self.job_image.job_builds

        self.assertEqual(len(builds), 1)
        self.assertEqual(builds[0], self.image_build)

    def test_image_job_last_build_date(self):
        """ Test image returns correct most recent run date. """
        self.assertEqual(
            self.job_image.job_last_build_date,
            self.last_build_date,
        )

        new_last_build_date = datetime.datetime.now()
        new_build = JenkinsBuild.objects.create(
            job=self.job,
            ran_at=new_last_build_date,
            failed=False,
            build_number=2,
        )
        self.assert_(new_build)

        new_imagebuild = self.job_image.imagebuild_set.create(
            build=new_build,
        )
        self.assert_(new_imagebuild)

        self.assertEqual(
            self.job_image.job_last_build_date,
            new_last_build_date,
        )
