# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.db import models
from django.views.decorators.http import require_GET
from django.template import RequestContext

from django.http import HttpResponse
from django.shortcuts import (
    render_to_response,
    get_object_or_404,
    redirect,
)

from django_tables2 import RequestConfig

from common.models import Bug

import matplotlib
matplotlib.use('Agg')
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.dates import date2num
from pylab import figure, gca

from bootspeed.models import (
    Image,
    Machine,
    ImageResult,
    ImageBuild,
    JobImage,
)

from bootspeed.tables import MachineTable

CHART_WIDTH_DAYS = 60


def _get_bootspeed_arches():
    arches = []
    for image in Image.objects.values(
        'arch'
    ).distinct().exclude(publish=False).order_by('-arch'):
        arches.append(image['arch'])

    return arches


def _get_bootspeed_releases():
    releases = []
    for image in Image.objects.values('release').distinct().exclude(
        publish=False
    ).order_by('-release'):
        releases.append(image['release'])

    return releases


@require_GET
def overview(request, release=None):
    images = Image.objects.all()
    releases = _get_bootspeed_releases()

    if release is not None:
        images = images.filter(release=release)

    data = {
        'images': images,
        'releases': releases,
    }

    return render_to_response(
        'bootspeed/overview.html',
        data,
        RequestContext(request)
    )


@require_GET
def image_overview_old(request, image_id):
    image = get_object_or_404(Image, id=image_id)

    data = {
        'image': image,
        'results': image.imageresult_set.values('ran_at').annotate(
            kernel_init_avg=models.Avg('kernel_init'),
            kernel_init_min=models.Min('kernel_init'),
            kernel_init_max=models.Max('kernel_init'),

            kernel_avg=models.Avg('kernel'),
            kernel_min=models.Min('kernel'),
            kernel_max=models.Max('kernel'),

            plumbing_avg=models.Avg('plumbing'),
            plumbing_min=models.Min('plumbing'),
            plumbing_max=models.Max('plumbing'),

            xorg_avg=models.Avg('xorg'),
            xorg_min=models.Min('xorg'),
            xorg_max=models.Max('xorg'),

            desktop_avg=models.Avg('desktop'),
            desktop_min=models.Min('desktop'),
            desktop_max=models.Max('desktop'),

            boot_avg=models.Avg('boot'),
            boot_min=models.Min('boot'),
            boot_max=models.Max('boot'),
        ),
    }

    return render_to_response(
        'bootspeed/image_overview.html',
        data,
        RequestContext(request)
    )


@require_GET
def build_boot_chart(request, image_id):
    image = Image.objects.get(id=image_id)

    end = datetime.datetime.now()
    start = end - datetime.timedelta(days=CHART_WIDTH_DAYS)
    image = image.get_chart_data
    x_data, avg_data, max_data, min_data, low_error, hi_error = image(
        start=start,
        end=end,
    )

    f = figure(figsize=(12, 9))  # inches
    f.patch.set_alpha(0.0)  # transparent background
    #plot_date(date2num(x_data), avg_data, 'b-')
    #plot_date(date2num(x_data), max_data, 'g-')
    #plot_date(date2num(x_data), min_data, 'r-')
    ax = gca()
    ax.bar(
        x_data,
        avg_data,
        color='#dd4814',
        yerr=[low_error, hi_error],
        ecolor='b'
    )
    ax.set_title('Bootspeed (sec.)')
    labels = ax.get_xticklabels()
    for label in labels:
        label.set_rotation(90)
        label.set_size('small')
    ax.set_xlim(date2num(start), date2num(end))
    ax.xaxis.set_major_locator(
        matplotlib.dates.DayLocator()
    )
    ax.xaxis.set_major_formatter(
        matplotlib.dates.DateFormatter('%Y%m%d')
    )
    response = HttpResponse(content_type='image/png')
    canvas = FigureCanvasAgg(f)
    canvas.print_png(response, transparent=True)
    matplotlib.pyplot.close(f)

    return response


@require_GET
def image_overview(request, image_id):
    image = get_object_or_404(Image, id=image_id)
    api_url = 'image'
    data = {
        'image': image,
        'api_url': api_url,
        'image_id': image.id
    }
    return render_to_response(
        'bootspeed/image_chart.html',
        data,
        RequestContext(request)
    )


@require_GET
def arch_overview(request, arch=None):
    if arch is None:
        return redirect('arch/amd64')

    result = ImageResult.objects.filter(image__arch=arch).distinct(
        'machine__name'
    )
    arches = _get_bootspeed_arches()
    data = {
        'result': result,
        'arch': arch,
        'arches': arches,
        'active': arch,
    }
    return render_to_response(
        'bootspeed/arch_overview.html',
        data,
        RequestContext(request)
    )


@require_GET
def arch_image_list(request, arch):
    images = Image.objects.filter(arch=arch)
    arches = _get_bootspeed_arches()
    data = {
        'images': images,
        'arches': arches,
    }

    return render_to_response(
        'bootspeed/image_list.html',
        data,
        RequestContext(request)
    )


@require_GET
def machine_data(request, machine_id, arch):
    machine = get_object_or_404(Machine, id=machine_id)

    data = {
        'name': machine.name,
        'id': machine.id,
        'arch': arch,
    }
    return render_to_response(
        "bootspeed/machine_data.html",
        data,
        RequestContext(request)
    )


@require_GET
def machine_raw_data(request, machine_id, arch):
    machine = get_object_or_404(Machine, id=machine_id)

    qs = ImageResult.objects.filter(machine__id=machine_id, image__arch=arch)

    data = []
    for q in qs:
        data.append(dict(
            machine=q.machine,
            image=q.image,
            image__build_number=q.image.build_number,
            image__arch=q.image.arch,
            image__release=q.image.release,
            image__variant=q.image.variant,
            kernel=q.kernel,
            kernel_init=q.kernel_init,
            plumbing=q.plumbing,
            xorg=q.xorg,
            desktop=q.desktop,
            boot=q.boot,
            files=q.log_links,
        ))

    table = MachineTable(data)
    RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'name': machine.name,
        'id': machine.id,
        'arch': arch,
        'table': table,
    }
    return render_to_response("bootspeed/machine_raw_data.html", data,
                              RequestContext(request))


@require_GET
def image_bugs_overview(request):
    images = JobImage.objects.annotate(
        imagebuild_count=models.Count(
            'imagebuild'
        )
    ).filter(imagebuild_count__gt=0).order_by(
        '-build_number',
        'arch'
    ).exclude(build_number="UNKNOWN")

    unknown_images = JobImage.objects.annotate(
        imagebuild_count=models.Count(
            'imagebuild')
    ).filter(imagebuild_count__gt=0).order_by(
        '-build_number',
        'arch'
    ).filter(build_number="UNKNOWN")

    data = {
        'images': images,
        'unknown_images': unknown_images,
    }

    return render_to_response(
        "bootspeed/image_bugs_overview.html",
        data,
        RequestContext(request)
    )


@require_GET
def image_bugs_detail(request, image_id):
    image_builds = ImageBuild.objects.filter(job_image=image_id)
    image = get_object_or_404(JobImage, id=image_id)

    data = {
        'image_builds': image_builds,
        'image': image,
    }

    return render_to_response(
        "bootspeed/image_bugs_detail.html",
        data,
        RequestContext(request)
    )


@require_GET
def bug_overview(request):
    bugs = Bug.objects.annotate(
        build_count=models.Count('build_bugs')
    ).filter(build_count__gt=0).order_by('-build_count')

    bug_numbers = []
    for bug in bugs:
        if bug.bug_no not in bug_numbers:
            bug_numbers.append(bug.bug_no)

    data = {
        'bugs': bugs,
    }

    return render_to_response(
        "bootspeed/bug_overview.html",
        data,
        RequestContext(request)
    )


@require_GET
def bug_detail(request, bug_id):
    bug = get_object_or_404(Bug, id=bug_id)

    data = {
        'builds': bug.build_bugs.all(),
        'bug': bug,
    }

    return render_to_response(
        "bootspeed/bug_detail.html",
        data,
        RequestContext(request)
    )
