# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import json
import logging
import os
import re
import shutil
import urllib
import urllib2
import yaml
import zipfile

from django.core.management.base import BaseCommand
from optparse import make_option

BUILD_COUNT_MAX = 20  # XXX: DEBUG
JOB_COUNT_MAX = 20  # XXX: DEBUG
DEBUG_LIMITS = False

from common.utils import regexes
from common.models import JenkinsJob, Bug

from eventstat.models import eventstat_models


def jenkins_get(loc, as_json=True):
    req = urllib2.Request(loc)
    response = ''
    try_count = 0

    while try_count < 5:
        try:
            f = urllib2.urlopen(req)
            response = f.read()
            f.close()
            break
        except urllib2.HTTPError:
            try_count += 1
        except urllib2.URLError:
            try_count += 1

    if response != '' and as_json:
        try:
            return json.loads(response)
        except ValueError:
            logging.warn("jenkins_get: response: {}".format(response))
            # handle this exception
            raise
    else:
        return response


def process_class_list(args):
    """ Handle app names and class names.

    :returns: A list of model names.

    """

    klass_list = []
    for arg in args:
        if arg == 'bootspeed':
            klass_list += [
                'Image',
                'Machine',
                'ImageResult',
                'ImageMachineAgg',
                'JobImage',
                'ImageBuild',
            ]
        elif arg == 'smoke':
            klass_list += [
                'Build',
                'Run',
                'Result',
                'ResultLog',
            ]
        elif arg == 'sru':
            klass_list += [
                'Kernel',
                'SRUResult',
                'SRUResultLog',
            ]
        elif arg == 'common':
            klass_list += [
                'Bug',
                'JenkinsJob',
                'JenkinsBuild',
            ]
        elif arg == 'power':
            klass_list += [
                'PowerImage',
                'PowerLog',
                'PowerMachine',
                'PowerMetric',
            ]
        elif arg == 'memory':
            klass_list += [
                'MemoryBuild',
                'MemoryDetail',
                'MemoryImage',
                'MemoryLog',
                'MemoryMachine',
                'MemoryMetric',
                'MemoryProcess',
                'MemoryResult',
                'MemoryUpgrade',
            ]
        elif arg == 'eventstat':
            klass_list += [x.__name__ for x in eventstat_models]
        else:
            klass_list.append(arg)

    return klass_list


def skip_build(build):
    """ Whether or not to skip the provided build. """

    build_desc = build.get('description')
    if build_desc is None:
        return False

    if regexes['skip_build'].search(build_desc):
        return True

    return False


def parse_build_description(build_desc):
    """ Get data from build description. """

    if build_desc is None:
        build_desc = ''

    tokens = regexes['build_desc'].findall(build_desc)

    build_number = None
    bugs = []
    job_bugs = []
    other_tokens = []

    for token in tokens:
        word = token.strip()

        # the first build_number found wins.
        if regexes['buildid'].match(word) and build_number is None:
            build_number = word
        elif regexes['lpbug'].match(word):
            bug = word[4:]
            if bug not in bugs:
                bugs.append(bug)
        elif regexes['lpbug-job'].match(word):
            job_bug = word[8:]
            if job_bug not in job_bugs:
                job_bugs.append(job_bug)
        else:
            other_tokens.append(word)

    return dict(
        build_number=build_number,
        bugs=bugs,
        job_bugs=job_bugs,
        other_tokens=other_tokens,
    )


def profile(fn):
    """ Decorator for time profiling function calls.

    usage:
        @profile
        def some_method():
            ...

    """

    def wrapped(*args, **kwargs):
        import datetime
        start = datetime.datetime.now()
        result = fn(*args, **kwargs)
        end = datetime.datetime.now()
        logging.info("PROFILE: %s %s", fn.__name__, end - start)

        return result

    return wrapped


class JenkinsBaseCommand(BaseCommand):
    help = "Gather bootspeed results from jenkins.qa.ubuntu.com"
    args = "[job name] [job name]..."

    option_list = BaseCommand.option_list + (
        make_option(
            '-I', '--initial-run',
            action='store_true',
            dest='initial_run',
            default=False,
            help='Initial run, therefore process all results',
        ),
        make_option(
            '-d', '--days',
            dest='history_days',
            help='The number of days to pull data',
            default=7,
            type=int,
        ),
        make_option(
            '-w', '--wait-for',
            dest='waitfor',
            help='Skip all jobs until the named job is found',
        ),
    )

    # To be overridden by subclasses
    job_regex = re.compile(ur'^$')

    create_jenkins_job = False

    builds = []
    skip_builds = []

    JOB_BUILD_SEP = ':'

    def process_job(self, build):
        pass

    def create_jenkins_job(self, job):
        jenkins_job, new_jenkins_job = JenkinsJob.objects.get_or_create(
            name=job['name'],
            url=job['url'],
        )

        self.jenkins_job = jenkins_job
        self.new_jenkins_job = new_jenkins_job

    def preprocess_builds(self, builds):
        """ Add the build to the either the list of builds to add or
        the list of builds to skip.

        """

        self.skip_builds = []
        self.builds = []

        for build in builds:
            if skip_build(build):
                self.skip_builds.append(build)
            else:
                self.builds.append(build)

    def create_jenkins_build(self, build):
        """ Add build and parse results. """

        failed = build['result'] == 'FAILURE'
        logging.info(
            "Adding build %s for job %s",
            build['number'],
            self.jenkins_job.name,
        )

        build_set = self.jenkins_job.jenkinsbuild_set
        jenkins_build, new_jenkins_build = build_set.get_or_create(
            build_number=build['number'],
            defaults={
                'failed': failed,
                'ran_at': self.build_date,
            },
        )

        self.install_data['jenkins_build'] = jenkins_build

    def process_build(self, build):
        pass

    def post_process_build(self, build):
        pass

    def process_add_result(self, build):
        self._process_logs(
            build=build,
            build_date=self.build_date,
            install_data=self.install_data,
            func=self.add_result,
        )

    def remove_skipped_build(self, build):
        """ Remove builds marked 'dash_ignore'. """

        logging.info(
            "Skipping build %s for job %s",
            build['number'],
            self.jenkins_job.name,
        )

        jenkins_builds = self.jenkins_job.jenkinsbuild_set.filter(
            build_number=build['number'],
        )

        for jenkins_build in jenkins_builds.iterator():
            jenkins_build.bugs.all().delete()
            jenkins_build.delete()

    def process_remove_result(self, build):
        self._process_logs(
            build=build,
            build_date=self.build_date,
            install_data=self.install_data,
            func=self.remove_result,
        )

    def add_job_bugs(self, build):
        build_desc = build.get('description', '')
        build_desc_data = parse_build_description(build_desc)
        for job_bug in build_desc_data['job_bugs']:
            job_bug, new_job_bug = Bug.objects.get_or_create(
                bug_no=job_bug,
                defaults={
                    'status': 'open',
                },
            )
            self.install_data['jenkins_build'].bugs.add(job_bug)

    def process_skipped_build(self, build):
        pass

    def extract_data(self, name):
        pass

    def post_jobs(self):
        pass

    def get_build_date(self, build):
        return datetime.datetime.fromtimestamp(build['timestamp']/1000)

    def get_artifacts(self, build):
        """ Pull the artifacts from jenkins locally and unzip them. """
        artifact_zip_path = "{}artifact/*zip*/artifact.zip".format(
            build['url'],
        )

        home = os.path.expanduser("~")
        unzip_path = os.path.join(home, 'tmp', 'dashboard_artifacts')

        if os.path.exists(unzip_path):
            shutil.rmtree(unzip_path)

        artifact_zip_path_local = os.path.join(
            home,
            "dashboard_artifact.zip",
        )
        artifact_zip_path_local = os.path.expanduser(artifact_zip_path_local)

        urllib.urlretrieve(
            artifact_zip_path,
            filename=artifact_zip_path_local,
        )
        art_zip = zipfile.ZipFile(artifact_zip_path_local, 'r')
        art_zip.extractall(unzip_path)

        logging.info("artifact_zip_path : %s", artifact_zip_path)
        logging.info("artifact_zip_path_local : %s", artifact_zip_path_local)

        return unzip_path

    def get_kernel(self, build_url, artifact):
        """ Try to find a kernel in a utah log artifact. """
        retval = "unknown"

        if artifact is None:
            return retval

        log_path = "{}artifact/{}".format(
            build_url,
            artifact['relativePath'],
        )

        data = jenkins_get(log_path, as_json=False)
        try:
            data = yaml.load(data)
        except yaml.YAMLError:
            logging.warn("failed to parse data: %s", data)
            raise

        kernel = data['uname'][2]

        return kernel

    def get_local_file(self, path):
        """ Get a local files contents. """

        output = ""

        with open(path, 'r') as f:
            output = f.read()

        return output

    def handle(self, *args, **options):
        start_time = datetime.datetime.now()
        initial_run = options.get('initial_run')
        verbosity = int(options.get('verbosity'))
        waitfor = options.get('waitfor')
        history_days = options.get('history_days')

        if verbosity == 2:
            logging.basicConfig(level=logging.INFO)
        elif verbosity > 2:
            logging.basicConfig(level=logging.DEBUG)

        job_list = list(args)
        logging.info("job_list: {}".format(job_list))

        only_jobs = []
        for job in job_list:
            job_data = job.split(self.JOB_BUILD_SEP, 2)
            job_len = len(job_data)

            if job_len == 0:
                logging.warn("skipping invalid job: %s", job)
                continue

            job_name = job_data[0]
            job_build = None

            if job_len == 2:
                job_build = job_data[1]

            only_jobs.append(
                dict(
                    name=job_name,
                    build=job_build,
                ),
            )

        url = "https://jenkins.qa.ubuntu.com/api/json"
        jobs_top = jenkins_get(url)

        jobs = jobs_top['jobs']

        waiting = waitfor is not None
        waitfor_data = []

        waitfor_build = None
        waitfor_job = ""

        if waiting:
            waitfor_data = waitfor.split(self.JOB_BUILD_SEP, 2)

            if len(waitfor_data) == 0:
                logging.error("invalid job to wait for: %s", waitfor)
                return

            waitfor_job = waitfor_data[0]

        if len(waitfor_data) >= 2:
            waitfor_build = waitfor_data[1]

        for job in jobs:

            name = job['name']
            # skip jobs until the one we're waiting for
            if waiting:
                if name == waitfor_job:
                    logging.info("found job %s", waitfor_job)
                    waiting = False
                else:
                    continue

            # if passed a list of jobs skip those that don't match
            if len(only_jobs) > 0 and name not in [
                x['name'] for x in only_jobs
            ]:
                logging.debug("Skipping job %s", name)
                continue

            # skip jobs that don't match a subclasses job regex.
            m = self.job_regex.match(name)

            if not m:
                continue

            logging.info("Job name: {}".format(name))

            self.extract_data(name)
            self.create_jenkins_job(job)
            self.process_job(job)

            job_info = jenkins_get(
                "{}/api/json/?depth=1".format(
                    job['url']
                )
            )

            # Ignore jobs that haven't been run yet.
            if job_info['lastBuild'] is None:
                continue

            last_build = datetime.datetime.fromtimestamp(
                job_info['lastBuild']['timestamp']/1000
            )

            # for update runs only check jobs that have built in the
            # last week.
            if (
                not initial_run and datetime.datetime.now(
                ) - last_build > datetime.timedelta(days=history_days)
            ):
                continue

            builds = job_info['builds']

            self.preprocess_builds(builds)

            for build in self.skip_builds:
                self.build_date = self.get_build_date(build)
                logging.info("Skipping Job: {}, Build: {}".format(
                    job['name'], build['number']))
                self.remove_skipped_build(build)
                self.process_skipped_build(build)
                self.process_remove_result(build)

            for build in self.builds:
                if waitfor_build and build['number'] != waitfor_build:
                    continue

                # reset so we start processing builds at the specified one
                waitfor_build = None

                only_builds = [
                    int(x['build']) for x in only_jobs
                    if x['name'] == name and x['build'] is not None
                ]

                if len(only_builds) > 0 and build['number'] not in only_builds:
                    logging.info("Skipping build %s", build['number'])
                    continue

                building = build['building']

                # Skip this Jenkins build if it hasn't finished yet
                if building:
                    continue

                self.build_date = self.get_build_date(build)

                if (
                    not initial_run and datetime.datetime.now(
                    ) - self.build_date > datetime.timedelta(
                        days=history_days
                    )
                ):
                    continue
                self.create_jenkins_build(build)
                self.add_job_bugs(build)
                self.process_build(build)
                self.process_add_result(build)
                self.post_process_build(build)

        self.post_jobs()
        logging.debug("time: {}".format(datetime.datetime.now() - start_time))
