# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.db import models


class DashboardBaseModel(models.Model):
    """
    Common base class that provides fields needed on every model.
    """
    class Meta:
        abstract = True

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    internal = models.BooleanField(default=True)
    publish = models.BooleanField(default=True)


class Bug(DashboardBaseModel):
    class Meta:
        db_table = 'bugs'
    bug_no = models.CharField(max_length=255)
    status = models.CharField(max_length=255)

    def __unicode__(self):
        return self.bug_no

    @property
    def debug_str(self):
        return "{}-{}-{}-{}-{}".format(
            self.id,
            self.bug_no,
            self.status,
            self.created_at,
            self.updated_at,
        )

    @property
    def affected_job_count(self):
        """ Get the number of JenkinsJobs affected by this bug. """

        return self.build_bugs.annotate(
            job_count=models.Count('job')
        ).aggregate(models.Sum('job_count')
                    )['job_count__sum'] or 0

    @property
    def affected_build_count(self):
        """ Get the number of JenkinsJobs affected by this bug. """

        return self.build_bugs.count()


class JenkinsJob(DashboardBaseModel):
    """ Jenkins job. """

    class Meta:
        db_table = "jenkins_jobs"

    name = models.CharField(max_length=200, unique=True)
    url = models.URLField()

    def __unicode__(self):
        return self.name

    def bug_count(self):
        a = self.jenkinsbuild_set.annotate(
            bug_count=models.Count('bugs')
        ).aggregate(models.Sum('bug_count')
                    )
        return a['bug_count__sum'] or 0


class JenkinsBuild(DashboardBaseModel):
    """ Jenkins build. """

    class Meta:
        db_table = "jenkins_builds"
        unique_together = ('job', 'build_number')

    job = models.ForeignKey(JenkinsJob)
    failed = models.BooleanField(default=True)
    bugs = models.ManyToManyField(Bug, related_name='build_bugs')
    ran_at = models.DateTimeField('date run')
    build_number = models.CharField(max_length=50)

    def __unicode__(self):
        return "{} - {}".format(
            self.job,
            self.build_number,
        )

    @property
    def url(self):
        return "{}{}".format(
            self.job.url,
            self.build_number,
        )

    def bug_count(self):
        """ Get number of bugs for this build. """

        return self.bugs.count()
