# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.test import TestCase

from django.db import IntegrityError

from common.models import (
    Bug,
    JenkinsJob,
    JenkinsBuild,
)

from common.management import skip_build


class TestManagement(TestCase):
    """ Test management methods. """

    def setUp(self):

        self.build = {
            'description': 'dash_ignore',
        }

    def test_skip_build(self):
        """ Test that skip_build returns correct results. """

        self.assertTrue(skip_build(self.build))

    def test_skip_build_none_description(self):
        """ Test skip_build returns false when description is None. """

        self.build['description'] = None
        self.assertFalse(skip_build(self.build))

    def test_skip_build_missing_description(self):
        """ Test skip_build returns false when description is missing. """

        self.build = {}
        self.assertFalse(skip_build(self.build))


class TestJenkins(TestCase):
    """ Test Job and JobBug models. """

    def setUp(self):
        """ Common setup. """

        self.name = "bootspeed-backfill-desktop-amd64-aver-veriton-01"
        self.job = JenkinsJob.objects.create(
            name=self.name,
            url="http://jenkins.qa.ubuntu.com/job/{}".format(self.name),
        )
        self.assert_(self.job)

        self.build_failed = True
        self.build_number = 1
        self.build = JenkinsBuild.objects.create(
            job=self.job,
            failed=self.build_failed,
            ran_at=datetime.datetime.now(),
            build_number=self.build_number,
        )
        self.assert_(self.build)

        self.extra_name = "bootspeed-backfill-desktop-amd64-aver-veriton-02"
        self.extra_job = JenkinsJob.objects.create(
            name=self.extra_name,
            url="http://jenkins.qa.ubuntu.com/job/{}".format(self.extra_name),
        )
        self.assert_(self.extra_job)

        self.extra_build = self.extra_job.jenkinsbuild_set.create(
            failed=self.build_failed,
            ran_at=datetime.datetime.now(),
            build_number=self.build_number,
        )
        self.assert_(self.extra_build)

        self.bug_no = '12345'
        self.bug_status = 'open'

        self.build.bugs.create(
            bug_no=self.bug_no,
            status=self.bug_status,
        )
        self.assertEqual(self.build.bugs.count(), 1)

        self.build_bug = self.build.bugs.all()[0]

    def test_jenkins_job(self):
        """ Test JenkinsJob model. """

        self.assertEqual(self.name, self.job.name)

    def test_jenkins_build(self):
        """ Test JenkinsBuild model. """

        self.assertEqual(self.build.job, self.job)
        self.assertEqual(self.build.failed, self.build_failed)

    def test_jenkins_build_bug(self):
        """ Test JenkinsBuildBug model. """

        self.assertEqual(self.build_bug.build_bugs.all()[0], self.build)
        self.assertEqual(self.build_bug.bug_no, self.bug_no)
        self.assertEqual(self.build_bug.status, self.bug_status)

    def test_bug_agg(self):
        """ Test that bugs reference their builds. """

        self.assertEqual(self.job.bug_count(), 1)
        self.assertEqual(self.build.bug_count(), 1)

    def test_add_bug_agg(self):
        """ Test that bug count increments for job and build. """

        new_bug = self.build.bugs.create(
            bug_no='2345',
            status='open',
        )

        self.assert_(new_bug)

        self.assertEqual(self.build.bug_count(), 2)
        self.assertEqual(self.job.bug_count(), 2)

    def test_bug_affects_jobs(self):
        """ Test that a bug can tell how many jobs are affected. """

        affected_job_count = self.build_bug.affected_job_count

        self.assertEqual(affected_job_count, 1)

        self.extra_build.bugs.add(self.build_bug)

        self.assertEqual(self.build_bug.affected_job_count, 2)

    def test_bug_afftects_builds(self):
        """ Test taht a bug can tell how many builds are affected. """

        affected_build_count = self.build_bug.affected_build_count

        self.assertEqual(affected_build_count, 1)

        self.extra_build.bugs.add(self.build_bug)

        self.assertEqual(self.build_bug.affected_build_count, 2)

    def test_build_constraints(self):
        """
        Test that two builds can't exist with
        the same job and build number.
        """

        with self.assertRaises(IntegrityError):
            self.job.jenkinsbuild_set.create(
                build_number=self.build_number,
                ran_at=datetime.datetime.now(),
                failed=self.build_failed,
            )

    def test_job_name_contraint(self):
        """ Test two jobs can't have the same name. """

        with self.assertRaises(IntegrityError):
            JenkinsJob.objects.create(
                name=self.name,
                url="doesntmatter",
            )
