# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import json

from django.http import HttpResponse
from django.views.decorators.http import require_GET

from django.core.serializers.json import Serializer

regexes = {
    'bootspeed': re.compile(
        ur"^bootspeed-(milestone|backfill|raring)-(desktop)-([^-]*)-(.*)"
    ),
    'power': re.compile(
        ur"^power-(milestone|backfill|raring)-(desktop)-([^-]*)-(.*)"
    ),
    'upgrade': re.compile(
        ur".*-upgrade-.*"
    ),
    'smoke': re.compile(
        ur"^(precise|quantal)-(desktop|server|alternate)"
    ),
    'utah-smoke': re.compile(
        ur"^(raring)-(desktop|server)-.*-smoke-.*"
    ),
    'utah-smoke-data': re.compile(
        ur"^(raring)-(desktop|server)-([^-]*)-smoke-(.*)"
    ),
    'smoke-ec2': re.compile(
        ur"(.*)-(.*)-ec2.*"
    ),
    'smoke-ec2-daily': re.compile(
        ur".*-daily$"
    ),
    'smoke-ec2-daily-data': re.compile(
        ur"^(.*)-(.*).*"
    ),
    'smoke-data': re.compile(
        ur"^(precise|quantal)-(.*)-([^_]*)_(.*)"
    ),
    'smoke-buildid': re.compile(
        r"^[0-9]{8}(\.[0-9]+)?$"
    ),
    'smoke-lpbug': re.compile(
        ur"LP:#[0-9]+$"
    ),
    'smoke-build_desc': re.compile(
        ur"[^,\s]+"
    ),
    'kernel_sru': re.compile(
        ur"^(sru_kernel)(|_backport)-(lucid|maveric|natty|oneiric|precise|quantal|raring)(|_lts_hwe)-(.*)-(.*)-(.*).*"
    ),
    'kernel_sru-version': re.compile(
        ur"^([^,\s]+)(\s+|,)"
    ),
    'lpbug': re.compile(
        ur"LP:#[0-9]+$"
    ),
    'lpbug-job': re.compile(
        ur"JOB-LP:#[0-9]+$"
    ),
    'buildid': re.compile(
        r"^[0-9]{8}(\.[0-9]+)?$"
    ),
    'build_desc': re.compile(
        ur"[^,\s]+"
    ),
    'build-number': re.compile(
        ur"^.*\((.+)\)$"
    ),
    'console-build-number': re.compile(
        ur'.*Media Info:.*\(([0-9.]+)\)'
    ),
    'artifact-ignore': re.compile(
        ur"^tests\/"
    ),
    'skip_build': re.compile(
        ur"dash_ignore"
    ),
}


def get_sru_data(name):
    """
    Parses a testcase name and returns a dictionary of sru test data
    of None if data is missing.
    """

    m = regexes['kernel_sru'].match(name)
    if not m:
        return None

    pre_job_type = m.group(1)
    post_job_type = m.group(2)
    job_type = "{}{}".format(pre_job_type, post_job_type)

    pre_release = m.group(3)
    post_release = m.group(4)
    release = "{}{}".format(pre_release, post_release)

    variant, arch, gpu = m.group(5, 6, 7)

    return {
        'release': release,
        'variant': variant,
        'arch': arch,
        'gpu': gpu,
    }


def get_smoke_data(name):
    """
    Parses a testcase name and returns a dictionary of smoke test data
    or None if data is missing.
    """

    if not regexes['smoke'].match(name):
        return None

    if regexes['smoke-ec2'].match(name):
        return None

    m = regexes['smoke-data'].match(name)
    release, variant, arch, name = m.group(1, 2, 3, 4)

    return {
        'release': release,
        'variant': variant,
        'arch': arch,
        'name': name,
    }


def require_API_TOKEN(the_func=None):
    """
    Make sure that request.POST has a valid token in it.
    """
    def _dec(view_func):
        def _view(request, *args, **kwargs):
            from qa_dashboard.settings import API_TOKEN

            token = request.POST.get('token', '')

            if API_TOKEN == 'DISABLED' or token != API_TOKEN:
                return JSONResponse(json.dumps([{
                    'result': 'failure',
                    'message': 'Invalid API token',
                }]))
            else:
                return view_func(request, *args, **kwargs)

        _view.__name__ = view_func.__name__
        _view.__dict__ = view_func.__dict__
        _view.__doc__ = view_func.__doc__

        return _view

    if the_func is not None:
        return _dec(the_func)
    else:
        return _dec


def JSONResponse(data):
    """
    Convenience method.
    """
    return HttpResponse(data, content_type='application/json')

# From: http://stackoverflow.com/questions/5453237/override-django-object-serializer-to-get-rid-of-specified-model


class MySerializer(Serializer):
    def __init__(self, use_extra_fields=False):
        super(MySerializer, self).__init__()

        self.use_extra_fields = use_extra_fields

    def end_object(self, obj):
        self._current['id'] = obj._get_pk_val()

        # Add any property fields listed to the object representation
        # XXX: This slows things down considerably, look for a better way.
        if self.use_extra_fields and hasattr(obj, 'extra_fields'):
            attrs = getattr(obj, 'extra_fields')

            for attr in attrs:
                if hasattr(obj, attr):
                    self._current[attr] = getattr(obj, attr)

        self.objects.append(self._current)


@require_GET
def get_verbose(request, default=False):
    verbose = request.GET.get('verbose')

    if verbose is not None:
        verbose = verbose.lower() in ['1', 'true']

    return verbose if verbose is not None else default
