# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import csv
import logging
import os
import re

from django.db import models

from common.management import (
    jenkins_get,
    JenkinsBaseCommand,
)
from eventstat.models import (
    EventstatImage,
    EventstatLog,
    EventstatMachine,
    EventstatMetric,
    EventstatResult,
    EventstatUpgrade,
    TASK_TYPE_CHOICES,
)
from performance.management import (
    get_image_build_number,
    get_run_number,
)


def whoami(fn):

    def wrapped(*args, **kwargs):
        logging.warn("fn: %s", fn.__name__)

        return fn(*args, **kwargs)

    return wrapped


class Command(JenkinsBaseCommand):

    job_regex = re.compile(
        ur"eventstat-([^-]+)-([^-]+)-([^-]+)-(install|upgrade)-([^-]+)-(.*)",
    )

    history_days = 30

    def _get_eventstat_logs(self, artifacts):
        """ Check for valid eventstat results.

        :returns: a list of eventstat logs.

        """

        logs = {}
        eventstat_log_regex = re.compile(
            ur'^(results\/|)(\d+\/|)eventstat.csv',
        )
        log_regex = re.compile(ur'^(results\/|)(\d+\/|)eventstat.(csv,log)')
        utah_log_regex = re.compile(ur'^(clientlogs\/|)utah.*.yaml')

        for artifact in artifacts:
            logging.info("artifact: %s", artifact)
            path = artifact['relativePath']
            num = get_run_number(path, prefix="results\/")

            if utah_log_regex.match(path):
                logging.debug("found %s", path)
                logs['utah'] = artifact

            if num is None:
                num = 1

            if num not in logs:
                logs[num] = {
                    'data_file': None,
                    'logs': [],
                }

            logging.debug("artifact: %s", path)
            if eventstat_log_regex.match(path):
                logging.debug("found %s", path)
                logs[num]['data_file'] = artifact
            if log_regex.match(path):
                logging.debug("found eventstat log %s", path)
                logs[num]['logs'].append(artifact)

        return logs

    def _process_logs(self, build, build_date, install_data, func):
        """ Process eventstat logs calling func() for each log. """

        if not hasattr(func, '__call__'):
            raise Exception("invalid callable: {}".format(func))

        unzip_path = self.get_artifacts(build)
        install_data['unzip_path'] = unzip_path

        logs = self._get_eventstat_logs(build['artifacts'])
        if 'data_file' in logs:
            logging.info("logs: %s", logs)

        image_build_number = get_image_build_number(build)

        utah_log = logs.pop('utah', None)

        kernel = self.get_kernel(build['url'], utah_log)

        unzip_path = '{}artifact'.format(build['url'])
        online_path = unzip_path

        if 'unzip_path' in install_data:
            unzip_path = "{}/archive".format(install_data['unzip_path'])

        for num in logs:
            # skip incomplete logs
            if logs[num]['data_file'] is None:
                logging.info(
                    "missing data: %s",
                    logs[num]['data_file'],
                )
                continue

            log_name = logs[num]['data_file']['relativePath']
            log_path = "{}/{}".format(
                unzip_path,
                log_name,
            )
            log_path_url = "{}/{}".format(
                online_path,
                log_name,
            )
            logging.info("log_path_url: %s", log_path_url)

            install_type = install_data['method']
            dashboard_data = dict(
                release=install_data['release'],
                variant=install_data['variant'],
                arch=install_data['arch'],
                build_date=build_date,
                kernel=kernel,
            )
            if 'jenkins_build' in install_data:
                dashboard_data['jenkins_build'] = install_data['jenkins_build']

            if install_type == 'install':
                dashboard_data['build_number'] = image_build_number
            else:
                dashboard_data['date'] = build_date

                # XXX: if we start having builds that upgrade across
                # releases we will need to update this
                dashboard_data['from_release'] = install_data['release']

            # Assume we need to retrieve the logs if there's no 'unzip_path'
            if 'unzip_path' not in install_data:
                logging.warn("Getting logs from jenkins directly...")
                dashboard_data['output'] = jenkins_get(
                    log_path_url,
                    as_json=False,
                )
            else:
                dashboard_data['output'] = self.get_local_file(log_path)
                dashboard_data['csv'] = log_path

            dashboard_data['log_path'] = log_path
            dashboard_data['log_path_url'] = log_path_url
            dashboard_data['log_name'] = log_name

            func(dashboard_data)

    def _parse_task(self, task):

        if task[0] == '[':
            return task.split()[0]

        return os.path.basename(task.split()[0])

    def _task_type(self, task):

        if task[0] == '[':
            return TASK_TYPE_CHOICES[0][1]
        else:
            return TASK_TYPE_CHOICES[1][1]

    def _parse_log(self, filename):
        """ Parse an eventstat.csv file. """

        mapping = {
            'Task': 'task',
            'Init Function': 'init_function',
            'Callback': 'callback',
            'Total': 'total',
            'Max': 'max',
            'Min': 'min',
            'Average': 'avg',
            'Std.Dev.': 'stddev',
        }

        data = []
        with open(filename, 'rb') as f:

            reader = csv.reader(f)

            for row in reader:
                row_key = row[0].rstrip(':')
                logging.info("row_key type: %s", type(row_key))

                try:
                    interval = int(row_key)
                except ValueError:
                    interval = None

                if row_key not in mapping and interval is None:
                    logging.warn("unkown row: %s", row_key)
                    continue

                field = mapping.get(row_key, row_key)

                # strip off row key
                row = row[1:]
                for i in range(len(row)):
                    if len(data) <= i:
                        data.insert(i, dict())

                    value = row[i]

                    # Add interval values to a sub-array.
                    if interval is not None:
                        if 'intervals' not in data[i]:
                            data[i]['intervals'] = []

                        data[i]['intervals'].append(
                            dict(
                                interval=interval,
                                value=value,
                            ),
                        )
                        continue

                    if field == 'task':
                        data[i]['command'] = value
                        value = self._parse_task(value)
                        logging.info("value: %s", value)
                        data[i]['task_type'] = self._task_type(value)

                    data[i][field] = value

        return data

    def _valid_log_format(self, filename):
        """ Determine if an eventstat log is valid.
        Right now this means it has a 'Max:' row.

        """

        valid = False
        with open(filename, 'rb') as f:

            reader = csv.reader(f)

            for row in reader:
                if row[0].rstrip(':') == 'Max':
                    valid = True
                    break

        return valid

    def add_result(self, dashboard_data):
        """ Add eventstat data to the database. """

        if not self._valid_log_format(dashboard_data['csv']):
            logging.info(
                "Invalid log format: %s",
                dashboard_data['log_path_url'],
            )
            return

        logging.info("Adding eventstat result.")
        logging.info(
            "Found good log: %s",
            dashboard_data['log_path_url'],
        )

        items = self._parse_log(dashboard_data['csv'])
        logging.info("items: %s", items)

        if 'build_number' in dashboard_data:
            image, new_image = EventstatImage.objects.get_or_create(
                release=dashboard_data['release'],
                variant=dashboard_data['variant'],
                arch=dashboard_data['arch'],
                build_number=dashboard_data['build_number'],
                # fake an md5 since we don't have the data
                md5="{}{}{}{}".format(
                    self.arch,
                    self.release,
                    self.variant,
                    dashboard_data['build_number'],
                )
            )
            self.image = image
            self.upgrade = None
        else:  # upgrade
            upg, new_upgr = EventstatUpgrade.objects.get_or_create(
                release=dashboard_data['release'],
                variant=dashboard_data['variant'],
                arch=dashboard_data['arch'],
                date=dashboard_data['date'],
                from_release=dashboard_data['from_release'],
            )
            self.upgrade = upg
            self.image = None

        log, new_log = EventstatLog.objects.get_or_create(
            path=dashboard_data['log_path_url'],
            name=dashboard_data['log_name'],
        )

        jenkins_url = ""
        if 'jenkins_build' in dashboard_data:
            jenkins_url = dashboard_data['jenkins_build'].url

        metrics = {}
        task_types = [x[1] for x in TASK_TYPE_CHOICES]
        task_types.append('total')
        for task_type in task_types:
            metric, new_metric = EventstatMetric.objects.get_or_create(
                image=self.image,
                upgrade=self.upgrade,
                machine=self.machine,
                name=task_type,
                kernel=dashboard_data['kernel'],
                defaults=dict(
                    ran_at=dashboard_data['build_date'],
                    publish=False,
                ),
            )

            metrics[task_type] = metric

        for item in items:
            logging.info("item: %s", item)
            task_type = item['task_type']

            for interval in item['intervals']:
                value = int(float(interval['value']))
                interval = int(float(interval['interval']))

                result, new_result = EventstatResult.objects.get_or_create(
                    image=self.image,
                    upgrade=self.upgrade,
                    machine=self.machine,
                    ran_at=dashboard_data['build_date'],
                    kernel=dashboard_data['kernel'],
                    metric=metrics[task_type],
                    interval=interval,
                    task=item['task'],
                    name=item['task'][:200],
                    task_type=task_type,
                    init_function=item['init_function'],
                    callback=item['callback'],
                    value=value,
                    log=log,
                    jenkins_url=jenkins_url,
                )

        for task_type in task_types:
            values = [
                'image',
                'upgrade',
                'machine',
            ]

            results = EventstatResult.objects.filter(
                publish=True,
                image=self.image,
                upgrade=self.upgrade,
                machine=self.machine,
            )

            if task_type != "total":
                results = results.filter(
                    task_type=task_type,
                )
                values.append('task_type')

            results = results.values(*values).annotate(
                models.Avg('value'),
                models.Min('value'),
                models.Max('value'),
                models.StdDev('value'),
            )

            agg_len = len(results)
            if agg_len != 1:
                logging.warn("Strange aggregation result count: %s", agg_len)
                logging.warn("  machine.id: %s", self.machine.id)
                if self.image is not None:
                    logging.warn("  image.id: %s", self.image.id)
                if self.upgrade is not None:
                    logging.warn("  upgrade.id: %s", self.upgrade.id)
                logging.warn("  task_type: %s", task_type)
                return

            result = results[0]

            metrics[task_type].average = result['value__avg']
            metrics[task_type].minimum = result['value__min']
            metrics[task_type].maximum = result['value__max']
            metrics[task_type].stddev = result['value__stddev']
            metrics[task_type].publish = True
            metrics[task_type].save()

    def remove_result(self, dashboard_data):
        """ Remove eventstat data from the database. """

        logging.info("Removing eventstat result.")

    def extract_data(self, name):
        m = self.job_regex.match(name)

        if m:
            self.release = m.group(1)
            self.variant = m.group(2)
            self.arch = m.group(3)
            self.method = m.group(4)
            self.workload = m.group(5)
            self.machine_name = m.group(6)

            logging.info("name: %s", name)
            logging.info("method: %s", self.method)

            self.install_data = dict(
                release=self.release,
                variant=self.variant,
                arch=self.arch,
                method=self.method,
            )

    def process_job(self, job):
        logging.info("job.name: %s", job['name'])
        self.machine, new_machine = EventstatMachine.objects.get_or_create(
            name=self.machine_name,
        )
