# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.db import models

from performance.models import (
    ImageBase,
    UpgradeBase,
    MachineBase,
    MetricBase,
    ResultBase,
)

from common.models import DashboardBaseModel


class EventstatMachine(MachineBase):

    class Meta:
        db_table = "eventstat_machines"

    @property
    def metric_count(self):
        return self.eventstatmetric_set.count()

    @property
    def metrics(self):
        return self.eventstatmetric_set.filter(
            publish=True,
        )

    @property
    def image_metrics(self):
        return self.eventstatmetric_set.filter(
            publish=True,
            upgrade=None,
        )

    @property
    def upgrade_metrics(self):
        return self.eventstatmetric_set.filter(
            publish=True,
            image=None,
        )

    @property
    def has_image(self):
        images = self.metrics.values('image')
        images = images.filter(upgrade=None).distinct()

        return images.count() > 0

    @property
    def has_upgrade(self):
        upgrades = self.metrics.values('upgrade')
        upgrades = upgrades.filter(image=None).distinct()

        return upgrades.count() > 0


class EventstatImage(ImageBase):

    class Meta:
        db_table = "eventstat_image"

    @property
    def metric_count(self):
        return self.eventstatmetric_set.count()

    @property
    def metrics(self):
        return self.eventstatmetric_set.all()


class EventstatUpgrade(UpgradeBase):

    class Meta:
        db_table = "eventstat_upgrade"

    @property
    def metric_count(self):
        return self.eventstatmetric_set.count()

    @property
    def metrics(self):
        return self.eventstatmetric_set.all()

TASK_TYPE_CHOICES = [
    ('Kernel', 'kernel'),
    ('Userland', 'userland'),
]


class EventstatDetail(DashboardBaseModel):
    """ Eventstat raw data. """

    class Meta:
        db_table = "eventstat_details"

    image = models.ForeignKey(EventstatImage, null=True)
    upgrade = models.ForeignKey(EventstatUpgrade, null=True)
    machine = models.ForeignKey(EventstatMachine)
    task = models.CharField(max_length=1024)
    command = models.CharField(max_length=4096)
    task_type = models.CharField(
        max_length=100,
        choices=TASK_TYPE_CHOICES,
        default=TASK_TYPE_CHOICES[:-1],
    )
    init_function = models.CharField(max_length=1024)
    callback = models.CharField(max_length=1024)
    total = models.PositiveIntegerField()
    interval = models.PositiveIntegerField()
    ran_at = models.DateTimeField("date run")
    result = models.ForeignKey("EventstatResult")

    def __unicode__(self):
        return "{}:{}:{}:{}".format(
            self.task,
            self.init_function,
            self.callback,
            self.total,
        )


class EventstatResult(ResultBase):
    """ Eventstat result model. """

    class Meta:
        db_table = "eventstat_results"

    image = models.ForeignKey(EventstatImage, null=True)
    upgrade = models.ForeignKey(EventstatUpgrade, null=True)
    machine = models.ForeignKey(EventstatMachine)
    kernel = models.CharField(max_length=100, null=True)
    metric = models.ForeignKey('EventstatMetric')
    log = models.ForeignKey('EventstatLog', null=True)
    interval = models.PositiveIntegerField()
    task = models.CharField(max_length=1024)
    command = models.CharField(max_length=4096)
    task_type = models.CharField(
        max_length=100,
        choices=TASK_TYPE_CHOICES,
        default=TASK_TYPE_CHOICES[:-1],
    )
    init_function = models.CharField(max_length=1024)
    callback = models.CharField(max_length=1024)

    @property
    def logs(self):
        return EventstatLog.objects.filter(
            eventstatresult__image=self.image,
            eventstatresult__upgrade=self.upgrade,
            eventstatresult__machine=self.machine,
        ).distinct()

    @property
    def log_count(self):
        return self.logs.count()

    @property
    def method(self):
        if self.image is not None:
            return "install"
        else:
            return "upgrade"

    @property
    def arch(self):
        if self.image is not None:
            return self.image.arch
        else:
            return self.upgrade.arch


class EventstatMetric(MetricBase):
    """ Eventzstat metric model. """

    class Meta:
        db_table = "eventstat_metrics"

    image = models.ForeignKey(EventstatImage, null=True)
    upgrade = models.ForeignKey(EventstatUpgrade, null=True)
    machine = models.ForeignKey(EventstatMachine)
    kernel = models.CharField(max_length=100, null=True)

    @property
    def logs(self):
        return EventstatLog.objects.filter(
            eventstatresult__image=self.image,
            eventstatresult__upgrade=self.upgrade,
            eventstatresult__machine=self.machine,
        ).distinct()

    @property
    def log_count(self):
        return self.logs.count()


class EventstatLog(DashboardBaseModel):
    """ Eventstat logs for a build. """

    class Meta:
        db_table = "eventstat_logs"

    path = models.CharField(max_length=512)
    name = models.CharField(max_length=100)


eventstat_models = [
    EventstatMachine,
    EventstatImage,
    EventstatUpgrade,
    EventstatDetail,
    EventstatResult,
    EventstatMetric,
    EventstatLog,
]
