# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import django_tables2 as tables
from django_tables2.utils import Accessor as A

from eventstat.models import (
    EventstatResult,
    EventstatMetric,
)


class ResultTable(tables.Table):

    class Meta:
        attrs = {'class': 'basic wide'}
        model = EventstatResult
        fields = (
            'value',
        )

    value = tables.TemplateColumn(
        "{{ record.value|floatformat }}",
        verbose_name="Wakeups",
        attrs={'td': {'class': 'num'}},
    )


class MetricTable(tables.Table):
    class Meta:
        attrs = {'class': 'basic wide'}
        model = EventstatMetric
        fields = (
            'average',
            'minimum',
            'maximum',
            'stddev',
        )

    def __init__(self, *args, **kwargs):

        days = kwargs.pop('days', None)

        if days is not None:
            for column in self.base_columns.values():
                if isinstance(column, tables.LinkColumn):
                    column.kwargs['days'] = days
        super(MetricTable, self).__init__(*args, **kwargs)

    average = tables.TemplateColumn(
        "{{ record.average|floatformat }}",
        verbose_name="Average",
        attrs={'td': {'class': 'num'}},
    )
    minimum = tables.TemplateColumn(
        "{{ record.minimum|floatformat }}",
        verbose_name="Minimum",
        attrs={'td': {'class': 'num'}},
    )
    maximum = tables.TemplateColumn(
        "{{ record.maximum|floatformat }}",
        verbose_name="Maximum",
        attrs={'td': {'class': 'num'}},
    )
    stddev = tables.TemplateColumn(
        "{{ record.stddev|floatformat }}",
        verbose_name="Standard Deviation",
        attrs={'td': {'class': 'num'}},
    )


class TaskInstallTable(MetricTable):

    class Meta(MetricTable.Meta):
        fields = (
            "task",
        )
        order_by = ("-average", "task")

    task = tables.LinkColumn(
        "eventstat.views.task_overview",
        kwargs={
            "arch": A("image__arch"),
            "machine_id":  A("machine_id"),
            "method": "install",
            "task": A("task"),
        },
    )


class TaskUpgradeTable(MetricTable):

    class Meta(MetricTable.Meta):
        fields = (
            "task",
        )
        order_by = ("-average", "task")

    task = tables.LinkColumn(
        "eventstat.views.task_overview",
        kwargs={
            "arch": A("upgrade__arch"),
            "machine_id":  A("machine_id"),
            "method": "upgrade",
            "task": A("task"),
        },
    )


class TaskInstallDetailTable(ResultTable):

    class Meta(ResultTable.Meta):
        fields = (
            "image__build_number",
        )
        order_by = ("-image__build_number", "-value")

    image__build_number = tables.LinkColumn(
        'eventstat.views.image_task_details',
        accessor="image.build_number",
        kwargs={
            'task': A('task'),
            'image_id': A('image.id'),
            'machine_id': A('machine.id'),
        },
        verbose_name="Build Number",
    )


class TaskUpgradeDetailTable(ResultTable):

    class Meta(ResultTable.Meta):
        fields = (
            "upgrade__date",
        )
        order_by = ("-upgrade__date", "-value")

    upgrade__date = tables.LinkColumn(
        'eventstat.views.upgrade_task_details',
        accessor='upgrade.date',
        kwargs={
            'upgrade_id': A('upgrade.id'),
            'task': A('task'),
        },
        verbose_name='Upgrade Date',
    )


class TaskUpgradeOverviewTable(MetricTable):

    class Meta(MetricTable.Meta):
        order_by = ("-upgrade__date",)
        fields = (
            'upgrade__date',
        )

    upgrade__date = tables.LinkColumn(
        'eventstat.views.upgrade_task_details',
        accessor="upgrade__date",
        kwargs={
            'task': A('task'),
            'upgrade_id': A('upgrade__id'),
        },
        verbose_name="Upgrade Date",
    )


class TaskInstallOverviewTable(MetricTable):

    class Meta(MetricTable.Meta):
        order_by = ("-image__build_number",)
        fields = (
            "image__build_number",
        )

    image__build_number = tables.LinkColumn(
        'eventstat.views.image_task_details',
        kwargs={
            'task': A('task'),
            'image_id': A('image__id'),
            'machine_id': A('machine__id'),
        },
        verbose_name="Build Number",
    )


class UpgradeOverviewTable(MetricTable):

    class Meta(MetricTable.Meta):
        order_by = ("-upgrade__date",)
        fields = (
            'upgrade__date',
        )

    upgrade__date = tables.LinkColumn(
        'eventstat.views.upgrade_details',
        accessor="upgrade.date",
        kwargs={
            'task_type': 'all',
            'upgrade_id': A('upgrade.id'),
            'machine_id': A('machine.id'),
        },
        verbose_name="Upgrade Date",
    )


class InstallOverviewTable(MetricTable):

    class Meta(MetricTable.Meta):
        order_by = ("-image__build_number",)
        fields = (
            "image__build_number",
        )

    image__build_number = tables.LinkColumn(
        'eventstat.views.image_details',
        accessor="image.build_number",
        kwargs={
            'task_type': 'all',
            'image_id': A('image.id'),
            'machine_id': A('machine.id'),
        },
        verbose_name="Build Number",
    )


class InstallDetailTable(ResultTable):

    class Meta(ResultTable.Meta):
        fields = (
            'task',
        )
        order_by = ('-value')

    task = tables.LinkColumn(
        'eventstat.views.arch_breakdown',
        kwargs={
            'machine_id': A('machine.id'),
            'arch': A('arch'),
            'method': A('method'),
            'task_type': A('task_type'),
        }
    )
'''
    class Meta(ResultTable.Meta):
        fields = (
            "image__build_number",
        )
        order_by = ("-image__build_number", "-value")

    image__build_number = tables.LinkColumn(
        'eventstat.views.image_task_details',
        accessor="image.build_number",
        kwargs={
            'task': A('task'),
            'image_id': A('image.id'),
            'machine_id': A('machine.id'),
        },
        verbose_name="Build Number",
    )
'''
