# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import datetime
import json

from django.test import TestCase
from django.test.client import Client

from eventstat.models import (
    EventstatDetail,
    EventstatImage,
    EventstatLog,
    EventstatMachine,
    EventstatMetric,
    EventstatResult,
    EventstatUpgrade,
)
from eventstat.utils import (
    get_arches,
    get_images,
    get_machines,
)


def _setUp(self):
    """ Setup commonly used objects. """

    self.release = 'raring'
    self.variant = 'desktop'
    self.arch = 'amd64'
    self.build_number = '20130201'
    self.md5 = 'blah'

    self.image = EventstatImage.objects.create(
        release=self.release,
        variant=self.variant,
        arch=self.arch,
        build_number=self.build_number,
        md5=self.md5,
    )

    self.assert_(self.image)

    self.upgrade_data = {
        'release': 'raring',
        'from_release': 'precise',
        'date': datetime.datetime.now(),
        'variant': 'desktop',
        'arch': 'amd64',
    }

    self.upgrade = EventstatUpgrade.objects.create(
        release=self.upgrade_data['release'],
        from_release=self.upgrade_data['from_release'],
        arch=self.upgrade_data['arch'],
        variant=self.upgrade_data['variant'],
        date=self.upgrade_data['date'],
    )

    self.assert_(self.upgrade)

    self.machine_name = 'eventstat-test-1'
    self.machine_mac_address = '00:00:00:00:01'
    self.machine = EventstatMachine.objects.create(
        name=self.machine_name,
        mac_address=self.machine_mac_address,
    )

    self.assert_(self.machine)

    self.metric_name = "pss"

    self.metric = EventstatMetric.objects.create(
        image=self.image,
        machine=self.machine,
        ran_at=datetime.datetime.now(),
        name=self.metric_name,
    )

    self.assert_(self.metric)

    self.log = EventstatLog.objects.create(
        name="blah",
        path="http://jenkins.qa.ubuntu.com/blah",
    )

    self.assert_(self.log)

    self.result = EventstatResult.objects.create(
        image=self.image,
        machine=self.machine,
        value=0,
        ran_at=datetime.datetime.now(),
        name=self.metric.name,
        metric=self.metric,
        log=self.log,
        interval=20,
    )

    self.assert_(self.result)

    self.detail_init_function = "init_function"
    self.detail_callback = "callback"
    self.detail_total = 100
    self.detail = EventstatDetail.objects.create(
        image=self.image,
        machine=self.machine,
        init_function=self.detail_init_function,
        callback=self.detail_callback,
        total=self.detail_total,
        ran_at=datetime.datetime.now(),
        result=self.result,
        interval=20,
    )

    self.assert_(self.detail)

    self.metric2 = EventstatMetric.objects.create(
        upgrade=self.upgrade,
        machine=self.machine,
        name=self.metric_name,
        ran_at=datetime.datetime.now(),
    )

    self.assert_(self.metric2)


class TestEventstat(TestCase):
    """ Tests for eventstat models. """

    def setUp(self):
        """ Common setup. """

        _setUp(self)

    def test_machine_has_image(self):
        """ Test that a machine has image metrics. """

        self.assertTrue(self.machine.has_image)

    def test_machine_has_image_none(self):
        """ Test has_image for no metrics. """

        self.image.metrics.delete()
        self.assertFalse(self.machine.has_image)

    def test_machine_has_upgrade_none(self):
        """ Test has_upgrade for no metrics. """

        self.upgrade.metrics.delete()
        self.assertFalse(self.machine.has_upgrade)

    def test_machine_has_upgrade(self):
        """ Test that a machine has upgrade metrics. """

        self.assertTrue(self.machine.has_upgrade)

    def test_detail(self):
        """ Test the detail model. """

        self.assertEqual(self.detail_init_function, self.detail.init_function)
        self.assertEqual(self.detail_callback, self.detail.callback)
        self.assertEqual(self.detail_total, self.detail.total)

    def test_image(self):
        """ Test the image model. """

        self.assertEqual(self.image.release, self.release)
        self.assertEqual(self.image.variant, self.variant)
        self.assertEqual(self.image.arch, self.arch)
        self.assertEqual(self.image.build_number, self.build_number)
        self.assertEqual(self.image.md5, self.md5)

    def test_upgrade(self):
        """ Test the upgrade model. """

        self.assertEqual(
            self.upgrade_data['release'],
            self.upgrade.release,
        )
        self.assertEqual(
            self.upgrade_data['from_release'],
            self.upgrade.from_release,
        )
        self.assertEqual(self.upgrade_data['arch'], self.upgrade.arch)
        self.assertEqual(self.upgrade_data['variant'], self.upgrade.variant)
        self.assertEqual(self.upgrade_data['date'], self.upgrade.date)

    def test_machine(self):
        """ Test the machine model. """

        self.assertEqual(self.machine.name, self.machine_name)
        self.assertEqual(self.machine.mac_address, self.machine_mac_address)

    def test_image_metric(self):
        """ Test that a metric has valid values. """

        metric = self.result.metric

        self.assertEqual(metric.average, self.result.value)
        self.assertEqual(metric.minimum, self.result.value)
        self.assertEqual(metric.maximum, self.result.value)
        self.assertEqual(metric.stddev, 0)
        self.assertEqual(metric.name, self.result.name)

    def test_metric_logs(self):
        """ Test that metric log counts are correct. """

        self.assertEqual(self.metric.log_count, 1)


class TestAPI(TestCase):
    """ Test Eventstat API. """

    def setUp(self):
        """ Common setup. """

        _setUp(self)
        self.url_prefix = '/api/power/eventstat/'

    def url(self, part):
        """ URL creator. """

        return "{}{}".format(
            self.url_prefix,
            part,
        )

    def test_images(self):
        """ Test images api. """

        c = Client()
        url = self.url('image/')
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = json.loads(response.content)

        self.assertEqual(len(data), 1)

    def test_image_arch(self):
        """ Test images api. """

        c = Client()
        url = self.url('image/arch/amd64/')
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = json.loads(response.content)

        self.assertEqual(len(data), 1)

    def test_image_nonexistent_arch(self):
        """ Test images api. """

        c = Client()
        url = self.url('image/arch/i386/')
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = json.loads(response.content)

        self.assertEqual(len(data), 0)


class TestUtils(TestCase):
    """ Test eventstat.utils methods. """

    def setUp(self):
        _setUp(self)

    def test_get_arches(self):
        """ Test that get_arches() returns correct results. """

        self.assertEqual(len(get_arches()), 1)

    def test_get_machines(self):
        """ Test that get_machines() returns correct results. """

        machines = EventstatMachine.objects.filter(publish=True)

        machines = machines.filter(eventstatmetric__upgrade=None)

        self.assertEqual(machines.count(), 1)

    def test_get_machines_with_arch(self):
        """
        Test that get_machines() returns correct results when supplied an arch.

        """

        self.assertEqual(get_machines(arch="amd64").count(), 1)

    def test_get_machines_with_nonexistent_arch(self):
        """ Test that get_machines() returns zero results for bad arch. """

        self.assertEqual(len(get_machines(arch="fake")), 0)

    def test_get_images(self):
        """ Test that get_images() returns correct results. """

        self.assertEqual(get_images().count(), 1)

    def test_get_images_with_arch(self):
        """
        Test that get_images() returns correct results when given an arch.

        """

        self.assertEqual(get_images(arch="amd64").count(), 1)

    def test_get_images_with_invalid_arch(self):
        """
        Test that get_images() returns zero results when given an invalid arch.

        """

        self.assertEqual(get_images(arch="fake").count(), 0)
