# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.core.urlresolvers import reverse
from django.db import models
from django.shortcuts import (
    redirect,
    render_to_response,
    get_object_or_404,
)
from django.template import RequestContext
from django.views.decorators.http import require_GET


from django_tables2 import RequestConfig

from eventstat.models import (
    EventstatImage,
    EventstatMachine,
    EventstatMetric,
    EventstatResult,
    EventstatUpgrade,
    TASK_TYPE_CHOICES,
)
from eventstat.tables import (
    TaskInstallTable,
    TaskUpgradeTable,
    TaskInstallDetailTable,
    TaskUpgradeDetailTable,
    TaskInstallOverviewTable,
    TaskUpgradeOverviewTable,
    InstallOverviewTable,
    UpgradeOverviewTable,
)
from eventstat.utils import (
    get_arches,
    get_images,
    get_machines,
    get_upgrades,
)

# number of days worth of results to display/calculate
DAY_OPTIONS = [7, 30, 90]
DAYS = DAY_OPTIONS[0]
TASK_TYPES = ['all'] + [x[1] for x in TASK_TYPE_CHOICES]


def _generate_day_urls(url, kwargs, days=DAY_OPTIONS):

    retval = []

    for day in days:
        kwargs['days'] = day
        ret_url = reverse(
            url,
            kwargs=kwargs,
        )
        retval.append((day, ret_url))

    return retval


def _generate_task_type_urls(url, kwargs, task_types=TASK_TYPES):

    retval = []

    for task_type in task_types:
        kwargs['task_type'] = task_type
        ret_url = reverse(
            url,
            kwargs=kwargs,
        )
        retval.append((task_type, ret_url))

    return retval


@require_GET
def arch_overview(request, arch=None, days=None):
    arches = get_arches()

    if days is None:
        days = DAYS
    else:
        days = int(days)

    if arch is None:
        arch = arches[0] if len(arches) > 0 else 'amd64'
        return redirect(
            reverse(
                'eventstat_arch_overview',
                kwargs={
                    'arch': arch,
                    'days': days,
                }
            ),
            #'arch/{}'.format(arches[0] if len(arches) > 0 else "amd64"),
        )

    images = get_images(arch)
    upgrades = get_upgrades(arch)
    machines = get_machines(arch)

    for machine in machines:
        metrics = EventstatMetric.objects.filter(
            publish=True,
            machine=machine,
            image__arch=arch,
            name='total',
            ran_at__gt=datetime.datetime.now() - datetime.timedelta(days=days),
        ).order_by('-image__build_number')

        table = InstallOverviewTable(
            metrics,
            days=days,
            prefix="{}-install".format(machine.id),
        )
        RequestConfig(request, paginate={"per_page": 100}).configure(table)
        machine.install_table = table

        metrics = EventstatMetric.objects.filter(
            publish=True,
            machine=machine,
            upgrade__arch=arch,
            name='total',
            ran_at__gt=datetime.datetime.now() - datetime.timedelta(days=days),
        ).order_by('-upgrade__date')

        table = UpgradeOverviewTable(
            metrics,
            days=days,
            prefix="{}-upgrade".format(machine.id),
        )
        RequestConfig(request, paginate={"per_page": 100}).configure(table)
        machine.upgrade_table = table

    data = {
        'images': images,
        'upgrades': upgrades,
        'arches': arches,
        'machines': machines,
        'arch': arch,
        'days': days,
        'day_urls': _generate_day_urls(
            'eventstat_arch_overview',
            kwargs={
                'arch': arch,
            }
        )
    }

    return render_to_response(
        'power/eventstat/arch_overview.html',
        data,
        RequestContext(request),
    )


@require_GET
def arch_breakdown(
    request,
    machine_id,
    arch,
    method,
    task_type=None,
    days=None,
):

    if task_type is None:
        task_type = "all"

    if days is None:
        days = DAYS
    else:
        days = int(days)

    machine = get_object_or_404(EventstatMachine, id=machine_id)

    values = [
        'task',
        'minimum',
        'maximum',
        'average',
        'stddev',
        'machine_id',
    ]

    if method == "install":
        tasks = EventstatResult.objects.filter(
            publish=True,
            machine__id=machine_id,
            upgrade=None,
            image__arch=arch,
            ran_at__gt=datetime.datetime.now() - datetime.timedelta(days=days)
        )
        values.append('image__arch')
    else:
        tasks = EventstatResult.objects.filter(
            publish=True,
            machine__id=machine_id,
            image=None,
            upgrade__arch=arch,
            ran_at__gt=datetime.datetime.now() - datetime.timedelta(days=days)
        )
        values.append('upgrade__arch')

    if task_type != "all":
        tasks = tasks.filter(task_type=task_type)

    tasks = tasks.values('task').annotate(
        minimum=models.Min('value'),
        maximum=models.Max('value'),
        average=models.Avg('value'),
        stddev=models.StdDev('value'),
    )

    tasks = tasks.values(
        *values
    )

    if method == "install":
        table = TaskInstallTable(tasks, days=days)
    else:
        table = TaskUpgradeTable(tasks, days=days)
    RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'table': table,
        'arch': arch,
        'machine': machine,
        'task_type': task_type,
        'method': method,
        'days': days,
        'task_type_urls': _generate_task_type_urls(
            'eventstat_arch_breakdown',
            kwargs=dict(
                machine_id=machine_id,
                arch=arch,
                method=method,
                task_type=task_type,
                days=days
            ),
        ),
    }

    return render_to_response(
        "power/eventstat/arch_breakdown.html",
        data,
        RequestContext(request),
    )


@require_GET
def task_overview(request, machine_id, arch, method, task, days=None):

    machine = get_object_or_404(EventstatMachine, id=machine_id)

    if days is None:
        days = DAYS
    else:
        days = int(days)

    if method == 'install':
        tasks = EventstatResult.objects.filter(
            publish=True,
            task=task,
            machine=machine,
            image__arch=arch,
            upgrade=None,
            ran_at__gt=datetime.datetime.now() - datetime.timedelta(days=days),
        ).values(
            'machine__id',
            'image__build_number',
            'image__arch',
            'image__id',
            'task',
        ).annotate(
            minimum=models.Min('value'),
            maximum=models.Max('value'),
            average=models.Avg('value'),
            stddev=models.StdDev('value'),
        )

        table = TaskInstallOverviewTable(tasks)
    else:
        tasks = EventstatResult.objects.filter(
            publish=True,
            task=task,
            machine=machine,
            upgrade__arch=arch,
            image=None,
            ran_at__gt=datetime.datetime.now() - datetime.timedelta(days=days),
        ).values(
            'machine__id',
            'upgrade__date',
            'upgrade__arch',
            'upgrade__id',
            'task',
        ).annotate(
            minimum=models.Min('value'),
            maximum=models.Max('value'),
            average=models.Avg('value'),
            stddev=models.StdDev('value'),
        )
        table = TaskUpgradeOverviewTable(tasks)

    RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'table': table,
        'arch': arch,
        'machine': machine,
        'task': task,
        'days': days,
        'day_urls': _generate_day_urls(
            'eventstat_task_overview',
            kwargs={
                'arch': arch,
                'machine_id': machine_id,
                'method': method,
                'task': task,
            }
        ),
    }

    return render_to_response(
        "power/eventstat/task_overview.html",
        data,
        RequestContext(request),
    )


@require_GET
def image_details(request, image_id, task_type, machine_id, days=None):
    image = get_object_or_404(EventstatImage, id=image_id)
    machine = get_object_or_404(EventstatMachine, id=machine_id)

    if days is None:
        days = 7
    else:
        days = int(days)

    results = EventstatResult.objects.filter(
        publish=True,
        image=image,
        machine=machine,
        ran_at__gt=datetime.datetime.now() - datetime.timedelta(days=days),
    ).values(
        'task',
        'image_id',
        'image__arch',
        'machine_id',
    ).annotate(
        average=models.Avg('value'),
        minimum=models.Min('value'),
        maximum=models.Max('value'),
        stddev=models.StdDev('value'),
    )

    if task_type != 'all':
        results = results.filter(task_type=task_type)

    table = TaskInstallTable(results, days=days)

    RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'table': table,
        'image': image,
        'task_type': task_type,
        'machine': machine,
        'method': "install",
        'days': days,
        'day_urls': _generate_day_urls(
            'eventstat_image_details',
            kwargs={
                'image_id': image_id,
                'machine_id': machine_id,
                'task_type': task_type,
            }
        ),
        'task_type_urls': _generate_task_type_urls(
            'eventstat_image_details',
            kwargs=dict(
                machine_id=machine_id,
                image_id=image_id,
                task_type=task_type,
                days=days
            ),
        ),
    }

    return render_to_response(
        "power/eventstat/image_details.html",
        data,
        RequestContext(request),
    )


@require_GET
def upgrade_details(request, upgrade_id, task_type, machine_id, days=None):
    upgrade = get_object_or_404(EventstatUpgrade, id=upgrade_id)
    machine = get_object_or_404(EventstatMachine, id=machine_id)

    if days is None:
        days = 7
    else:
        days = int(days)

    results = EventstatResult.objects.filter(
        publish=True,
        upgrade=upgrade,
        machine=machine,
        ran_at__gt=datetime.datetime.now() - datetime.timedelta(days=days),
    ).values(
        'task',
        'upgrade_id',
        'upgrade__arch',
        'machine_id',
    ).annotate(
        average=models.Avg('value'),
        minimum=models.Min('value'),
        maximum=models.Max('value'),
        stddev=models.StdDev('value'),
    )

    if task_type != 'all':
        results = results.filter(task_type=task_type)

    table = TaskUpgradeTable(results, days=days)

    RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'table': table,
        'upgrade': upgrade,
        'task_type': task_type,
        'machine': machine,
        'method': "upgrade",
        'days': days,
        'day_urls': _generate_day_urls(
            'eventstat_upgrade_details',
            kwargs={
                'upgrade_id': upgrade_id,
                'machine_id': machine_id,
                'task_type': task_type,
            }
        ),
        'task_type_urls': _generate_task_type_urls(
            'eventstat_upgrade_details',
            kwargs=dict(
                machine_id=machine_id,
                upgrade_id=upgrade_id,
                task_type=task_type,
                days=days
            ),
        ),
    }

    return render_to_response(
        "power/eventstat/upgrade_details.html",
        data,
        RequestContext(request),
    )


@require_GET
def image_task_details(request, image_id, task, machine_id):
    image = get_object_or_404(EventstatImage, id=image_id)
    machine = get_object_or_404(EventstatMachine, id=machine_id)

    tasks = EventstatResult.objects.filter(
        publish=True,
        image=image,
        task=task,
    )
    table = TaskInstallDetailTable(tasks)

    RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'table': table,
        'image': image,
        'task': task,
        'machine': machine,
    }

    return render_to_response(
        "power/eventstat/task_details.html",
        data,
        RequestContext(request),
    )


@require_GET
def upgrade_task_details(request, upgrade_id, task):
    tasks = EventstatResult.objects.filter(
        publish=True,
        upgrade__id=upgrade_id,
        task=task,
    )
    table = TaskUpgradeDetailTable(tasks)

    RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'table': table,
        'upgrade_id': upgrade_id,
    }

    return render_to_response(
        "power/eventstat/task_details.html",
        data,
        RequestContext(request),
    )


@require_GET
def task_details(request, machine_id, arch, method, task, days=None):

    machine = get_object_or_404(EventstatMachine, id=machine_id)

    if days is None:
        days = DAYS
    else:
        days = int(days)

    if method == 'install':
        tasks = EventstatResult.objects.filter(
            publish=True,
            task=task,
            machine=machine,
            image__arch=arch,
            upgrade=None,
            ran_at=datetime.datetime.now() - datetime.timedelta(days=days)
        )
        table = TaskInstallDetailTable(tasks)
    else:

        tasks = EventstatResult.objects.filter(
            publish=True,
            task=task,
            machine=machine,
            upgrade__arch=arch,
            image=None,
        )
        table = TaskUpgradeDetailTable(tasks)

    RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'table': table,
        'arch': arch,
        'machine': machine,
        'task': task,
    }

    return render_to_response(
        "power/eventstat/task_details.html",
        data,
        RequestContext(request),
    )


@require_GET
def machine_upgrade_details(request, arch, machine_id, result_id):
    pass


@require_GET
def machine_install_details(request, arch, machine_id, result_id):
    pass


@require_GET
def result_details(request, arch, machine_id, result_id, method):
    pass
