# QA Dashboard

# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import json
from django.views.decorators.http import require_GET
from django.shortcuts import get_object_or_404

from common.utils import (
    JSONResponse,
    MySerializer,
)

from memory.models import (
    MemoryImage,
    MemoryMachine,
    MemoryMetric,
    MemoryProcess,
    MemoryResult,
    MemoryUpgrade,
)

MAX_RESULTS = 180


@require_GET
def images(request, arch=None):
    images = MemoryImage.objects.filter(publish=True)

    if arch is not None:
        images = images.filter(arch=arch)

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize(images)

    return JSONResponse(data)


@require_GET
def image_detail(request, image_id):
    image = get_object_or_404(MemoryImage, id=image_id)

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize([image])

    return JSONResponse(data)


@require_GET
def upgrades(request, arch=None):
    upgrades = MemoryUpgrade.objects.filter(publish=True)

    if arch is not None:
        upgrades = upgrades.filter(arch=arch)

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize(upgrades)

    return JSONResponse(data)


@require_GET
def upgrade_detail(request, upgrade_id):
    upgrade = get_object_or_404(MemoryUpgrade, id=upgrade_id)

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize([upgrade])

    return JSONResponse(data)


@require_GET
def machines(request, arch=None):
    machines = MemoryMachine.objects.filter(publish=True)

    if arch is not None:
        machines = machines.filter(arch=arch)

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize(machines)

    return JSONResponse(data)


@require_GET
def machine_detail(request, machine_id):
    machine = get_object_or_404(MemoryMachine, id=machine_id)

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize([machine])

    return JSONResponse(data)


@require_GET
def image_machine_results(request, image_id, machine_id):

    results = MemoryResult.objects.filter(
        image=image_id,
        machine=machine_id,
        publish=True,
    )

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize(results)

    return JSONResponse(data)


@require_GET
def upgrade_machine_results(request, upgrade_id, machine_id):

    results = MemoryResult.objects.filter(
        upgrade=upgrade_id,
        machine=machine_id,
        publish=True,
    )

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize(results)

    return JSONResponse(data)


@require_GET
def result_detail(request, result_id):

    results = MemoryResult.objects.filter(id=result_id)

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize(results)

    return JSONResponse(data)


@require_GET
def image_machine_metrics(request, image_id, machine_id):

    metrics = MemoryMetric.objects.filter(
        image=image_id,
        machine=machine_id,
        publish=True,
    )

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize(metrics)

    return JSONResponse(data)


@require_GET
def upgrade_machine_metrics(request, upgrade_id, machine_id):

    metrics = MemoryMetric.objects.filter(
        upgrade=upgrade_id,
        machine=machine_id,
        publish=True,
    )

    serializer = MySerializer(use_extra_fields=True)
    data = serializer.serialize(metrics)

    return JSONResponse(data)


@require_GET
def upgrade_machine_arch_metrics(request, machine_id, arch, variant,
                                 metric='pss'):
    res = MemoryMetric.objects.filter(
        machine=machine_id,
        upgrade__arch=arch,
        upgrade__variant=variant,
        name=metric,
        publish=True,
    ).order_by('upgrade__date')[:MAX_RESULTS]

    last_releases = MemoryImage.objects.filter(
        memorymetric__machine=machine_id,
        arch=arch,
        image_type='release',
        publish=True,
    ).order_by('build_number')

    if last_releases.count() > 0:
        last_release_avg = last_releases[0].boot_avg
        last_release_name = last_releases[0].image.release
    else:
        last_release_avg = 0.0
        last_release_name = 'Unknown'

    bundle = {'last_release_avg': last_release_avg,
              'last_release_name': last_release_name}

    image_data = []
    for r in res:
        image_info = dict(
            bottom=r.minimum,
            min=r.minimum,
            max=r.maximum,
            avg=r.average,
            std=r.stddev,
            build_number=r.upgrade.date.isoformat(),
        )
        image_data.append(image_info)

    bundle['count'] = res.count()
    bundle['image_data'] = image_data

    data = json.dumps(bundle)

    return JSONResponse(data)


@require_GET
def image_machine_arch_metrics(request, machine_id, arch, variant,
                               metric='pss'):
    res = MemoryMetric.objects.filter(
        machine=machine_id,
        image__arch=arch,
        name=metric,
        publish=True,
    ).order_by('image__build_number')[:MAX_RESULTS]

    last_releases = MemoryImage.objects.filter(
        memorymetric__machine=machine_id,
        arch=arch,
        image_type='release',
        publish=True,
    ).order_by('-build_number')

    if last_releases.count() > 0:
        last_release_avg = last_releases[0].boot_avg
        last_release_name = last_releases[0].image.release
    else:
        last_release_avg = 0.0
        last_release_name = 'Unknown'

    bundle = {'last_release_avg': last_release_avg,
              'last_release_name': last_release_name}

    image_data = []
    for r in res:
        image_info = dict(
            bottom=r.minimum,
            min=r.minimum,
            max=r.maximum,
            avg=r.average,
            std=r.stddev,
            build_number=r.image.build_number,
        )
        image_data.append(image_info)

    bundle['count'] = res.count()
    bundle['image_data'] = image_data

    data = json.dumps(bundle)

    return JSONResponse(data)


@require_GET
def image_machine_arch_processes(request, machine_id, arch, process=None):
    procs = MemoryProcess.objects.filter(
        machine=machine_id,
        image__arch=arch,
        publish=True,
    ).order_by('image__build_number')

    if process is not None:
        procs = procs.filter(process=process)

    proc_data = []
    for p in procs:
        proc_info = dict(
            bottom=p.minimum,
            min=p.minimum,
            max=p.maximum,
            avg=p.average,
            std=p.stddev,
            process=p.process,
            build_number=p.image.build_number,
        )
        proc_data.append(proc_info)

    bundle = dict(
        proc_data=proc_data,
        count=procs.count(),
    )

    data = json.dumps(bundle)

    return JSONResponse(data)


@require_GET
def upgrade_machine_arch_processes(request, machine_id, arch, process=None):
    procs = MemoryProcess.objects.filter(
        machine=machine_id,
        upgrade__arch=arch,
        publish=True,
    ).order_by('upgrade__date')

    if process is not None:
        procs = procs.filter(process=process)

    proc_data = []
    for p in procs:
        proc_info = dict(
            bottom=p.minimum,
            min=p.minimum,
            max=p.maximum,
            avg=p.average,
            std=p.stddev,
            process=p.process,
            build_number=p.upgrade.date.strftime("%Y-%m-%d"),
        )
        proc_data.append(proc_info)

    bundle = dict(
        proc_data=proc_data,
        count=procs.count(),
    )

    data = json.dumps(bundle)

    return JSONResponse(data)
