# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import logging

from django.core.management.base import BaseCommand
from django.db import models

from memory.models import (
    MemoryResult,
    MemoryMetric,
)


class Command(BaseCommand):
    help = "Do result calculations."

    def _aggregate(self, name, image, upgrade, machine):
        """ Aggregate values by name. """

        metric, new_metric = MemoryMetric.objects.get_or_create(
            image=image,
            upgrade=upgrade,
            machine=machine,
            name=name,
        )

        results = MemoryResult.objects.filter(
            name=name,
            image=image,
            upgrade=upgrade,
            machine=machine,
            publish=True,
        )

        agg_results = results.values(
            'image',
            'upgrade',
            'machine',
            'name',
        ).annotate(
            models.Avg('value'),
            models.Min('value'),
            models.Max('value'),
            models.StdDev('value'),
        )

        # Remove unlinked MemoryMetrics
        # XXX: if we add support for direct MemoryMetric insertion
        # this logic will need to updated.
        MemoryMetric.objects.filter(publish=True).annotate(
            models.Count('memoryresult')
        ).filter(memoryresult__count=0).exclude(
            id=metric.id
        ).delete()

        if results.count() == 0:
            MemoryMetric.objects.filter(
                machine=self.machine,
                image=self.image,
                upgrade=self.upgrade,
                name=self.name,
            ).delete()
            return

        agg_len = len(agg_results)
        if agg_len != 1:
            logging.warn("strange aggregation result count: {}".format(
                agg_len,
            ))
            return

        agg_results = agg_results[0]

        latest_ran_at = results.aggregate(
            models.Max('ran_at')
        )['ran_at__max']

        metric.stddev = agg_results['value__stddev']
        metric.minimum = agg_results['value__min']
        metric.maximum = agg_results['value__max']
        metric.average = agg_results['value__avg']
        metric.ran_at = latest_ran_at
        metric.publish = True
        metric.save()

        return metric

    def handle(self, *args, **options):

        verbosity = int(options.get('verbosity'))

        log_level = logging.WARN
        if verbosity == 2:
            log_level = logging.INFO
        if verbosity > 2:
            log_level = logging.DEBUG

        logging.basicConfig(level=log_level)

        logging.info("Starting...")
        results = MemoryResult.objects.filter(publish=True)
        results = results.distinct('name', 'image', 'upgrade', 'machine')

        logging.info("results: {}".format(results))
        for result in results.iterator():
            logging.info("name: {}".format(result.name))
            metric = self._aggregate(
                result.name,
                result.image,
                result.upgrade,
                result.machine,
            )

            logging.info("metric: {}".format(metric))
