# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.


import logging
import re

from django.core.management import call_command

from performance.management import (
    get_image_build_number,
    get_run_number,
)

from common.management import (
    jenkins_get,
    JenkinsBaseCommand,
)

from memory.models import (
    MemoryBuild,
    MemoryDetail,
    MemoryImage,
    MemoryMachine,
    MemoryMetric,
    MemoryLog,
    MemoryResult,
    MemoryUpgrade,
)


class Command(JenkinsBaseCommand):

    job_regex = re.compile(
        ur"smem-([^-]+)-([^-]+)-([^-]+)-(install|upgrade)-([^-]+)-(.*)",
    )
    history_days = 30

    def _get_memory_logs(self, artifacts):
        """ Check for valid memory results.

        :returns: a list of memory logs.

        """

        logs = {}
        smem_log_regex = re.compile(ur'^results\/(\d*)\/smem.log')
        log_regex = re.compile(ur'^results\/(\d*)\/(smem.log|meminfo)')
        utah_log_regex = re.compile(ur'^(clientlogs\/|)utah.*.yaml')

        for artifact in artifacts:
            path = artifact['relativePath']
            num = get_run_number(path, prefix="results\/")

            if utah_log_regex.match(path):
                logging.debug("found %s", path)
                logs['utah'] = artifact

            if num is None:
                continue

            if num not in logs:
                logs[num] = {
                    'data_file': None,
                    'logs': [],
                }

            logging.debug("artifact: %s", path)
            if smem_log_regex.match(path):
                logging.debug("found %s", path)
                logs[num]['data_file'] = artifact
            if log_regex.match(path):
                logging.debug("found memory log %s", path)
                logs[num]['logs'].append(artifact)

        return logs

    def _parse_log(self, dashboard_data):
        """ Parse an smem.log file. """

        logging.debug("dashboard_data: %s", dashboard_data)

        lines = dashboard_data.split('\n')

        items = []

        count = 0
        for line in lines:
            count += 1

            # skip leading header line.
            if count == 1:
                continue

            if line == '':
                continue

            logging.debug("line: %s", line)
            data = line.split('\t')
            logging.debug("data: %s", data)
            items.append(
                dict(
                    name=data[0],
                    command=data[1],
                    uss=int(data[2]),
                    pss=int(data[3]),
                    rss=int(data[4]),
                    vss=int(data[5]),
                ),
            )

        return items

    def add_result(self, dashboard_data):
        """ Add memory data to the database. """

        logging.info("Adding memory result.")

        items = self._parse_log(dashboard_data['output'])
        logging.debug("items: %s", items)

        if 'build_number' in dashboard_data:
            image, new_image = MemoryImage.objects.get_or_create(
                release=dashboard_data['release'],
                variant=dashboard_data['variant'],
                arch=dashboard_data['arch'],
                build_number=dashboard_data['build_number'],
                # fake an md5 since we don't have the data
                md5="{}{}{}{}".format(
                    self.arch,
                    self.release,
                    self.variant,
                    dashboard_data['build_number'],
                )
            )
            self.image = image
            self.upgrade = None
        else:  # upgrade
            upg, new_upgr = MemoryUpgrade.objects.get_or_create(
                release=dashboard_data['release'],
                variant=dashboard_data['variant'],
                arch=dashboard_data['arch'],
                date=dashboard_data['date'],
                from_release=dashboard_data['from_release'],
            )
            self.upgrade = upg
            self.image = None

        log, new_log = MemoryLog.objects.get_or_create(
            path=dashboard_data['log_path_url'],
            name=dashboard_data['log_name'],
        )

        build, new_build = MemoryBuild.objects.get_or_create(
            image=self.image,
            upgrade=self.upgrade,
            machine=self.machine,
            build=dashboard_data['jenkins_build']
        )

        # Add PSS data
        metric, new_metric = MemoryMetric.objects.get_or_create(
            image=self.image,
            upgrade=self.upgrade,
            machine=self.machine,
            name='pss',
            kernel=dashboard_data['kernel'],
            defaults=dict(
                ran_at=dashboard_data['build_date'],
                publish=False,
            ),
        )

        result, new_result = MemoryResult.objects.get_or_create(
            image=self.image,
            upgrade=self.upgrade,
            machine=self.machine,
            ran_at=dashboard_data['build_date'],
            kernel=dashboard_data['kernel'],
            metric=metric,
            name=metric.name,
            log=log,
            defaults=dict(
                publish=False,
                value=0,
            )
        )

        for item in items:
            logging.debug("item: %s", item)
            logging.debug("item['uss']: %s", item['uss'])

            md, new_md = MemoryDetail.objects.get_or_create(
                image=self.image,
                upgrade=self.upgrade,
                machine=self.machine,
                name=item['name'],
                command=item['command'],
                uss=item['uss'],
                pss=item['pss'],
                rss=item['rss'],
                vss=item['vss'],
                ran_at=dashboard_data['build_date'],
                result=result,
            )

    def remove_result(self, dashboard_data):
        """ Remove memory data from the database. """

        logging.info("Removing memory result.")

        items = self._parse_log(dashboard_data['output'])

        image = None
        upgrade = None
        if self.method == 'install':
            images = MemoryImage.objects.filter(
                release=dashboard_data['release'],
                variant=dashboard_data['variant'],
                arch=dashboard_data['arch'],
                build_number=dashboard_data['build_number'],
            )

            if images.count() > 0:
                image = images[0]

        else:  # upgrade
            upgrades = MemoryUpgrade.objects.filter(
                release=dashboard_data['release'],
                variant=dashboard_data['variant'],
                arch=dashboard_data['arch'],
                date=dashboard_data['date'],
                from_release=dashboard_data['from_release'],
            )

            if upgrades.count() > 0:
                upgrade = upgrades[0]

        # XXX: here we should be able to just delete the MemoryResult
        # and have it delete all it's details.

        for item in items:
            logging.debug("item: %s", item)
            logging.debug("item['uss']: %s", item['uss'])

            details = MemoryDetail.objects.filter(
                image=image,
                upgrade=upgrade,
                machine=self.machine,
                name=item['name'],
                ran_at=dashboard_data['build_date'],
            )

            details.delete()

            if (
                image is not None and
                image.memorydetail_set.count() == 0 and
                image.memoryresult_set.count() == 0 and
                image.memorymetric_set.count() == 0
            ):
                image.delete()

            if (
                upgrade is not None and
                upgrade.memorydetail_set.count() == 0 and
                upgrade.memoryresult_set.count() == 0 and
                upgrade.memorymetric_set.count() == 0
            ):
                upgrade.delete()

    def _process_logs(self, build, build_date, install_data, func):
        """ Process memory logs calling func() for each log. """

        if not hasattr(func, '__call__'):
            raise Exception("invalid callable: {}".format(func))

        unzip_path = self.get_artifacts(build)
        install_data['unzip_path'] = unzip_path

        logs = self._get_memory_logs(build['artifacts'])
        logging.info("logs: %s", logs)

        image_build_number = get_image_build_number(build)

        utah_log = logs.pop('utah', None)

        kernel = self.get_kernel(build['url'], utah_log)

        unzip_path = '{}artifact'.format(build['url'])
        online_path = unzip_path

        if 'unzip_path' in install_data:
            unzip_path = "{}/archive".format(install_data['unzip_path'])

        for num in logs:
            # skip incomplete logs
            if logs[num]['data_file'] is None:
                logging.warn(
                    "missing data: %s",
                    logs[num]['data_file'],
                )
                continue
            log_path = "{}/{}".format(
                unzip_path,
                logs[num]['data_file']['relativePath'],
            )
            log_path_url = "{}/{}".format(
                online_path,
                logs[num]['data_file']['relativePath'],
            )
            log_name = logs[num]['data_file']['relativePath']

            install_type = install_data['method']
            dashboard_data = dict(
                release=install_data['release'],
                variant=install_data['variant'],
                arch=install_data['arch'],
                build_date=build_date,
                kernel=kernel,
            )
            if 'jenkins_build' in install_data:
                dashboard_data['jenkins_build'] = install_data['jenkins_build']

            if install_type == 'install':
                dashboard_data['build_number'] = image_build_number
            else:
                dashboard_data['date'] = build_date
                # XXX: if we start having builds that upgrade across
                # releases we will need to update this
                dashboard_data['from_release'] = install_data['release']

            # Assume we need to retrieve the logs if there's no 'unzip_path'
            if 'unzip_path' not in install_data:
                logging.warn("Getting logs from jenkins directly...")
                dashboard_data['output'] = jenkins_get(
                    log_path_url,
                    as_json=False,
                )
            else:
                dashboard_data['output'] = self.get_local_file(log_path)

            dashboard_data['log_path'] = log_path
            dashboard_data['log_path_url'] = log_path_url
            dashboard_data['log_name'] = log_name

            func(dashboard_data)

    def extract_data(self, name):
        m = self.job_regex.match(name)

        if m:
            self.release = m.group(1)
            self.variant = m.group(2)
            self.arch = m.group(3)
            self.method = m.group(4)
            self.workload = m.group(5)
            self.machine_name = m.group(6)

            logging.info("name: %s", name)
            logging.info("method: %s", self.method)

            self.install_data = dict(
                release=self.release,
                variant=self.variant,
                arch=self.arch,
                method=self.method,
            )

    def process_job(self, job):
        self.machine, new_machine = MemoryMachine.objects.get_or_create(
            name=self.machine_name,
        )

    def post_jobs(self):
        """ Do post job aggregations. """

        call_command('agg_processes')
        call_command('agg_results')
        call_command('agg_metrics')
