# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.


from django.db import models

from common.models import (
    DashboardBaseModel,
    JenkinsBuild,
)

from performance.models import (
    ImageBase,
    MachineBase,
    MetricBase,
    ResultBase,
    UpgradeBase,
)


class InvalidInstance(Exception):
    pass


class MemoryUpgrade(UpgradeBase):
    """ Memory upgrade model. """

    class Meta:
        db_table = "memory_upgrades"

    @property
    def metric_count(self):
        return self.memorymetric_set.count()

    @property
    def metrics(self):
        return self.memorymetric_set.all()


class MemoryImage(ImageBase):
    """ Memory Image model. """

    class Meta:
        db_table = "memory_images"

    @property
    def metric_count(self):
        return self.memorymetric_set.count()

    @property
    def metrics(self):
        return self.memorymetric_set.all()


class MemoryMachine(MachineBase):
    """ Memory Machine model. """

    class Meta:
        db_table = "memory_machines"

    @property
    def metric_count(self):
        return self.memorymetric_set.count()

    @property
    def metrics(self):
        return self.memorymetric_set.all()

    @property
    def has_image(self):
        images = self.metrics.values('image')
        images = images.filter(upgrade=None).distinct()

        return images.count() > 0

    @property
    def has_upgrade(self):
        upgrades = self.metrics.values('upgrade')
        upgrades = upgrades.filter(image=None).distinct()

        return upgrades.count() > 0


class MemoryLog(DashboardBaseModel):
    """ Memory logs for a build. """

    class Meta:
        db_table = "memory_logs"

    path = models.CharField(max_length=512)
    name = models.CharField(max_length=100)


class MemoryDetail(DashboardBaseModel):
    """ Memory raw data. """

    class Meta:
        db_table = "memory_details"

    image = models.ForeignKey(MemoryImage, null=True)
    upgrade = models.ForeignKey(MemoryUpgrade, null=True)
    machine = models.ForeignKey(MemoryMachine)
    name = models.CharField(max_length=1024)
    command = models.CharField(max_length=4096)
    uss = models.PositiveIntegerField()
    pss = models.PositiveIntegerField()
    rss = models.PositiveIntegerField()
    vss = models.PositiveIntegerField()
    ran_at = models.DateTimeField("date run")
    result = models.ForeignKey('MemoryResult')

    def __unicode__(self):
        return "{}:{}:{}:{}:{}:{}".format(
            self.name,
            self.command,
            self.uss,
            self.pss,
            self.rss,
            self.vss,
        )

    def save(self, *args, **kwargs):
        """
        Overridden to update MemoryResult information.
        """

        if self.image is not None and self.upgrade is not None:
            raise InvalidInstance("MemoryDetail can't have image and upgrade")

        if self.image is None and self.upgrade is None:
            raise InvalidInstance(
                "MemoryDetail must have an image or an upgrade"
            )

        super(MemoryDetail, self).save(*args, **kwargs)

    def _clean_logs(self):
        """
        Clean up logs without MemoryResults.
        """

        MemoryLog.objects.annotate(models.Count('memoryresult')).filter(
            memoryresult__count=0).delete()

    @property
    def difference(self):

        process = MemoryProcess.objects.get(
            machine=self.machine,
            image=self.image,
            upgrade=self.upgrade,
            process=self.name,
        )

        return process.difference


class MemoryResult(ResultBase):
    """ Memory result model. """

    class Meta:
        db_table = "memory_results"

    image = models.ForeignKey(MemoryImage, null=True)
    upgrade = models.ForeignKey(MemoryUpgrade, null=True)
    machine = models.ForeignKey(MemoryMachine)
    kernel = models.CharField(max_length=100, null=True)
    metric = models.ForeignKey('MemoryMetric')
    log = models.ForeignKey('MemoryLog')

    def save(self, *args, **kwargs):
        """
        Overridden to update MemoryMetric information.
        """

        if self.image is not None and self.upgrade is not None:
            raise InvalidInstance("MemoryResult can't have image and upgrade")

        super(MemoryResult, self).save(*args, **kwargs)

    @property
    def log_link(self):
        """ Get links to logs for this result. """

        name = self.log.name
        name = name.split('/')[2]

        link = '<a href="{}" style="display: block;">{}</a>'.format(
            self.log.path,
            name,
        )

        return link

    @property
    def detail_count(self):
        return self.memorydetail_set.count()


class MemoryMetric(MetricBase):
    """ Memory metric model. """

    class Meta:
        db_table = "memory_metrics"

    image = models.ForeignKey(MemoryImage, null=True)
    upgrade = models.ForeignKey(MemoryUpgrade, null=True)
    machine = models.ForeignKey(MemoryMachine)
    kernel = models.CharField(max_length=100, null=True)

    def save(self, *args, **kwargs):
        """
        Overridden to prevent image and upgrade instances.
        """

        if self.image is not None and self.upgrade is not None:
            raise InvalidInstance("MemoryMetric can't have image and upgrade")

        super(MemoryMetric, self).save(*args, **kwargs)

    @property
    def log_links(self):
        """ Get links to logs for this result. """

        links = ""
        logs = MemoryLog.objects.filter(
            memoryresult__image=self.image,
            memoryresult__upgrade=self.upgrade,
            memoryresult__machine=self.machine,
        ).distinct()

        for log in logs:
            name = log.name
            name = name.split('/')[2]

            links = " ".join([
                links,
                '<a href="{}" style="display: block;">{}</a>'.format(
                    log.path,
                    name,
                )
            ])

        return links

    @property
    def logs(self):
        return MemoryLog.objects.filter(
            memoryresult__image=self.image,
            memoryresult__upgrade=self.upgrade,
            memoryresult__machine=self.machine,
        ).distinct()

    @property
    def log_count(self):
        return self.logs.count()


class MemoryBuild(DashboardBaseModel):
    """ Jenkins Build image/upgrade. """

    class Meta:
        db_table = "memory_builds"

    image = models.ForeignKey(MemoryImage, null=True)
    upgrade = models.ForeignKey(MemoryUpgrade, null=True)
    machine = models.ForeignKey(MemoryMachine)
    build = models.ForeignKey(JenkinsBuild)

    @property
    def job_bugs(self):
        """ Get a list of job bugs for this image/upgrade. """

        return self.build.bugs

    @property
    def job_bug_count(self):
        """ Get a count of job bugs for this image/upgrade. """
        return self.job_bugs.count()


class MemoryProcess(MetricBase):
    """ Per-process metrics. """

    class Meta:
        db_table = "memory_processes"

    image = models.ForeignKey(MemoryImage, null=True)
    upgrade = models.ForeignKey(MemoryUpgrade, null=True)
    machine = models.ForeignKey(MemoryMachine)
    command = models.CharField(max_length=4096)
    process = models.CharField(max_length=1024)

    @property
    def count(self):
        """ Get number of times this process showed up in the logs. """
        return MemoryDetail.objects.filter(
            machine=self.machine,
            image=self.image,
            upgrade=self.upgrade,
            name=self.process,
        ).count()

    @property
    def previous(self):
        """ The previous process data. """

        prev = MemoryProcess.objects.filter(
            machine=self.machine,
            name=self.name,
            process=self.process,
            command=self.command,
        ).exclude(id=self.id)

        if self.upgrade is not None:
            prev = prev.filter(upgrade__date__lte=self.upgrade.date)
            prev = prev.exclude(upgrade=None).order_by('-upgrade__date')
        else:
            prev = prev.filter(
                image__build_number__lte=self.image.build_number,
            )
            prev = prev.exclude(image=None).order_by('-image__build_number')

        if prev.count() == 0:
            return None

        prev = prev[0]

        return prev

    @property
    def difference(self):
        """ The difference between this image/upgrade and the one before. """

        prev = self.previous

        if prev is None:
            return 0

        return self.average - prev.average
