# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import datetime
import json

from django.core.management import call_command
from django.db.models import Q
from django.test import TestCase
from django.test.client import Client

from memory.models import (
    InvalidInstance,
    MemoryDetail,
    MemoryImage,
    MemoryLog,
    MemoryMachine,
    MemoryMetric,
    MemoryProcess,
    MemoryResult,
    MemoryUpgrade,
)


def _setUp(self):
    """ Setup commonly used objects. """

    self.release = 'raring'
    self.variant = 'desktop'
    self.arch = 'amd64'
    self.build_number = '20130201'
    self.md5 = 'blah'

    self.image = MemoryImage.objects.create(
        release=self.release,
        variant=self.variant,
        arch=self.arch,
        build_number=self.build_number,
        md5=self.md5,
    )

    self.assert_(self.image)

    self.upgrade_data = {
        'release': 'raring',
        'from_release': 'precise',
        'date': datetime.datetime.now(),
        'variant': 'desktop',
        'arch': 'amd64',
    }

    self.upgrade = MemoryUpgrade.objects.create(
        release=self.upgrade_data['release'],
        from_release=self.upgrade_data['from_release'],
        arch=self.upgrade_data['arch'],
        variant=self.upgrade_data['variant'],
        date=self.upgrade_data['date'],
    )

    self.assert_(self.upgrade)

    self.machine_name = 'memory-test-1'
    self.machine_mac_address = '00:00:00:00:01'
    self.machine = MemoryMachine.objects.create(
        name=self.machine_name,
        mac_address=self.machine_mac_address,
    )

    self.assert_(self.machine)

    self.metric_name = "pss"

    self.metric = MemoryMetric.objects.create(
        image=self.image,
        machine=self.machine,
        ran_at=datetime.datetime.now(),
        name=self.metric_name,
    )

    self.assert_(self.metric)

    self.log = MemoryLog.objects.create(
        name="blah",
        path="http://jenkins.qa.ubuntu.com/blah",
    )

    self.assert_(self.log)

    self.result = MemoryResult.objects.create(
        image=self.image,
        machine=self.machine,
        value=0,
        ran_at=datetime.datetime.now(),
        name=self.metric.name,
        metric=self.metric,
        log=self.log
    )

    self.assert_(self.result)

    self.detail_uss = 100
    self.detail_pss = 150
    self.detail_rss = 200
    self.detail_vss = 50
    self.detail = MemoryDetail.objects.create(
        image=self.image,
        machine=self.machine,
        uss=self.detail_uss,
        pss=self.detail_pss,
        rss=self.detail_rss,
        vss=self.detail_vss,
        ran_at=datetime.datetime.now(),
        result=self.result,
    )

    self.assert_(self.detail)

    self.metric2 = MemoryMetric.objects.create(
        upgrade=self.upgrade,
        machine=self.machine,
        name=self.metric_name,
        ran_at=datetime.datetime.now(),
    )

    self.assert_(self.metric2)

    self.log2 = MemoryLog.objects.create(
        name="blah2",
        path="http://jenkins.qa.ubuntu.com/blah2",
    )

    self.assert_(self.log2)

    self.result2 = MemoryResult.objects.create(
        upgrade=self.upgrade,
        machine=self.machine,
        value=0,
        ran_at=datetime.datetime.now(),
        name=self.metric2.name,
        metric=self.metric2,
        log=self.log2,
    )

    self.assert_(self.result2)

    self.detail2 = MemoryDetail.objects.create(
        upgrade=self.upgrade,
        machine=self.machine,
        uss=self.detail_uss,
        pss=self.detail_pss,
        rss=self.detail_rss,
        vss=self.detail_vss,
        ran_at=datetime.datetime.now(),
        result=self.result2,
    )

    self.assert_(self.detail2)

    call_command('agg_results')
    call_command('agg_metrics')
    call_command('agg_processes')

    # get updated instances from the DB
    self.result = MemoryResult.objects.get(id=self.result.id)
    self.result2 = MemoryResult.objects.get(id=self.result2.id)
    self.metric = MemoryMetric.objects.get(id=self.metric.id)
    self.metric2 = MemoryMetric.objects.get(id=self.metric2.id)


class TestMemory(TestCase):
    """ Tests for memory models. """

    def setUp(self):
        """ Common setup. """

        _setUp(self)

    def test_machine_has_image(self):
        """ Test that a machine has image metrics. """

        self.assertTrue(self.machine.has_image)

    def test_machine_has_image_none(self):
        """ Test has_image for no metrics. """

        self.image.metrics.delete()
        self.assertFalse(self.machine.has_image)

    def test_machine_has_upgrade_none(self):
        """ Test has_upgrade for no metrics. """

        self.upgrade.metrics.delete()
        self.assertFalse(self.machine.has_upgrade)

    def test_machine_has_upgrade(self):
        """ Test that a machine has upgrade metrics. """

        self.assertTrue(self.machine.has_upgrade)

    def test_detail(self):
        """ Test the detail model. """

        self.assertEqual(self.detail_uss, self.detail.uss)
        self.assertEqual(self.detail_pss, self.detail.pss)
        self.assertEqual(self.detail_rss, self.detail.rss)
        self.assertEqual(self.detail_vss, self.detail.vss)

    def test_image(self):
        """ Test the image model. """

        self.assertEqual(self.image.release, self.release)
        self.assertEqual(self.image.variant, self.variant)
        self.assertEqual(self.image.arch, self.arch)
        self.assertEqual(self.image.build_number, self.build_number)
        self.assertEqual(self.image.md5, self.md5)

    def test_upgrade(self):
        """ Test the upgrade model. """

        self.assertEqual(
            self.upgrade_data['release'],
            self.upgrade.release,
        )
        self.assertEqual(
            self.upgrade_data['from_release'],
            self.upgrade.from_release,
        )
        self.assertEqual(self.upgrade_data['arch'], self.upgrade.arch)
        self.assertEqual(self.upgrade_data['variant'], self.upgrade.variant)
        self.assertEqual(self.upgrade_data['date'], self.upgrade.date)

    def test_machine(self):
        """ Test the machine model. """

        self.assertEqual(self.machine.name, self.machine_name)
        self.assertEqual(self.machine.mac_address, self.machine_mac_address)

    def test_image_metric(self):
        """ Test that a metric has valid values. """

        metric = self.result.metric

        self.assertEqual(metric.average, self.result.value)
        self.assertEqual(metric.minimum, self.result.value)
        self.assertEqual(metric.maximum, self.result.value)
        self.assertEqual(metric.stddev, 0)
        self.assertEqual(metric.name, self.result.name)

    def test_upgrade_metric(self):
        """ Test that an upgrade metric has valid values. """

        metrics = MemoryMetric.objects.filter(
            upgrade=self.upgrade,
            machine=self.machine,
            name="pss",
        )

        self.assertEqual(metrics.count(), 1)
        metric = metrics[0]

        self.assertEqual(metric.average, self.detail2.pss)
        self.assertEqual(metric.minimum, self.detail2.pss)
        self.assertEqual(metric.maximum, self.detail2.pss)
        self.assertEqual(metric.stddev, 0)
        self.assertEqual(metric.name, "pss")

    def test_metric_logs(self):
        """ Test that metric log counts are correct. """

        self.assertEqual(self.metric.log_count, 1)

    def test_remove_result(self):
        """ Test that removing a result adjusts metrics. """

        result_name = self.result.name

        metric_count = MemoryMetric.objects.filter(
            image=self.image,
            machine=self.machine,
            name=result_name,
        ).count()

        self.result.delete()

        call_command("agg_metrics")

        metrics = MemoryMetric.objects.filter(
            image=self.image,
            machine=self.machine,
            name=result_name,
        )

        self.assertEqual(metrics.count(), metric_count - 1)

    def test_remove_result_all(self):
        """ Test that removing a result queryset adjusts metrics. """

        result_name = self.result.name

        MemoryResult.objects.filter(
            image=self.image,
            machine=self.machine,
            name=result_name,
        ).delete()

        call_command("agg_metrics")

        metrics = MemoryMetric.objects.filter(
            image=self.image,
            machine=self.machine,
            name=result_name,
        )

        self.assertEqual(metrics.count(), 0)

    def test_have_proccesses(self):
        """ Test that agg_processes created some processes. """

        processes = MemoryProcess.objects.all()

        self.assertGreater(processes.count(), 0)

    def test_process_count(self):
        """ Test that a process has at least one detail. """

        process = MemoryProcess.objects.all()[0]

        self.assertGreater(process.count, 0)

    def test_process_difference(self):
        """ Test that a process returns a correct difference. """

        release = "raring"
        variant = "desktop"
        arch = "amd64"

        image1 = MemoryImage.objects.create(
            release=release,
            variant=variant,
            arch=arch,
            build_number="20120314",
            md5="aaa",
        )

        image2 = MemoryImage.objects.create(
            release=release,
            variant=variant,
            arch=arch,
            build_number="20120315",
            md5="bbb",
        )

        p1 = MemoryProcess.objects.create(
            machine=self.machine,
            image=image1,
            process="test",
            average=100,
            ran_at=datetime.datetime.now(),
        )

        p2 = MemoryProcess.objects.create(
            machine=self.machine,
            image=image2,
            process="test",
            average=200,
            ran_at=datetime.datetime.now(),
        )

        difference = p2.difference

        if p1.average > p2.average:
            self.assertLess(difference, 0)
        elif p1.average < p2.average:
            self.assertGreater(difference, 0)
        else:
            self.assertEqual(difference, 0)


class TestAPI(TestCase):
    """ Test Memory API. """

    def setUp(self):
        """ Common setup. """

        _setUp(self)
        self.url_prefix = '/api/memory/'

    def url(self, part):
        """ URL creator. """

        return "{}{}".format(
            self.url_prefix,
            part,
        )

    def test_images(self):
        """ Test images api. """

        c = Client()
        url = self.url('images/')
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        self.assertEqual(len(data), 1)

    def test_image_detail(self):
        """ Test image_detail api. """

        c = Client()
        url = self.url(
            'image/{}/'.format(
                self.image.id,
            )
        )
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        self.assertEqual(len(data), 1)
        data = data[0]

        self.assertEqual(data['release'], self.image.release)
        self.assertEqual(data['variant'], self.image.variant)
        self.assertEqual(data['arch'], self.image.arch)
        self.assertEqual(data['build_number'], self.image.build_number)

    def test_upgrades(self):
        """ Test upgrades api. """

        c = Client()
        url = self.url('upgrades/')
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        self.assertEqual(len(data), 1)

    def test_upgrade_detail(self):
        """ Test upgrade_detail api. """

        c = Client()
        url = self.url(
            'upgrade/{}/'.format(
                self.upgrade.id,
            )
        )
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        self.assertEqual(len(data), 1)
        data = data[0]

        self.assertEqual(data['release'], self.upgrade.release)
        self.assertEqual(data['variant'], self.upgrade.variant)
        self.assertEqual(data['arch'], self.upgrade.arch)

        # Due to differences between the json decoded date microseconds
        # convert both dates to strings without microseconds.
        json_date = datetime.datetime.strptime(
            data['date'],
            '%Y-%m-%dT%H:%M:%S.%f'
        )
        date_cmp_format = "%Y-%m-%d %H:%H:S"

        self.assertEqual(
            self.upgrade_data['date'].strftime(date_cmp_format),
            json_date.strftime(date_cmp_format),
        )

        self.assertEqual(data['from_release'], self.upgrade.from_release)

    def test_machines(self):
        """ Test machines api. """

        c = Client()
        url = self.url('machines/')
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        self.assertEqual(len(data), 1)

    def test_machine_detail(self):
        """ Test machine_detail api. """

        c = Client()
        url = self.url(
            'machine/{}/'.format(
                self.machine.id,
            )
        )
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        data = data[0]

        self.assertEqual(data['name'], self.machine.name)
        self.assertEqual(data['mac_address'], self.machine.mac_address)

    def test_image_machine_results(self):
        """ Test image_machine_results api. """

        c = Client()
        url = self.url(
            'image/{}/machine/{}/results/'.format(
                self.image.id,
                self.machine.id,
            )
        )
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        self.assertEqual(len(data), 1)

    def test_upgrade_machine_results(self):
        """ Test upgrade_machine_results api. """

        c = Client()
        url = self.url(
            'upgrade/{}/machine/{}/results/'.format(
                self.upgrade.id,
                self.machine.id,
            )
        )
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        self.assertEqual(len(data), 1)

    def test_result_detail(self):
        """ Test result_detail api. """

        c = Client()
        url = self.url(
            'result/{}/'.format(
                self.result.id,
            )
        )
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        self.assertEqual(len(data), 1)

        data = data[0]
        self.assertEqual(data['name'], self.result.name)
        self.assertEqual(data['value'], self.result.value)

    def test_image_machine_metrics(self):
        """ Test image_machine_metrics api. """

        c = Client()
        url = self.url(
            'image/{}/machine/{}/metrics/'.format(
                self.image.id,
                self.machine.id,
            )
        )
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        self.assertEqual(len(data), 1)

    def test_upgrade_machine_metrics(self):
        """ Test upgrade_machine_metrics api. """

        c = Client()
        url = self.url(
            'upgrade/{}/machine/{}/metrics/'.format(
                self.upgrade.id,
                self.machine.id,
            )
        )
        response = c.get(url)

        self.assertEqual(response.status_code, 200)

        data = response.content
        data = json.loads(data)

        self.assertEqual(len(data), 1)

    def test_image_upgrade_result(self):
        """ Test that no result exist with image and upgrade. """

        self.assertEqual(
            MemoryResult.objects.filter(
                ~Q(image=None) & ~Q(upgrade=None),
            ).count(),
            0,
        )

    def test_image_upgrade_metric(self):
        """ Test that no metric exist with image and upgrade. """

        self.assertEqual(
            MemoryMetric.objects.filter(
                ~Q(image=None) & ~Q(upgrade=None),
            ).count(),
            0,
        )

    def test_prevent_adding_image_upgrade_detail(self):
        """ Test detail can't have image and upgrade. """

        with self.assertRaises(InvalidInstance):
            MemoryDetail.objects.create(
                image=self.image,
                upgrade=self.upgrade,
                machine=self.machine,
                uss=1,
                pss=2,
                rss=3,
                vss=4,
                ran_at=datetime.datetime.now(),
            )

    def test_prevent_adding_image_upgrade_result(self):
        """ Test result can't have image and upgrade. """

        with self.assertRaises(InvalidInstance):
            MemoryResult.objects.create(
                image=self.image,
                upgrade=self.upgrade,
                machine=self.machine,
                name="pss",
                value=1,
                ran_at=datetime.datetime.now(),
            )

    def test_prevent_adding_image_upgrade_metric(self):
        """ Test metric can't have image and upgrade. """

        with self.assertRaises(InvalidInstance):
            MemoryMetric.objects.create(
                image=self.image,
                upgrade=self.upgrade,
                machine=self.machine,
                name="pss",
                average=1,
                minimum=1,
                maximum=1,
                stddev=0.01,
            )
