# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.core.urlresolvers import reverse
from django.db.models import Q
from django.db import models
from django.template import RequestContext
from django.views.decorators.http import require_GET

from django_tables2 import RequestConfig

from django.shortcuts import (
    render_to_response,
    get_object_or_404,
    redirect,
)

from memory.models import (
    MemoryMachine,
    MemoryMetric,
    MemoryProcess,
    MemoryResult,
)

from memory.tables import (
    ImageResultTable,
    DetailTable,
    DetailTotalTable,
    UpgradeResultTable,
)

#from power.tables import MetricTable


def _get_memory_arches():
    arches = []

    image_arches = MemoryMetric.objects.values('image__arch')
    image_arches = image_arches.distinct().exclude(publish=False)
    image_arches = image_arches.exclude(image__arch=None)
    image_arches = image_arches.order_by('-image__arch')

    for arch in image_arches:
        arches.append(arch['image__arch'])

    upgrade_arches = MemoryMetric.objects.values('upgrade__arch')
    upgrade_arches = upgrade_arches.distinct().exclude(publish=False)
    upgrade_arches = upgrade_arches.exclude(upgrade__arch=None)
    upgrade_arches = upgrade_arches.order_by('-upgrade__arch')

    for arch in upgrade_arches:
        if arch['upgrade__arch'] not in arches:
            arches.append(arch['upgrade__arch'])
    return arches


def _get_metrics(arch=None):
    mmetrics = MemoryMetric.objects.all()

    if arch is not None:
        mmetrics = mmetrics.filter(image__arch=arch)

    metrics = []
    for metric in mmetrics.values('name').distinct().exclude(publish=False):
        metrics.append(metric['name'])

    return metrics


@require_GET
def arch_overview(request, arch=None):
    arches = _get_memory_arches()

    chart_size = request.GET.get('size', '')

    if arch is None:
        return redirect(
            'arch/{}'.format(arches[0] if len(arches) > 0 else "amd64")
        )

    machines = MemoryMachine.objects.filter(
        Q(memorymetric__image__arch=arch) |
        Q(memorymetric__upgrade__arch=arch)
    ).distinct()

    image_types = MemoryMetric.objects.filter(
        publish=True
    ).exclude(
        image__variant=None, upgrade__variant=None
    ).values(
        'machine__name', 'image__variant', 'upgrade__variant',
    ).annotate()
    data = {
        'machines': machines,
        'arch': arch,
        'arches': arches,
        'active': arch,
        'chart_size': chart_size,
        'image_types': image_types,
    }

    return render_to_response('memory/arch_overview.html', data,
                              RequestContext(request))


@require_GET
def _process_overview(request, arch, machine_id, upgrade=False):
    arches = _get_memory_arches()

    machine = get_object_or_404(MemoryMachine, id=machine_id)
    chart_size = request.GET.get('size', '')
    process = request.GET.get('process')

    processes = MemoryProcess.objects.filter(
        machine=machine,
        publish=True,
    ).values('process').distinct()

    if upgrade:
        processes = processes.filter(upgrade__arch=arch)
    else:
        processes = processes.filter(image__arch=arch)

    if process is not None:
        processes = processes.filter(process=process)

    data = {
        'machine': machine,
        'processes': processes,
        'arch': arch,
        'arches': arches,
        'active': arch,
        'chart_size': chart_size,
        'upgrade': upgrade,
        'type': "upgrade" if upgrade else "image",
    }

    return render_to_response('memory/process_overview.html', data,
                              RequestContext(request))


@require_GET
def upgrade_process_overview(request, arch, machine_id):
    return _process_overview(request, arch, machine_id, upgrade=True)


@require_GET
def image_process_overview(request, arch, machine_id):
    return _process_overview(request, arch, machine_id, upgrade=False)


@require_GET
def machine_image_results(request, arch, machine_id, variant, metric):
    machine = get_object_or_404(MemoryMachine, id=machine_id)

    results = MemoryResult.objects.filter(publish=True)
    results = results.filter(
        machine=machine,
        image__arch=arch,
        image__variant=variant,
        name=metric,
    )

    table = ImageResultTable(results)
    RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'machine': machine,
        'arch': arch,
        'table': table,
        'variant': variant,
    }

    return render_to_response('memory/machine_results.html', data,
                              RequestContext(request))


@require_GET
def machine_upgrade_results(request, arch, machine_id, variant, metric):
    machine = get_object_or_404(MemoryMachine, id=machine_id)

    results = MemoryResult.objects.filter(publish=True)
    results = results.filter(
        machine=machine,
        upgrade__variant=variant,
        upgrade__arch=arch,
        name=metric,
    )

    table = UpgradeResultTable(results)
    RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'machine': machine,
        'arch': arch,
        'table': table,
        'variant': variant,
    }

    return render_to_response('memory/machine_results.html', data,
                              RequestContext(request))


@require_GET
def machine_image_details(request, arch, machine_id, result_id):
    result = get_object_or_404(MemoryResult, id=result_id)

    details = result.memorydetail_set.all()

    data = []
    for detail in details:
        movement = "movement"

        if detail.difference > 0:
            movement = "movement-up"
        elif detail.difference < 0:
            movement = "movement-down"

        url = reverse(
            'api_memory_image_machine_arch_processes',
            args=(machine_id, arch, detail.name),
        )

        data.append(
            dict(
                id=detail.id,
                machine=detail.machine,
                image=detail.image,
                image__build_number=detail.image.build_number,
                image__arch=detail.image.arch,
                name=detail.name,
                command=detail.command,
                uss=detail.uss,
                pss=detail.pss,
                rss=detail.rss,
                vss=detail.vss,
                url=url,
                movement=movement,
                change=detail.difference,
            )
        )

    table = DetailTable(data)
    RequestConfig(request, paginate=False).configure(table)

    detail_agg = details.aggregate(
        models.Sum('uss'),
        models.Sum('pss'),
        models.Sum('rss'),
        models.Sum('vss'),
    )

    totals_data = [
        dict(
            uss=detail_agg['uss__sum'],
            pss=detail_agg['pss__sum'],
            rss=detail_agg['rss__sum'],
            vss=detail_agg['vss__sum'],
        )
    ]

    totals = DetailTotalTable(totals_data)
    #RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'result': result,
        'arch': arch,
        'table': table,
        'totals': totals,
    }

    return render_to_response('memory/machine_details.html', data,
                              RequestContext(request))


@require_GET
def machine_upgrade_details(request, arch, machine_id, result_id):
    result = get_object_or_404(MemoryResult, id=result_id)

    details = result.memorydetail_set.filter(publish=True)

    data = []
    for detail in details:
        movement = "movement"

        if detail.difference > 0:
            movement = "movement-up"
        elif detail.difference < 0:
            movement = "movement-down"

        url = reverse(
            'api_memory_upgrade_machine_arch_processes',
            args=(machine_id, arch, detail.name),
        )

        data.append(
            dict(
                id=detail.id,
                machine=detail.machine,
                upgrade=detail.upgrade,
                upgrade__date=detail.upgrade.date,
                name=detail.name,
                command=detail.command,
                uss=detail.uss,
                pss=detail.pss,
                rss=detail.rss,
                vss=detail.vss,
                url=url,
                movement=movement,
                change=detail.difference,
            )
        )

    table = DetailTable(data)
    RequestConfig(request, paginate=False).configure(table)

    detail_agg = details.aggregate(
        models.Sum('uss'),
        models.Sum('pss'),
        models.Sum('rss'),
        models.Sum('vss'),
    )

    totals_data = [
        dict(
            uss=detail_agg['uss__sum'],
            pss=detail_agg['pss__sum'],
            rss=detail_agg['rss__sum'],
            vss=detail_agg['vss__sum'],
        )
    ]

    totals = DetailTotalTable(totals_data)
    #RequestConfig(request, paginate={"per_page": 100}).configure(table)

    data = {
        'result': result,
        'arch': arch,
        'table': table,
        'totals': totals,
    }

    return render_to_response('memory/machine_details.html', data,
                              RequestContext(request))
