# QA Dashboard
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version
# 3, as published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.


import logging
import re

from common.management import (
    jenkins_get,
    JenkinsBaseCommand,
)

from common.utils import regexes

from performance.management import get_image_build_number

from power.models import (
    PowerImage,
    PowerLog,
    PowerMachine,
    PowerMetric,
)

regexes.update({
    'power_result': re.compile(
        ur'^qa-power-testing/results/.*power_consumption.json'
    ),
    'power-metrics': re.compile(
        ur'^(.*)_(current_drawn_mA|test_duration)_(.*){perf}'
    ),
})


def _get_power_logs(artifacts):
    """ Check for valid power results.

    :returns: a list of power logs.

    """
    json_regex = regexes['power_result']

    logs = []
    for artifact in artifacts:
        path = artifact['relativePath']
        logging.debug("path: {}".format(path))
        if json_regex.match(path):
            logging.info("artifact: %s", artifact['relativePath'])
            logs.append(artifact)

    return logs


def _valid_metric(metric):
    """ Check that a metric is valid. """
    keys = metric.keys()

    if 'minimum' not in keys:
        return False
    if 'maximum' not in keys:
        return False
    if 'average' not in keys:
        return False
    if 'stddev' not in keys:
        return False
    if 'seconds_average' not in keys:
        return False

    return True


class Command(JenkinsBaseCommand):
    """ The command. """

    job_regex = regexes['power']
    job_types = ['milestone', 'backfill']

    def extract_data(self, name):
        m = self.job_regex.match(name)

        if m:
            self.job_type = m.group(1)
            self.variant = m.group(2)
            self.arch = m.group(3)
            self.machine_name = m.group(4)

            self.install_data = dict(
                job_type=self.job_type,
                variant=self.variant,
                arch=self.arch,
                machine=self.machine_name,
            )

    def process_job(self, job):
        self.machine, new_machine = PowerMachine.objects.get_or_create(
            name=self.machine_name,
        )

        if new_machine:
            logging.debug("Added new machine: %s", self.machine_name)

    def _process_logs(self, build, build_date, install_data, func):

        if not hasattr(func, '__call__'):
            raise Exception("invalid callable: %s", func)

        logs = _get_power_logs(build['artifacts'])
        self.image_build_number = get_image_build_number(build)
        self.build_url = build['url']

        for log in logs:
            log_path = "{}artifact/{}".format(
                build['url'],
                log['relativePath'],
            )
            dashboard_data = jenkins_get(log_path)

            logging.info("dashboard_data: {}".format(dashboard_data))
            logging.info("metrics type: {}".format(type(
                dashboard_data['metrics'])))

            self.release = dashboard_data['meta']['distro-release']
            self.kernel = dashboard_data['meta']['kernel']

            func(build, dashboard_data)

    def add_result(self, build, dashboard_data):

        build_logs = []
        for artifact in build['artifacts']:
            log, new_log = PowerLog.objects.get_or_create(
                path="{}artifact/{}".format(
                    build['url'],
                    artifact['relativePath'],
                ),
                name=artifact['relativePath'],
            )

            if new_log:
                logging.info("Adding new log: {}".format(log))

            build_logs.append(log)

        self.image, new_image = PowerImage.objects.get_or_create(
            arch=self.arch,
            release=self.release,
            variant=self.variant,
            build_number=self.image_build_number,
            # fake an md5 since we don't have the data
            md5="{}{}{}{}".format(
                self.arch,
                self.release,
                self.variant,
                self.image_build_number,
            )
        )

        if new_image:
            logging.debug("Added new image: %s", self.image)

        metrics = {}
        for metric in dashboard_data['metrics'].keys():
            m = regexes['power-metrics'].match(metric)
            if m:
                metric_name = m.group(1)
                metric_units = m.group(2)
                metric_type = m.group(3)
                value = dashboard_data['metrics'][metric]

                logging.info(
                    "metric: %s %s %s: %s",
                    metric_name,
                    metric_units,
                    metric_type,
                    dashboard_data['metrics'][metric],
                )

                if metric_name not in metrics:
                    metrics[metric_name] = {}

                metrics[metric_name][metric_type] = value

        for metric_name in metrics.keys():
            if not _valid_metric(metrics[metric_name]):
                continue

            metric, new_metric = PowerMetric.objects.get_or_create(
                image=self.image,
                machine=self.machine,
                name=metric_name,
                ran_at=self.build_date,
                kernel=self.kernel,
                jenkins_url=self.build_url,
                minimum=metrics[metric_name]['minimum'],
                maximum=metrics[metric_name]['maximum'],
                average=metrics[metric_name]['average'],
                stddev=metrics[metric_name]['stddev'],
                duration=metrics[metric_name]['seconds_average'],
            )

            if new_metric:
                logging.info("Added new metric: %s", metric)

            for log in build_logs:
                metric.logs.add(log)

    def remove_result(self, build, dashboard_data):
        """ Remove power data from the database """
        logging.info("Removing power result")

        try:
            image = PowerImage.objects.get(
                arch=self.arch,
                release=self.release,
                variant=self.variant,
                build_number=self.image_build_number,
                md5="{}{}{}{}".format(
                    self.arch,
                    self.release,
                    self.variant,
                    self.image_build_number,
                )
            )
            metrics = PowerMetric.objects.filter(image=image)

            for metric in metrics:
                metric.logs.all().delete()
                metric.delete()
            image.delete()
        except PowerImage.DoesNotExist:
            pass
