# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import urllib2
import json
import re
import time
import logging
from datetime import datetime, timedelta
from StringIO import StringIO
from django.core.management.base import BaseCommand, CommandError

from optparse import make_option

from smoke.models import (
    Build,
    Result,
    ResultLog,
    Run,
)

from common.management import (
    jenkins_get,
    skip_build,
    BUILD_COUNT_MAX,
    JOB_COUNT_MAX,
    DEBUG_LIMITS,
)

from common.models import Bug

from smoke.utah_utils import process_smoke_log

#l = logging.getLogger('django.db.backends')
#l.setLevel(logging.DEBUG)
#l.addHandler(logging.StreamHandler())

from common.utils import regexes


def _get_utah_logs(artifacts):
    """
    Check if there are files in 'clientlogs'.
    Returns a list of utah logs found.
    """

    logs = []

    for artifact in artifacts:
        path = artifact['relativePath']
        utah_log_regex = re.compile(ur'^(clientlogs\/|)utah.*.yaml')
        logging.debug(path)
        if utah_log_regex.match(path):
            logs.append(artifact)

    return logs


def _process_utah_logs(logs, jenkins_build=None, jenkins_url=None, name=None):

    runs = []
    builds = []
    results = []

    for log in logs:
        logging.debug("log: {}".format(log))
        run, build, result = process_smoke_log(
            log['relativePath'],
            jenkins_build=jenkins_build,
            jenkins_url=jenkins_url,
            name=name,
        )

        if run:
            runs.append(run)
        if build:
            builds.append(build)
        if result:
            results.append(result)

    return runs, builds, results


def _parse_bugs(lp_bugs, result):
    """
    Add bugs to the result.
    """

    for bug_no in lp_bugs:
        bug = None
        try:
            bug = Bug.objects.get(bug_no=bug_no)
        except Bug.DoesNotExist:
            bug = Bug.objects.create(bug_no=bug_no, status='unknown')
        if bug not in list(result.bugs.all()):
            result.bugs.add(bug)
            result.save()


def _parse_logs(artifacts, build_url, result):
    """
    Add logs to result.
    """

    for artifact in artifacts:
        path = artifact['relativePath']

        # skip artifacts we don't care about.
        if regexes['artifact-ignore'].match(path):
            continue

        remote_url = "%s/artifact/%s" % (build_url, path)
        art_args = {
            'display_name': path,
            'remote_url': remote_url,
            'path': path,
        }
        res_log, new_res = result.resultlog_set.get_or_create(**art_args)


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option(
            '-I', '--initial-run',
            action='store_true',
            dest='initial_run',
            default=False,
            help='Initial run processes all data (takes longer)'
        ),
        make_option(
            '-w', '--wait-for',
            dest='waitfor',
            help='Skip all jobs until the named job is found'
        ),
    )

    args = "[job name] [job name]..."
    help = 'Pull data from jenkins'

    def get_build_number(self, build_url):
        media_info_url = "{}artifact/media_info".format(build_url)
        try:
            response = jenkins_get(media_info_url, as_json=False)
        except urllib2.HTTPError:
            response = ""

        m = regexes['build-number'].match(response)

        if m:
            return m.group(1)
        else:
            return None

    def handle(self, *args, **options):
        verbosity = int(options.get('verbosity'))
        if verbosity > 1:
            logging.basicConfig(level=logging.DEBUG)

        job_list = list(args)
        logging.debug("job_list: {}".format(job_list))

        starttime = datetime.now()
        initial_run = options.get('initial_run')
        waitfor = options.get('waitfor')

        logging.debug("waitfor: {}".format(waitfor))

        # for the initial run include older releases as well.
        if initial_run:
            regexes['smoke'] = re.compile(
                ur"^(lucid|natty|oneiric|precise|quantal)-(desktop|server|alternate)"
            )

        jobs_top = jenkins_get('https://jenkins.qa.ubuntu.com/api/json')
        jobs = jobs_top['jobs']

        job_count = 0  # XXX: DEBUG
        waiting = waitfor is not None
        logging.info("waiting for job {}".format(waitfor))
        for job in jobs:
            utah_job = False
            name = job['name']
            url = job['url']

            # skip jobs until the one we're waiting for
            if waiting:
                if name == waitfor:
                    logging.info("found job {}".format(waitfor))
                    waiting = False
                else:
                    continue

            # if passed a list of jobs skip those that don't match
            if len(job_list) > 0 and name not in job_list:
                continue

            if regexes['smoke'].match(name) or regexes['utah-smoke'].match(name):
                logging.info("Smoke test %s" % name)
                flavor = 'ubuntu'
                if regexes['smoke-ec2'].match(name):
                    if regexes['smoke-ec2-daily'].match(name):
                        m = regexes['smoke-ec2-daily-data'].match(name)
                        release, variant = m.group(1, 2)
                else:
                    # first check if it's a utah based job
                    m = regexes['utah-smoke-data'].match(name)
                    orig_name = name
                    if m:
                        release, variant, arch, name = m.group(1, 2, 3, 4)
                        logging.debug(
                            "Got a UTAH job: %s %s %s %s",
                            release,
                            variant,
                            arch,
                            name
                        )
                        utah_job = True
                    else:
                        m = regexes['smoke-data'].match(name)
                        if m:
                            release, variant, arch, name = m.group(1, 2, 3, 4)
                        else:
                            logging.warn("Strange smoke name: %s", name)
                            continue

                    # XXX: debugging limit
                    job_count += 1
                    if DEBUG_LIMITS and job_count > JOB_COUNT_MAX:
                        logging.info("Stopping due to DEBUG_LIMITS")
                        break

                    logging.info(
                        "Release: %s, variant: %s, arch: %s, name: %s",
                        release,
                        variant,
                        arch,
                        name
                    )

                    # Ignore static validation per Gema's request via IRC.
                    if name == "static_validation":
                        logging.info(
                            "Skipping static validation test: %s",
                            orig_name
                        )
                        continue

                    # XXX: See if adding '?depth=1' makes this faster.

                    job_info = jenkins_get("%s/api/json/?depth=1" % url)
                    builds = job_info['builds']
                    last_build = datetime.fromtimestamp(
                        job_info['lastBuild']['timestamp']/1000
                    )

                    # for update runs only check jobs that have built in the
                    # last week.
                    if not initial_run and datetime.now() - last_build > timedelta(days=7):
                        continue

                    build_count = 0  # XXX: DEBUG
                    for build in builds:
                        if skip_build(build):
                            logging.info(
                                "skipping build: %s %s",
                                job['name'],
                                build['number']
                            )
                            continue

                        # XXX: debugging limit
                        build_count += 1
                        if DEBUG_LIMITS and build_count > BUILD_COUNT_MAX:
                            logging.info("Stopping due to DEBUG_LIMITS")
                            break

                        build_number = build['number']
                        build_url = build['url']
                        #build_info = jenkins_get("%s/api/json" % build_url)
                        building = build['building']
                        build_desc = build['description']
                        build_date = datetime.fromtimestamp(
                            build['timestamp']/1000
                        )
                        artifacts = build['artifacts']
                        actions = build['actions']

                        # Skip this Jenkins build if it hasn't finished yet
                        if building:
                            continue

                        # XXX: find out what TZ jenkins uses and use correct one here
                        if not initial_run and datetime.now() - build_date > timedelta(days=7):
                            continue

                        build_no = self.get_build_number(build_url)

                        lp_bugs = []

                        if build_desc != None:
                            b = build_desc.split(',')
                            b = regexes['build_desc'].findall(build_desc)
                            for word in b:
                                word = word.strip()
                                if regexes['buildid'].match(word) and build_no is None:
                                    # It's a build number
                                    # XXX: is this still being used, if not remove.
                                    build_no = "{} ?".format(word)
                                elif regexes['lpbug'].match(word):
                                    lp_bugs.append(word[4:])

                        if utah_job:  # A utah job
                            utah_logs = _get_utah_logs(artifacts)

                            if len(utah_logs) > 0:
                                runs, builds, results = _process_utah_logs(
                                    utah_logs,
                                    jenkins_build=build_number,
                                    jenkins_url=url,
                                    name=name,
                                )

                                for result in results:
                                    _parse_bugs(lp_bugs, result)
                                    _parse_logs(artifacts, build_url, result)

                            # If there were no logs or parsing failed
                            # create a zero entry.
                            if len(utah_logs) == 0 or len(results) == 0:
                                # a job without utah logs
                                # create an empty entry
                                run, new_run = Run.objects.get_or_create(
                                    release=release,
                                    flavor=flavor,
                                    build_no=build_no,
                                    defaults={
                                        'ran_at': build_date,
                                        'test_type': 0,
                                    },
                                )

                                build, new_build = run.build_set.get_or_create(
                                    variant=variant,
                                    flavor=flavor,
                                    arch=arch,
                                )

                                result, new_result = build.result_set.get_or_create(
                                    name=name,
                                    jenkins_build=build_number,
                                    defaults={
                                        'ran_at': build_date,
                                        'jenkins_url': build_url,
                                        'fail_count': 0,
                                        'error_count': 0,
                                        'pass_count': 0,
                                        'total_count': 0,
                                    },
                                )
                                logging.info(
                                    "No valid utah logs found, creating empty result..."
                                )
                                _parse_bugs(lp_bugs, result)
                                _parse_logs(artifacts, build_url, result)

                        # if all else fails set to the build_date
                        if build_no is None:
                            build_no = build_date.strftime("%Y%m%d ?")

                        logging.info(
                            "Build #%s, date: %s, build_no: %s",
                            build_number,
                            build_date,
                            build_no
                        )

                        result = None
                        if not utah_job:
                            logging.debug("Processing results as jenkins job")
                            # Start adding objects here.

                            run, new_run = Run.objects.get_or_create(
                                release=release,
                                flavor=flavor,
                                build_no=build_no,
                                defaults={
                                    'ran_at': build_date,
                                    'test_type': 0,
                                },
                            )

                            build, new_build = run.build_set.get_or_create(
                                variant=variant,
                                flavor=flavor,
                                arch=arch,
                            )

                            try:
                                result = build.result_set.get(
                                    name=name,
                                    jenkins_build=build_number
                                )
                                # break
                            except Result.DoesNotExist:
                                fail_count, error_count, pass_count, total_count = (0, 0, 0, 0)
                                for action in actions:
                                    try:
                                        fail_count, total_count = (
                                            action['failCount'],
                                            action['totalCount']
                                        )
                                        pass_count = total_count - fail_count
                                        break
                                    except KeyError:
                                        continue

                                result = build.result_set.create(
                                    name=name,
                                    jenkins_build=build_number,
                                    ran_at=build_date,
                                    jenkins_url=build_url,
                                    fail_count=fail_count,
                                    error_count=0,
                                    pass_count=pass_count,
                                    total_count=total_count
                                )

                            _parse_bugs(lp_bugs, result)
                            _parse_logs(artifacts, build_url, result)

        endtime = datetime.now()
        duration = endtime - starttime
        logging.info("time: %s", duration)
