# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.db import models

from common.models import DashboardBaseModel, Bug


class ResultLog(DashboardBaseModel):
    class Meta:
        db_table = 'result_logs'
    display_name = models.CharField(max_length=255, null=True)
    path = models.TextField()
    result = models.ForeignKey('Result')
    remote_url = models.URLField()

    def __unicode__(self):
        return self.path


class Run(DashboardBaseModel):
    class Meta:
        db_table = 'runs'

    build_no = models.CharField(max_length=200, db_index=True)
    release = models.CharField(max_length=200, db_index=True)
    flavor = models.CharField(max_length=200, db_index=True)
    test_type = models.IntegerField(null=True)
    ran_at = models.DateTimeField('date run')

    def __unicode__(self):
        return "{} - {}".format(
            self.release,
            self.build_no,
        )

    # Used by MySerializer to add property fields to the JSON representation
    # of an instance of this class.
    extra_fields = [
        'pass_count',
        'fail_count',
        'error_count',
        'total_count',
        'pass_rate',
        'pass_rate_pct',
        'bug_count',
    ]

    @property
    def pass_count(self):
        passes = 0
        for build in self.build_set.all():
            passes += build.pass_count
        return passes

    @property
    def fail_count(self):
        fails = 0
        for build in self.build_set.all():
            fails += build.fail_count
        return fails

    @property
    def error_count(self):
        errors = 0
        for build in self.build_set.all():
            errors += build.error_count
        return errors

    @property
    def total_count(self):
        total = 0
        for build in self.build_set.all():
            total += build.total_count
        return total

    @property
    def pass_rate(self):
        total = 0.0
        count = 0
        for build in self.build_set.all():
            total += build.pass_rate
            count += 1

        return total / count if count > 0 else 0

    @property
    def pass_rate_pct(self):
        return "%d%%" % int(self.pass_rate*100)

    @property
    def bug_count(self):
        bugs = 0
        for build in self.build_set.all():
            bugs = bugs + build.bug_count
        return bugs

    @property
    def bugs(self):
        bugs = Bug.objects.filter(
            result_bugs__build__run__id=self.id
        ).values('bug_no').distinct()

        return bugs


class Build(DashboardBaseModel):
    class Meta:
        db_table = 'builds'
    run = models.ForeignKey(Run)
    name = models.CharField(max_length=200, null=True)
    flavor = models.CharField(max_length=200)
    variant = models.CharField(max_length=200)
    arch = models.CharField(max_length=200)

    def __unicode__(self):
        return "{} - {} - {} - {}".format(
            self.flavor,
            self.variant,
            self.arch,
            self.run,
        )

    # Used by MySerializer to add property fields to the JSON representation
    # of an instance of this class.
    extra_fields = [
        'pass_count',
        'fail_count',
        'error_count',
        'total_count',
        'pass_rate',
        'pass_rate_pct',
        'bug_count',
    ]

    @property
    def debug_str(self):
        return "{}-{}-{}-{}-{}-{}".format(
            self.id,
            self.name,
            self.flavor,
            self.arch,
            self.created_at,
            self.updated_at,
        )

    @property
    def pass_count(self):
        return self.result_set.aggregate(
            models.Sum('pass_count')
        )['pass_count__sum'] or 0

    @property
    def fail_count(self):
        return self.result_set.aggregate(
            models.Sum('fail_count')
        )['fail_count__sum'] or 0

    @property
    def error_count(self):
        return self.result_set.aggregate(
            models.Sum('error_count')
        )['error_count__sum'] or 0

    @property
    def total_count(self):
        return self.result_set.aggregate(
            models.Sum('total_count')
        )['total_count__sum'] or 0

    @property
    def pass_rate(self):
        total = 0.0
        count = 0
        for result in self.result_set.all():
            total += result.pass_rate
            count += 1
        return total / count if count > 0 else 0

    @property
    def pass_rate_pct(self):
        return "%d%%" % int(self.pass_rate*100)

    @property
    def bug_count(self):
        return self.result_set.annotate(
            num_bugs=models.Count('bugs')
        ).aggregate(models.Sum('num_bugs'))['num_bugs__sum'] or 0

    @property
    def bugs(self):
        bugs = Bug.objects.filter(
            result_bugs__build__id=self.id
        ).values('bug_no').distinct()
        return bugs


class Result(DashboardBaseModel):
    class Meta:
        db_table = 'results'
    build = models.ForeignKey(Build)
    bugs = models.ManyToManyField(Bug, related_name="result_bugs")
    # django doesn't have proper polymorphic associations, so mis-use a
    # many-to-many relationship for the logs, even if each log can only
    # belong to one other object.
    #result_logs   = models.ManyToManyField(ResultLog)
    name = models.CharField(max_length=255)
    jenkins_build = models.CharField(max_length=255)
    jenkins_url = models.URLField()
    fail_count = models.IntegerField()
    error_count = models.IntegerField(null=False, default=0)
    pass_count = models.IntegerField()
    total_count = models.IntegerField()
    ran_at = models.DateTimeField('date run')

    def __unicode__(self):
        return "{} - {}".format(
            self.name,
            self.ran_at,
        )

    @property
    def debug_str(self):
        return "{}-{}-{}-{}-{}".format(
            self.id,
            self.name,
            self.build,
            self.created_at,
            self.updated_at,
        )

    # Used by MySerializer to add property fields to the JSON representation
    # of an instance of this class.
    extra_fields = [
        'pass_rate',
        'pass_rate_pct',
        'bug_count',
        'log_count',
    ]

    @property
    def pass_rate(self):
        if self.total_count > 0:
            return self.pass_count * 1.0 / self.total_count
        else:
            return 0

    @property
    def pass_rate_pct(self):
        return "%d%%" % int(self.pass_rate*100)

    @property
    def bug_count(self):
        return self.bugs.count()

    @property
    def log_count(self):
        return self.resultlog_set.count()
