# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.test import TestCase
from smoke.models import (
    Build,
    Result,
    ResultLog,
    Run,
)

from common.models import Bug


class TestSmoke(TestCase):
    """ Tests for the smoke models. """

    def setUp(self):
        """ Basic setup for each test. """

        self.run_without_results = Run.objects.create(
            build_no='20130101',
            release='raring',
            flavor='ubuntu',
            ran_at=datetime.datetime.now(),
        )

        self.assert_(self.run_without_results)

        self.build_without_results = Build.objects.create(
            run=self.run_without_results,
            flavor='ubuntu',
            variant='desktop',
            arch='amd64',
        )

        self.assert_(self.build_without_results)

        self.run_with_results = Run.objects.create(
            build_no='20130101',
            release='raring',
            flavor='ubuntu',
            ran_at=datetime.datetime.now(),
        )

        self.assert_(self.run_with_results)

        self.build_with_results = Build.objects.create(
            run=self.run_with_results,
            flavor='ubuntu',
            variant='desktop',
            arch='amd64',
        )

        self.assert_(self.build_with_results)

        self.result = Result.objects.create(
            build=self.build_with_results,
            name='default',
            jenkins_build='none',
            jenkins_url='none',
            fail_count=1,
            error_count=2,
            pass_count=3,
            total_count=6,
            ran_at=datetime.datetime.now(),
        )

        self.assert_(self.result)

        self.result_log = ResultLog.objects.create(
            path='somewhere',
            result=self.result,
            remote_url='http://jenkins.qa.ubuntu.com/',
        )

        self.assert_(self.result_log)

        self.bug = self.result.bugs.create(
            bug_no='123456',
            status='open',
        )

        self.assert_(self.bug)

    def test_run_default_values(self):
        """ Test default values for a simple Run model. """

        self.assertEqual(self.run_without_results.pass_count, 0)
        self.assertEqual(self.run_without_results.fail_count, 0)
        self.assertEqual(self.run_without_results.error_count, 0)
        self.assertEqual(self.run_without_results.total_count, 0)
        self.assertEqual(self.run_without_results.pass_rate, 0)
        self.assertEqual(self.run_without_results.pass_rate_pct, '0%')
        self.assertEqual(self.run_without_results.bug_count, 0)

    def test_build_default_values(self):
        """ Test default values for a simple Build model. """

        self.assertEqual(self.build_without_results.pass_count, 0)
        self.assertEqual(self.build_without_results.fail_count, 0)
        self.assertEqual(self.build_without_results.error_count, 0)
        self.assertEqual(self.build_without_results.total_count, 0)
        self.assertEqual(self.build_without_results.pass_rate, 0)
        self.assertEqual(self.build_without_results.pass_rate_pct, '0%')
        self.assertEqual(self.build_without_results.bug_count, 0)

    def test_result(self):
        """ Test statistics for a result. """

        self.assertEqual(self.result.pass_count, 3)
        self.assertEqual(self.result.error_count, 2)
        self.assertEqual(self.result.fail_count, 1)
        self.assertEqual(self.result.total_count, 6)
        self.assertEqual(self.result.log_count, 1)
        self.assertEqual(self.result.bug_count, 1)

    def test_build_with_results(self):
        """ Test aggregated stats for a Build. """

        self.assertEqual(self.build_with_results.pass_count, 3)
        self.assertEqual(self.build_with_results.error_count, 2)
        self.assertEqual(self.build_with_results.fail_count, 1)
        self.assertEqual(self.build_with_results.total_count, 6)
        self.assertEqual(self.build_with_results.bug_count, 1)

    def test_run_with_results(self):
        """ Test aggregated stats for a Run. """

        self.assertEqual(self.run_with_results.pass_count, 3)
        self.assertEqual(self.run_with_results.error_count, 2)
        self.assertEqual(self.run_with_results.fail_count, 1)
        self.assertEqual(self.run_with_results.total_count, 6)
        self.assertEqual(self.run_with_results.bug_count, 1)
