"""UTAH Client log parser."""
import argparse
import urllib2
import yaml


class ParserError(Exception):

    """Module level parse error exception."""

    def __init__(self, message, zero_entry=False):
        super(ParserError, self).__init__(message)
        self.zero_entry = zero_entry


class UTAHParser(object):

    """UTAH Client result parser class."""

    class ValidationError(Exception):

        """Class level validation exception."""

        pass

    class SoftValidationError(Exception):
        """Class level validation exception without error."""

        pass

    def _check_for_empty(self, data):
        """ Verify that enough data is available to indicate failure."""

        has_required_fields = True
        required_fields = [
            'errors',
            'failures',
            'passes',
        ]

        for field in required_fields:
            if not field in data:
                has_required_fields = False

        # If the required fields are present and all zeroes then
        # we have enough data to generate a blank result.
        if has_required_fields:
            if data['passes'] == 0 and data['failures'] == 0 and data['errors'] == 0:
                return True

        return False

    def validate(self, data):
        """ Verify that the required fields are in the dictionary. """

        missing_fields = []
        required_fields = [
            'runlist',
            'ran_at',
            'commands',
            'passes',
            'failures',
            'errors',
            'fetch_errors',
            'uname',
            'media-info',
            'install_type',
            'arch',
            'release',
            'build_number',
            'name',
        ]

        for field in required_fields:
            if not field in data:
                missing_fields.append(field)

        # Don't give errors for bad default YAML files since
        # we can handle generating an empty result ourselves.
        if len(missing_fields) > 0:
            missing_msg = ", ".join(missing_fields)
            if self._check_for_empty(data):
                exception = self.SoftValidationError(
                    "field(s) '{}' missing".format(missing_msg)
                )
            else:
                exception = self.ValidationError(
                    "field(s) '{}' missing".format(missing_msg)
                )
            raise exception

    def parse(self, logdata, as_string=False):
        """Parse utah client results.

        :param logdata:
            should be either a filename or a handle to a file object.

        :returns: the parsed yaml data or None.

        :raises: ParserError on exceptions.

        """

        if as_string:
            return self._parse_string(logdata)
        elif isinstance(logdata, str):
            return self._parse_logfile(logdata)
        else:
            return self._parse_stream(logdata)

    def _parse_stream(self, stream):
        """Parse client output from stream.

        :returns: validated decoded YAML data.

        :raises: ParseError

        """
        try:
            data = yaml.load(stream)
        except yaml.YAMLError as e:
            raise ParserError(e.message)

        # Allow empty yaml files
        if data is None:
            return data

        try:
            self.validate(data)
        except self.ValidationError as e:
            raise ParserError(e.message)
        except self.SoftValidationError as e:
            raise ParserError(e.message, zero_entry=True)

        return data

    def _parse_logfile(self, logfile):
        """Parse client output log.

        :returns: validated decoded YAML data.

        :raises: ParseError

        """
        data = None

        if logfile.startswith('http'):
            fp = urllib2.urlopen(logfile)
            data = self._parse_stream(fp)
        else:
            with open(logfile, 'r') as fp:
                data = self._parse_stream(fp)

        return data

    def _parse_string(self, log_data):
        """Parse client data from a string.

        :returns: validated decoded YAML data.

        :raises: ParseError

        """
        try:
            data = yaml.load(log_data)

            # Allow empty yaml files
            if data is None:
                return data
        except yaml.YAMLError as e:
            raise ParserError(e.message)

        try:
            self.validate(data)
        except self.ValidationError as e:
            raise ParserError(e.message)
        except self.SoftValidationError as e:
            raise ParserError(e.message, zero_entry=True)

        return data
