# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
UTAH result parsing code.
"""

import logging

from smoke.utah_parser import UTAHParser, ParserError
from smoke.models import (
    Build,
    Run,
    Result,
)

FLAVOR = 'ubuntu'


def process_smoke_log(
    logfile,
    jenkins_build=None,
    jenkins_url=None,
    name=None
):
    """
    Parse a utah client log for smoke test results.
    """

    run = None
    build = None
    result = None
    logfile_path = logfile

    if jenkins_url is not None and jenkins_build is not None:
        logfile_path = "{}{}/artifact/{}".format(
            jenkins_url,
            jenkins_build,
            logfile,
        )

    # set defaults
    if jenkins_build is None:
        jenkins_build = '0'

    if jenkins_url is None:
        jenkins_url = "http://jenkins.qa.ubuntu.com/"

    parser = UTAHParser()

    try:
        data = parser.parse(logfile_path)
    except ParserError as e:
        if not e.zero_entry:
            logging.error("{} {}".format(logfile_path, e))
        return run, build, result

    if data is None:
        logging.warn("Unable to parse {}".format(logfile_path))
        return run, build, result

    build_number = data['build_number']

    # Use the name from the logs if there is one, if there isn't one
    # and one is passed in use that.
    if 'name' not in data:
        data['name'] = 'unnamed'

    if name is None or data['name'] != 'unnamed':
        name = data['name']

    arch = data['arch']
    ran_at = data['ran_at']

    release = data['release']
    install_type = data['install_type']

    passes = data['passes']
    failures = data['failures']
    errors = data['errors']

    logging.debug("data.keys(): {}".format(data.keys()))

    logging.debug("name: {}".format(name))
    logging.debug("arch: {}".format(arch))
    logging.debug("ran_at: {}".format(ran_at))
    logging.debug("release: {}".format(release))
    logging.debug("build_number: {}".format(build_number))
    logging.debug("install_type: {}".format(install_type))

    run, new_run = Run.objects.get_or_create(
        build_no=build_number,
        release=release,
        flavor=FLAVOR,
        defaults={
            'ran_at': ran_at,
        },
    )

    logging.debug("run: {}".format(run))
    logging.debug("new_run: {}".format(new_run))

    build, new_build = Build.objects.get_or_create(
        run=run,
        flavor=FLAVOR,
        variant=install_type,
        arch=arch,
    )

    logging.debug("build: {}".format(build))
    logging.debug("new_build: {}".format(new_build))

    result, new_result = Result.objects.get_or_create(
        build=build,
        name=name,
        fail_count=failures,
        pass_count=passes,
        total_count=failures + passes + errors,
        ran_at=ran_at,
        #XXX: need to be provided from outside
        error_count=errors,
        jenkins_build=jenkins_build,
        jenkins_url="{}{}".format(jenkins_url, jenkins_build),
    )

    logging.debug("result: {}".format(result))
    logging.debug("new_result: {}".format(new_result))

    return run, build, result
