# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import matplotlib

from django.http import HttpResponse, Http404
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.views.decorators.http import require_GET

# XXX: This is needed due to a yet to be found import ordering bug.
matplotlib.use('Agg')
from matplotlib.backends.backend_agg import FigureCanvasAgg
from pylab import figure, pie

from smoke.models import (
    Build,
    Result,
    Run,
)


def _get_releases():
    releases = []
    for run in Run.objects.values(
        'release'
    ).order_by('-release').distinct().exclude(publish=False):
        releases.append(run['release'])

    return releases


@require_GET
def index(request):
    release = 'raring'
    data = {
        'release': release,
        'releases': _get_releases(),
        'is_flat': True,
        'active': release,
    }
    return _process_flat_smoke_overview(request, release, data, limit=100)


@require_GET
def overview(request, release, is_flat=False):
    data = {
        'release': release,
        'releases': _get_releases(),
        'is_flat': False,
        'active': release,
    }

    if request.GET.get('flat') is not None:
        is_flat = True

    if is_flat:
        return _process_flat_smoke_overview(request, release, data)
    else:
        return _process_combined_smoke_overview(request, release, data)


def _process_flat_smoke_overview(request, release, data, limit=None):
    data['is_flat'] = True
    if limit is not None:
        data['index'] = True
        build = Build.objects.filter(
            run__release=release
        ).order_by(
            '-run__build_no'
        ).select_related().exclude(publish=False)
        count = build.count()
        data['items'] = build[:limit]
        if count > limit:
            data['more'] = True
    else:
        data['items'] = Build.objects.filter(
            run__release=release
        ).select_related().exclude(publish=False)
    return _process_overview(request, data)


def _process_combined_smoke_overview(request, release, data):
    data['items'] = Run.objects.filter(
        release=release
    ).exclude(publish=False)
    return _process_overview(request, data)


def _process_overview(request, data):
    return render_to_response(
        'smoke/overview.html',
        data,
        RequestContext(request)
    )


@require_GET
def build_overview(request, release, run_id):
    try:
        run = Run.objects.exclude(publish=False).get(id=run_id)
    except Run.DoesNotExist:
        raise Http404

    # Again, eager loading of associations two levels deeper would be great.
    builds = run.build_set.all().exclude(publish=False)
    return render_to_response(
        'smoke/build_overview.html',
        locals(),
        RequestContext(request)
    )


@require_GET
def results(request, release, run_id, image_id):
    try:
        image = Build.objects.exclude(
            publish=False
        ).select_related().get(id=image_id)
    except Build.DoesNotExist:
        raise Http404

    results = image.result_set.all().exclude(publish=False)
    return render_to_response(
        'smoke/results.html',
        locals(),
        RequestContext(request)
    )


@require_GET
def results_logs(request, release, run_id, image_id, result_id):
    try:
        image = Build.objects.exclude(
            publish=False
        ).select_related().get(id=image_id)
        result = Result.objects.exclude(publish=False).get(id=result_id)
    except Result.DoesNotExist, Build.DoesNotExist:
        raise Http404

    logs = result.resultlog_set.all().exclude(publish=False)

    return render_to_response(
        'smoke/result_logs.html',
        locals(),
        RequestContext(request)
    )


def _make_pie_chart(obj):
    f = figure(figsize=(6, 6))  # in inches... (dpi=80)
    f.patch.set_alpha(0.0)  # Transparent background

    pre_data = [
        ('Passed', obj.pass_rate, 0.1, '#abffab'),
        ('Failed', 1.0 - obj.pass_rate, 0.1, '#ffabab'),
    ]

    data = []
    labels = []
    explode = []
    colors = []
    for label, value, expl, color in pre_data:
        if value > 0.0:
            data.append(value)
            labels.append(label)
            explode.append(expl)
            colors.append(color)

    pie(
        data,
        explode=explode,
        labels=labels,
        colors=colors,
        autopct='%1.1f%%',
        shadow=True
    )
    response = HttpResponse(content_type='image/png')
    canvas = FigureCanvasAgg(f)
    canvas.print_png(response, transparent=True)
    matplotlib.pyplot.close(f)
    return response


@require_GET
def build_pie(request, release, run_id):
    try:
        run = Run.objects.exclude(publish=False).get(id=run_id)
    except Run.DoesNotExist:
        raise Http404

    return _make_pie_chart(run)


@require_GET
def results_pie(request, release, run_id, image_id):
    try:
        image = Build.objects.exclude(publish=False).get(id=image_id)
    except Build.DoesNotExist:
        raise Http404

    return _make_pie_chart(image)
