# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from datetime import datetime, timedelta
from optparse import make_option
from django.core.management.base import BaseCommand

from common.management import (
    jenkins_get,
    skip_build,
    BUILD_COUNT_MAX,
    JOB_COUNT_MAX,
    DEBUG_LIMITS,
)

from common.utils import regexes
from common.models import Bug
from sru.models import Kernel


def _get_kernel_version(description, default="Unknown"):
    """
    Parse a jenkins build description for a kernel version
    """
    version = default

    if description is None:
        description = ''

    r = regexes['kernel_sru-version']

    m = r.match(description)

    if m:
        version = m.group(1)

    return version


def _parse_bugs(lp_bugs, result):
    """
    Add bugs to the sruresult.
    """
    sruresult = result
    status = 'unknown'
    for bug_no in lp_bugs:
        bug = None
        bug, new_bug = Bug.objects.get_or_create(
            bug_no=bug_no,
            defaults={'status': status}
        )
        if bug not in list(sruresult.bugs.all()):
            sruresult.bugs.add(bug)
            sruresult.save()


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option(
            '-I', '--initial-run',
            action='store_true',
            dest='initial_run',
            default=False,
            help='Initial run processes all data (takes longer)'
        ),
    )

    help = 'Pull data from jenkins'

    def handle(self, *args, **options):
        starttime = datetime.now()
        initial_run = options.get('initial_run')
        verbosity = int(options.get('verbosity'))
        if verbosity > 1:
            logging.basicConfig(level=logging.DEBUG)

        url = "https://jenkins.qa.ubuntu.com/api/json"
        jobs_top = jenkins_get(url)

        jobs = jobs_top['jobs']

        job_count = 0  # XXX: DEBUG
        for job in jobs:
            name = job['name']
            m = regexes['kernel_sru'].match(name)
            if m:
                pre_job_type = m.group(1)
                post_job_type = m.group(2)
                job_type = "{}{}".format(pre_job_type, post_job_type)

                pre_release = m.group(3)
                post_release = m.group(4)
                release = "{}{}".format(pre_release, post_release)

                variant, arch, gpu = m.group(5, 6, 7)

                # XXX: debugging limit
                job_count += 1
                if DEBUG_LIMITS and job_count > JOB_COUNT_MAX:
                    break

                logging.info("job: {}\n".format(name))
                logging.info("\tjob_type: {}\n".format(job_type))
                logging.info("\trelease: {}\n".format(release))
                logging.info("\tkernel: {}\n".format(variant))
                logging.info("\tarch: {}\n".format(arch))
                logging.info("\tgpu: {}\n".format(gpu))

                url = job['url']
                job_info = jenkins_get("{}/api/json/?depth=1".format(url))
                builds = job_info['builds']

                # ignore jobs that haven't been run yet.
                if job_info['lastBuild'] is None:
                    continue
                last_build = datetime.fromtimestamp(
                    job_info['lastBuild']['timestamp']/1000
                )
                # for update runs only check jobs that have built in the
                # last week.
                if not initial_run and datetime.now() - last_build > timedelta(
                    days=7
                ):
                    continue

                build_count = 0  # XXX: DEBUG
                for build in builds:
                    if skip_build(build):
                        logging.info(
                            "skipping build: %s %s",
                            job['name'],
                            build['number']
                        )
                        continue

                    # XXX: debugging limit
                    build_count += 1
                    if DEBUG_LIMITS and build_count > BUILD_COUNT_MAX:
                        break

                    build_number = build['number']
                    build_url = build['url']
                    building = build['building']
                    build_desc = build['description']
                    build_date = datetime.fromtimestamp(
                        build['timestamp']/1000
                    )
                    actions = build['actions']

                    if build_desc is None:
                        build_desc = ''

                    kernel_version = _get_kernel_version(
                        build_desc,
                        "Unknown. Ran on {}".format(
                            build_date.strftime("%Y-%m-%d %H:%M")
                        )
                    )

                    # Skip this Jenkins build if it hasn't finished yet
                    if building:
                        continue

                    # XXX: find out what TZ jenkins uses
                    # and use correct one here
                    too_old = datetime.now() - build_date > timedelta(days=7)
                    if not initial_run and too_old:
                        continue

                    lp_bugs = []

                    if build_desc is not None:
                        b = build_desc.split(',')
                        b = regexes['build_desc'].findall(build_desc)
                        for word in b:
                            word = word.strip()
                            regex = regexes['buildid'].match(word)
                            if regex and build_number is None:
                                # It's a build number
                                build_number = "{} ?".format(word)
                            elif regexes['lpbug'].match(word):
                                lp_bugs.append(word[4:])

                    kernel, new_kernel = Kernel.objects.get_or_create(
                        release=release,
                        variant=variant,
                        arch=arch,
                        gpu=gpu,
                        version=kernel_version,
                        job_type=job_type,
                    )

                    fail_count = 0
                    skip_count = 0
                    pass_count = 0
                    total_count = 0
                    for action in actions:
                        try:
                            fail_count = action['failCount']
                            skip_count = action['skipCount']
                            total_count = action['totalCount']
                            pass_count = total_count - fail_count - skip_count
                            break
                        except KeyError:
                            continue
                    result, new_result = kernel.sruresult_set.get_or_create(
                        name=name,
                        jenkins_build=build_number,
                        defaults={
                            'ran_at': build_date,
                            'jenkins_url': build_url,
                            'fail_count': fail_count,
                            'skip_count': skip_count,
                            'pass_count': pass_count,
                            'total_count': total_count,
                        }
                    )
                    _parse_bugs(lp_bugs, result)
        logging.debug("time: %s", datetime.now() - starttime)
