# QA Dashboard
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.shortcuts import render_to_response
from django.views.decorators.http import require_GET
from django.template import RequestContext
from datetime import date, timedelta

from sru.models import Kernel
from common.utils import (
    JSONResponse,
    MySerializer,
    get_verbose,
)


def _get_kernel_names():
    kernels = []
    for kernel in Kernel.objects.values('variant').distinct().exclude(
        publish=False
    ).order_by('variant'):
        kernels.append(kernel['variant'])

    return kernels


def _get_kernel_releases():
    releases = []
    for kernel in Kernel.objects.values('release').distinct().exclude(
        publish=False
    ).order_by('-release'):
        releases.append(kernel['release'])

    return releases


@require_GET
def api_overview(request, release='', version=''):
    kernels = Kernel.objects.all()

    if release != '':
        kernels = kernels.filter(release=release)

    if version != '':
        kernels = kernels.filter(version=version)

    serializer = MySerializer(use_extra_fields=get_verbose(request, True))
    data = serializer.serialize(kernels)

    return JSONResponse(data)


@require_GET
def overview(request, release=''):
    data = {
        'releases': _get_kernel_releases(),
        'active': release,
        'show_unknown': True
    }

    past_date = date.today() - timedelta(days=30)

    data['show_unknown'] = True
    show_unknown = request.GET.get('show_unknown')

    if show_unknown is None or show_unknown != "1":
        data['show_unknown'] = False
        data['kernels'] = Kernel.objects.exclude(
            version__startswith="Unknown"
        ).filter(
            sruresult__ran_at__gte=past_date
        ).select_related().exclude(publish=False)
    else:
        data['kernels'] = Kernel.objects.select_related().exclude(
            publish=False
        ).filter(sruresult__ran_at__gte=past_date)

    if release != '':
        data['kernels'] = data['kernels'].filter(release=release)

    return render_to_response(
        'sru/overview.html',
        data,
        RequestContext(request)
    )


@require_GET
def overview_kernel(request, release=None, kernel_version=None):
    kernels = Kernel.objects.filter(version=kernel_version).select_related()

    data = {
        'kernels': kernels,
        'kernel_version': kernel_version,
        'release': release
    }

    totals = {
        'skip_count': 0,
        'pass_count': 0,
        'fail_count': 0,
        'total_count': 0,
        'pass_rate': 0.0,
    }

    for k in kernels:
        totals['skip_count'] += k.skip_count
        totals['pass_count'] += k.pass_count
        totals['fail_count'] += k.fail_count
        totals['total_count'] += k.total_count

    if totals['total_count'] == 0:
        totals['pass_rate'] = 0.0
    else:
        totals['pass_rate'] = totals['pass_count'] * 1.0
        totals['pass_rate'] = totals['pass_rate'] / totals['total_count']
    totals['pass_rate_pct'] = "{:.2%}".format(totals['pass_rate'])

    data['totals'] = totals

    return render_to_response(
        'sru/overview_kernel.html',
        data,
        RequestContext(request)
    )
